/* Generated by Cython 0.29.21 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "name": "SpellChecker.SimpleSpellChecker",
        "sources": [
            "SpellChecker/SimpleSpellChecker.pyx"
        ]
    },
    "module_name": "SpellChecker.SimpleSpellChecker"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__SpellChecker__SimpleSpellChecker
#define __PYX_HAVE_API__SpellChecker__SimpleSpellChecker
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "SpellChecker/SimpleSpellChecker.pyx",
  "stringsource",
  "Word.pxd",
  "Sentence.pxd",
  "TrieNode.pxd",
  "TxtWord.pxd",
  "Trie.pxd",
  "Dictionary.pxd",
  "TxtDictionary.pxd",
  "LRUCache.pxd",
  "InflectionalGroup.pxd",
  "MorphologicalParse.pxd",
  "State.pxd",
  "FsmParse.pxd",
  "FiniteStateMachine.pxd",
  "MetamorphicParse.pxd",
  "FsmMorphologicalAnalyzer.pxd",
  "Language.pxd",
  "TurkishLanguage.pxd",
  "FsmParseList.pxd",
  "SpellChecker/Candidate.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_10Dictionary_4Word_Word;
struct __pyx_obj_6Corpus_8Sentence_Sentence;
struct __pyx_obj_10Dictionary_4Trie_8TrieNode_TrieNode;
struct __pyx_obj_10Dictionary_7TxtWord_TxtWord;
struct __pyx_obj_10Dictionary_4Trie_4Trie_Trie;
struct __pyx_obj_10Dictionary_10Dictionary_Dictionary;
struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary;
struct __pyx_obj_13DataStructure_5Cache_8LRUCache_LRUCache;
struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup;
struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse;
struct __pyx_obj_21MorphologicalAnalysis_5State_State;
struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse;
struct __pyx_obj_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine;
struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse;
struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer;
struct __pyx_obj_12SpellChecker_12SpellChecker_SpellChecker;
struct __pyx_obj_8Language_8Language_Language;
struct __pyx_obj_8Language_15TurkishLanguage_TurkishLanguage;
struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList;
struct __pyx_obj_12SpellChecker_9Candidate_Candidate;
struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker;
struct __pyx_opt_args_10Dictionary_4Trie_8TrieNode_8TrieNode_addWord;

/* "Dictionary/Trie/TrieNode.pxd":9
 *     cdef set __words
 * 
 *     cpdef addWord(self, str word, Word root, int index=*)             # <<<<<<<<<<<<<<
 *     cpdef TrieNode getChild(self, char ch)
 *     cpdef set getWords(self)
 */
struct __pyx_opt_args_10Dictionary_4Trie_8TrieNode_8TrieNode_addWord {
  int __pyx_n;
  int index;
};
struct __pyx_opt_args_21MorphologicalAnalysis_18FiniteStateMachine_18FiniteStateMachine_addTransition;

/* "MorphologicalAnalysis/FiniteStateMachine.pxd":12
 *     cpdef list getStates(self)
 *     cpdef State getState(self, str name)
 *     cpdef addTransition(self, State fromState, State toState, str _with, str withName, toPos=*)             # <<<<<<<<<<<<<<
 *     cpdef list getTransitions(self, State state)
 */
struct __pyx_opt_args_21MorphologicalAnalysis_18FiniteStateMachine_18FiniteStateMachine_addTransition {
  int __pyx_n;
  PyObject *toPos;
};
struct __pyx_opt_args_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_24FsmMorphologicalAnalyzer_morphologicalAnalysisRoot;

/* "MorphologicalAnalysis/FsmMorphologicalAnalyzer.pxd":34
 *     cpdef bint __parseExists(self, list fsmParse, str surfaceForm)
 *     cpdef list __parseWord(self, list fsmParse, maxLengthOrSurfaceForm)
 *     cpdef list morphologicalAnalysisRoot(self, str surfaceForm, TxtWord root, state=*)             # <<<<<<<<<<<<<<
 *     cpdef list generateAllParses(self, TxtWord root, int maxLength)
 *     cpdef str replaceRootWord(self, FsmParse parse, TxtWord newRoot)
 */
struct __pyx_opt_args_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_24FsmMorphologicalAnalyzer_morphologicalAnalysisRoot {
  int __pyx_n;
  PyObject *state;
};

/* "Dictionary/Word.pxd":1
 * cdef class Word:             # <<<<<<<<<<<<<<
 * 
 *     cdef str name
 */
struct __pyx_obj_10Dictionary_4Word_Word {
  PyObject_HEAD
  struct __pyx_vtabstruct_10Dictionary_4Word_Word *__pyx_vtab;
  PyObject *name;
};


/* "Corpus/Sentence.pxd":4
 * 
 * 
 * cdef class Sentence:             # <<<<<<<<<<<<<<
 * 
 *     cdef list words
 */
struct __pyx_obj_6Corpus_8Sentence_Sentence {
  PyObject_HEAD
  struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *__pyx_vtab;
  PyObject *words;
};


/* "Dictionary/Trie/TrieNode.pxd":4
 * 
 * 
 * cdef class TrieNode:             # <<<<<<<<<<<<<<
 * 
 *     cdef dict __children
 */
struct __pyx_obj_10Dictionary_4Trie_8TrieNode_TrieNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_10Dictionary_4Trie_8TrieNode_TrieNode *__pyx_vtab;
  PyObject *__pyx___children;
  PyObject *__pyx___words;
};


/* "Dictionary/TxtWord.pxd":4
 * 
 * 
 * cdef class TxtWord(Word):             # <<<<<<<<<<<<<<
 * 
 *     cdef list __flags
 */
struct __pyx_obj_10Dictionary_7TxtWord_TxtWord {
  struct __pyx_obj_10Dictionary_4Word_Word __pyx_base;
  PyObject *__pyx___flags;
  PyObject *morphology;
};


/* "Dictionary/Trie/Trie.pxd":6
 * 
 * 
 * cdef class Trie:             # <<<<<<<<<<<<<<
 * 
 *     cdef TrieNode __root_node
 */
struct __pyx_obj_10Dictionary_4Trie_4Trie_Trie {
  PyObject_HEAD
  struct __pyx_vtabstruct_10Dictionary_4Trie_4Trie_Trie *__pyx_vtab;
  struct __pyx_obj_10Dictionary_4Trie_8TrieNode_TrieNode *__pyx___root_node;
};


/* "Dictionary/Dictionary.pxd":4
 * 
 * 
 * cdef class Dictionary:             # <<<<<<<<<<<<<<
 * 
 *     cdef list words
 */
struct __pyx_obj_10Dictionary_10Dictionary_Dictionary {
  PyObject_HEAD
  struct __pyx_vtabstruct_10Dictionary_10Dictionary_Dictionary *__pyx_vtab;
  PyObject *words;
  PyObject *filename;
  PyObject *comparator;
};


/* "Dictionary/TxtDictionary.pxd":7
 * 
 * 
 * cdef class TxtDictionary(Dictionary):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict __misspelled_words
 */
struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary {
  struct __pyx_obj_10Dictionary_10Dictionary_Dictionary __pyx_base;
  PyObject *__pyx___misspelled_words;
};


/* "DataStructure/Cache/LRUCache.pxd":1
 * cdef class LRUCache(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef int __cacheSize
 */
struct __pyx_obj_13DataStructure_5Cache_8LRUCache_LRUCache {
  PyObject_HEAD
  struct __pyx_vtabstruct_13DataStructure_5Cache_8LRUCache_LRUCache *__pyx_vtab;
  int __pyx___cacheSize;
  PyObject *__pyx___map;
};


/* "MorphologicalAnalysis/InflectionalGroup.pxd":1
 * cdef class InflectionalGroup:             # <<<<<<<<<<<<<<
 * 
 *     cdef list __IG
 */
struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup {
  PyObject_HEAD
  struct __pyx_vtabstruct_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *__pyx_vtab;
  PyObject *__pyx___IG;
};


/* "MorphologicalAnalysis/MorphologicalParse.pxd":6
 * 
 * 
 * cdef class MorphologicalParse:             # <<<<<<<<<<<<<<
 * 
 *     cdef list inflectional_groups
 */
struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse {
  PyObject_HEAD
  struct __pyx_vtabstruct_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *__pyx_vtab;
  PyObject *inflectional_groups;
  struct __pyx_obj_10Dictionary_4Word_Word *root;
};


/* "MorphologicalAnalysis/State.pxd":1
 * cdef class State:             # <<<<<<<<<<<<<<
 * 
 *     cdef bint __start_state, __end_state
 */
struct __pyx_obj_21MorphologicalAnalysis_5State_State {
  PyObject_HEAD
  struct __pyx_vtabstruct_21MorphologicalAnalysis_5State_State *__pyx_vtab;
  int __pyx___start_state;
  int __pyx___end_state;
  PyObject *__pyx___name;
  PyObject *__pyx___pos;
};


/* "MorphologicalAnalysis/FsmParse.pxd":5
 * 
 * 
 * cdef class FsmParse(MorphologicalParse):             # <<<<<<<<<<<<<<
 * 
 *     cdef list __suffix_list
 */
struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse {
  struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse __pyx_base;
  PyObject *__pyx___suffix_list;
  PyObject *__pyx___form_list;
  PyObject *__pyx___transition_list;
  PyObject *__pyx___with_list;
  PyObject *__pyx___initial_pos;
  PyObject *__pyx___pos;
  PyObject *__pyx___form;
  PyObject *__pyx___verb_agreement;
  PyObject *__pyx___possesive_agreement;
};


/* "MorphologicalAnalysis/FiniteStateMachine.pxd":4
 * 
 * 
 * cdef class FiniteStateMachine:             # <<<<<<<<<<<<<<
 * 
 *     cdef list __states
 */
struct __pyx_obj_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine {
  PyObject_HEAD
  struct __pyx_vtabstruct_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine *__pyx_vtab;
  PyObject *__pyx___states;
  PyObject *__pyx___transitions;
};


/* "MorphologicalAnalysis/MetamorphicParse.pxd":6
 * 
 * 
 * cdef class MetamorphicParse:             # <<<<<<<<<<<<<<
 * 
 *     cdef list __meta_morpheme_list
 */
struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse {
  PyObject_HEAD
  struct __pyx_vtabstruct_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse *__pyx_vtab;
  PyObject *__pyx___meta_morpheme_list;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx___root;
};


/* "MorphologicalAnalysis/FsmMorphologicalAnalyzer.pxd":13
 * 
 * 
 * cdef class FsmMorphologicalAnalyzer:             # <<<<<<<<<<<<<<
 * 
 *     cdef Trie __dictionary_trie
 */
struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer {
  PyObject_HEAD
  struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *__pyx_vtab;
  struct __pyx_obj_10Dictionary_4Trie_4Trie_Trie *__pyx___dictionary_trie;
  struct __pyx_obj_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine *__pyx___finite_state_machine;
  struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *__pyx___dictionary;
  struct __pyx_obj_13DataStructure_5Cache_8LRUCache_LRUCache *__pyx___cache;
  PyObject *__pyx___most_used_patterns;
  PyObject *__pyx___parsed_surface_forms;
};


/* "SpellChecker/SpellChecker.pxd":4
 * 
 * 
 * cdef class SpellChecker:             # <<<<<<<<<<<<<<
 * 
 *     cpdef Sentence spellCheck(self, Sentence sentence)
 */
struct __pyx_obj_12SpellChecker_12SpellChecker_SpellChecker {
  PyObject_HEAD
  struct __pyx_vtabstruct_12SpellChecker_12SpellChecker_SpellChecker *__pyx_vtab;
};


/* "Language/Language.pxd":1
 * cdef class Language:             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_8Language_8Language_Language {
  PyObject_HEAD
};


/* "Language/TurkishLanguage.pxd":4
 * 
 * 
 * cdef class TurkishLanguage(Language):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_8Language_15TurkishLanguage_TurkishLanguage {
  struct __pyx_obj_8Language_8Language_Language __pyx_base;
};


/* "MorphologicalAnalysis/FsmParseList.pxd":4
 * from MorphologicalAnalysis.FsmParse cimport FsmParse
 * 
 * cdef class FsmParseList:             # <<<<<<<<<<<<<<
 * 
 *     cdef list __fsm_parses
 */
struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList {
  PyObject_HEAD
  struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList *__pyx_vtab;
  PyObject *__pyx___fsm_parses;
};


/* "SpellChecker/Candidate.pxd":3
 * from Dictionary.Word cimport Word
 * 
 * cdef class Candidate(Word):             # <<<<<<<<<<<<<<
 * 
 *     cdef object __operator
 */
struct __pyx_obj_12SpellChecker_9Candidate_Candidate {
  struct __pyx_obj_10Dictionary_4Word_Word __pyx_base;
  PyObject *__pyx___operator;
};


/* "SpellChecker/SimpleSpellChecker.pxd":7
 * 
 * 
 * cdef class SimpleSpellChecker(SpellChecker):             # <<<<<<<<<<<<<<
 * 
 *     cdef FsmMorphologicalAnalyzer fsm
 */
struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker {
  struct __pyx_obj_12SpellChecker_12SpellChecker_SpellChecker __pyx_base;
  struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *fsm;
  PyObject *__pyx___merged_words;
  PyObject *__pyx___split_words;
};



/* "Dictionary/Word.pxd":1
 * cdef class Word:             # <<<<<<<<<<<<<<
 * 
 *     cdef str name
 */

struct __pyx_vtabstruct_10Dictionary_4Word_Word {
  int (*charCount)(struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  PyObject *(*getName)(struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  PyObject *(*setName)(struct __pyx_obj_10Dictionary_4Word_Word *, PyObject *, int __pyx_skip_dispatch);
  int (*isPunctuation)(struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  PyObject *(*toCharacters)(struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10Dictionary_4Word_Word *__pyx_vtabptr_10Dictionary_4Word_Word;


/* "Corpus/Sentence.pxd":4
 * 
 * 
 * cdef class Sentence:             # <<<<<<<<<<<<<<
 * 
 *     cdef list words
 */

struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence {
  struct __pyx_obj_10Dictionary_4Word_Word *(*getWord)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, int, int __pyx_skip_dispatch);
  PyObject *(*getWords)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  PyObject *(*getStrings)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  int (*getIndex)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  int (*wordCount)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  PyObject *(*addWord)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  int (*charCount)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  PyObject *(*insertWord)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, int, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  PyObject *(*replaceWord)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, int, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  int (*safeIndex)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, int, int __pyx_skip_dispatch);
  PyObject *(*toString)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  PyObject *(*writeToFile)(struct __pyx_obj_6Corpus_8Sentence_Sentence *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *__pyx_vtabptr_6Corpus_8Sentence_Sentence;


/* "Dictionary/Trie/TrieNode.pxd":4
 * 
 * 
 * cdef class TrieNode:             # <<<<<<<<<<<<<<
 * 
 *     cdef dict __children
 */

struct __pyx_vtabstruct_10Dictionary_4Trie_8TrieNode_TrieNode {
  PyObject *(*addWord)(struct __pyx_obj_10Dictionary_4Trie_8TrieNode_TrieNode *, PyObject *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch, struct __pyx_opt_args_10Dictionary_4Trie_8TrieNode_8TrieNode_addWord *__pyx_optional_args);
  struct __pyx_obj_10Dictionary_4Trie_8TrieNode_TrieNode *(*getChild)(struct __pyx_obj_10Dictionary_4Trie_8TrieNode_TrieNode *, char, int __pyx_skip_dispatch);
  PyObject *(*getWords)(struct __pyx_obj_10Dictionary_4Trie_8TrieNode_TrieNode *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10Dictionary_4Trie_8TrieNode_TrieNode *__pyx_vtabptr_10Dictionary_4Trie_8TrieNode_TrieNode;


/* "Dictionary/TxtWord.pxd":4
 * 
 * 
 * cdef class TxtWord(Word):             # <<<<<<<<<<<<<<
 * 
 *     cdef list __flags
 */

struct __pyx_vtabstruct_10Dictionary_7TxtWord_TxtWord {
  struct __pyx_vtabstruct_10Dictionary_4Word_Word __pyx_base;
  PyObject *(*addFlag)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*removeFlag)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*getMorphology)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  PyObject *(*setMorphology)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*verbType)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*samePos)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isNominal)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isPassive)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isAbbreviation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isInterjection)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isDuplicate)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isCode)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isMetric)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isHeader)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isAdjective)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isPureAdjective)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isPronoun)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isQuestion)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isVerb)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isPortmanteau)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isDeterminer)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isConjunction)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isAdverb)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isPostP)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isPortmanteauEndingWithSI)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isPortmanteauFacedVowelEllipsis)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isPortmanteauFacedSoftening)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isSuffix)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isProperNoun)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isPlural)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isNumeral)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isReal)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isFraction)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isTime)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isDate)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isPercent)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isRange)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isOrdinal)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*notObeysVowelHarmonyDuringAgglutination)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*obeysAndNotObeysVowelHarmonyDuringAgglutination)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*rootSoftenDuringSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*rootSoftenAndNotSoftenDuringSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*verbSoftenDuringSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*nounSoftenDuringSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*endingKChangesIntoG)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*isExceptional)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*duplicatesDuringSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*duplicatesAndNotDuplicatesDuringSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*lastIdropsDuringSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*lastIDropsAndNotDropDuringSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*takesRelativeSuffixKi)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*takesRelativeSuffixKu)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*consonantSMayInsertedDuringPossesiveSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*lastIdropsDuringPassiveSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*vowelAChangesToIDuringYSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*vowelEChangesToIDuringYSuffixation)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*takesSuffixIRAsAorist)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*takesSuffixDIRAsFactitive)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*showsSuRegularities)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*containsFlag)(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10Dictionary_7TxtWord_TxtWord *__pyx_vtabptr_10Dictionary_7TxtWord_TxtWord;


/* "Dictionary/Trie/Trie.pxd":6
 * 
 * 
 * cdef class Trie:             # <<<<<<<<<<<<<<
 * 
 *     cdef TrieNode __root_node
 */

struct __pyx_vtabstruct_10Dictionary_4Trie_4Trie_Trie {
  PyObject *(*addWord)(struct __pyx_obj_10Dictionary_4Trie_4Trie_Trie *, PyObject *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  PyObject *(*getWordsWithPrefix)(struct __pyx_obj_10Dictionary_4Trie_4Trie_Trie *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *(*getCompundWordStartingWith)(struct __pyx_obj_10Dictionary_4Trie_4Trie_Trie *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10Dictionary_4Trie_4Trie_Trie *__pyx_vtabptr_10Dictionary_4Trie_4Trie_Trie;


/* "Dictionary/Dictionary.pxd":4
 * 
 * 
 * cdef class Dictionary:             # <<<<<<<<<<<<<<
 * 
 *     cdef list words
 */

struct __pyx_vtabstruct_10Dictionary_10Dictionary_Dictionary {
  struct __pyx_obj_10Dictionary_4Word_Word *(*getWord)(struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *, PyObject *, int __pyx_skip_dispatch);
  int (*getWordIndex)(struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*removeWord)(struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *, PyObject *, int __pyx_skip_dispatch);
  int (*size)(struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *, int __pyx_skip_dispatch);
  struct __pyx_obj_10Dictionary_4Word_Word *(*getWordWithIndex)(struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *, int, int __pyx_skip_dispatch);
  int (*longestWordSize)(struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *, int __pyx_skip_dispatch);
  int (*__pyx___getPosition)(struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  int (*getWordStartingWith)(struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10Dictionary_10Dictionary_Dictionary *__pyx_vtabptr_10Dictionary_10Dictionary_Dictionary;


/* "Dictionary/TxtDictionary.pxd":7
 * 
 * 
 * cdef class TxtDictionary(Dictionary):             # <<<<<<<<<<<<<<
 * 
 *     cdef dict __misspelled_words
 */

struct __pyx_vtabstruct_10Dictionary_13TxtDictionary_TxtDictionary {
  struct __pyx_vtabstruct_10Dictionary_10Dictionary_Dictionary __pyx_base;
  PyObject *(*addNumber)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*addRealNumber)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*addFraction)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*addTime)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  int (*addProperNoun)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  int (*addNoun)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  int (*addVerb)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  int (*addAdjective)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  int (*addAdverb)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  int (*addPronoun)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  int (*addWithFlag)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*mergeDictionary)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*__pyx___loadFromText)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*__pyx___loadMisspelledWords)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*__pyx___loadMorphologicalLexicon)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*getCorrectForm)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*saveAsTxt)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*__pyx___addWordWhenRootSoften)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, struct __pyx_obj_10Dictionary_4Trie_4Trie_Trie *, Py_UNICODE, PyObject *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  struct __pyx_obj_10Dictionary_4Trie_4Trie_Trie *(*prepareTrie)(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10Dictionary_13TxtDictionary_TxtDictionary *__pyx_vtabptr_10Dictionary_13TxtDictionary_TxtDictionary;


/* "DataStructure/Cache/LRUCache.pxd":1
 * cdef class LRUCache(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef int __cacheSize
 */

struct __pyx_vtabstruct_13DataStructure_5Cache_8LRUCache_LRUCache {
  int (*contains)(struct __pyx_obj_13DataStructure_5Cache_8LRUCache_LRUCache *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get)(struct __pyx_obj_13DataStructure_5Cache_8LRUCache_LRUCache *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*add)(struct __pyx_obj_13DataStructure_5Cache_8LRUCache_LRUCache *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_13DataStructure_5Cache_8LRUCache_LRUCache *__pyx_vtabptr_13DataStructure_5Cache_8LRUCache_LRUCache;


/* "MorphologicalAnalysis/InflectionalGroup.pxd":1
 * cdef class InflectionalGroup:             # <<<<<<<<<<<<<<
 * 
 *     cdef list __IG
 */

struct __pyx_vtabstruct_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup {
  PyObject *(*getTag)(struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *, int, int __pyx_skip_dispatch);
  int (*size)(struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *, int __pyx_skip_dispatch);
  PyObject *(*containsCase)(struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *, int __pyx_skip_dispatch);
  int (*containsPlural)(struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *, int __pyx_skip_dispatch);
  int (*containsTag)(struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *, PyObject *, int __pyx_skip_dispatch);
  int (*containsPossessive)(struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *__pyx_vtabptr_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup;


/* "MorphologicalAnalysis/MorphologicalParse.pxd":6
 * 
 * 
 * cdef class MorphologicalParse:             # <<<<<<<<<<<<<<
 * 
 *     cdef list inflectional_groups
 */

struct __pyx_vtabstruct_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse {
  struct __pyx_obj_10Dictionary_4Word_Word *(*getWord)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getTransitionList)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getInflectionalGroupString)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int, int __pyx_skip_dispatch);
  struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *(*getInflectionalGroup)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int, int __pyx_skip_dispatch);
  struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *(*getLastInflectionalGroup)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getTag)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int, int __pyx_skip_dispatch);
  int (*tagSize)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*size)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *(*firstInflectionalGroup)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup *(*lastInflectionalGroup)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  struct __pyx_obj_10Dictionary_4Word_Word *(*getWordWithPos)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getPos)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getRootPos)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*lastIGContainsCase)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*lastIGContainsTag)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, PyObject *, int __pyx_skip_dispatch);
  int (*lastIGContainsPossessive)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isCapitalWord)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isNoun)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isVerb)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isRootVerb)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isAdjective)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isProperNoun)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isPunctuation)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isCardinal)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isOrdinal)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isReal)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isNumber)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isTime)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isDate)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isHashTag)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isEmail)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isPercent)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isFraction)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isRange)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isPlural)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*isAuxiliary)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  int (*containsTag)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*getTreePos)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getPronType)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getNumType)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getReflex)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getNumber)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getCase)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getDefinite)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getDegree)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getPolarity)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getPerson)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getVoice)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getAspect)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getTense)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getMood)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getVerbForm)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
  PyObject *(*getUniversalDependencyFeatures)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*getUniversalDependencyPos)(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *__pyx_vtabptr_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse;


/* "MorphologicalAnalysis/State.pxd":1
 * cdef class State:             # <<<<<<<<<<<<<<
 * 
 *     cdef bint __start_state, __end_state
 */

struct __pyx_vtabstruct_21MorphologicalAnalysis_5State_State {
  PyObject *(*getName)(struct __pyx_obj_21MorphologicalAnalysis_5State_State *, int __pyx_skip_dispatch);
  PyObject *(*getPos)(struct __pyx_obj_21MorphologicalAnalysis_5State_State *, int __pyx_skip_dispatch);
  int (*isEndState)(struct __pyx_obj_21MorphologicalAnalysis_5State_State *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_21MorphologicalAnalysis_5State_State *__pyx_vtabptr_21MorphologicalAnalysis_5State_State;


/* "MorphologicalAnalysis/FsmParse.pxd":5
 * 
 * 
 * cdef class FsmParse(MorphologicalParse):             # <<<<<<<<<<<<<<
 * 
 *     cdef list __suffix_list
 */

struct __pyx_vtabstruct_21MorphologicalAnalysis_8FsmParse_FsmParse {
  struct __pyx_vtabstruct_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse __pyx_base;
  PyObject *(*constructInflectionalGroups)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*getVerbAgreement)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*getPossesiveAgreement)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*setAgreement)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*getLastLemmaWithTag)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*getLastLemma)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*addSuffix)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, struct __pyx_obj_21MorphologicalAnalysis_5State_State *, PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*getSurfaceForm)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  struct __pyx_obj_21MorphologicalAnalysis_5State_State *(*getStartState)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*getFinalPos)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*getInitialPos)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*setForm)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_21MorphologicalAnalysis_5State_State *(*getFinalSuffix)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*headerTransition)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*pronounTransition)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*transitionList)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*suffixList)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*withList)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*getWithList)(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_21MorphologicalAnalysis_8FsmParse_FsmParse *__pyx_vtabptr_21MorphologicalAnalysis_8FsmParse_FsmParse;


/* "MorphologicalAnalysis/FiniteStateMachine.pxd":4
 * 
 * 
 * cdef class FiniteStateMachine:             # <<<<<<<<<<<<<<
 * 
 *     cdef list __states
 */

struct __pyx_vtabstruct_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine {
  int (*isValidTransition)(struct __pyx_obj_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*getStates)(struct __pyx_obj_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine *, int __pyx_skip_dispatch);
  struct __pyx_obj_21MorphologicalAnalysis_5State_State *(*getState)(struct __pyx_obj_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*addTransition)(struct __pyx_obj_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine *, struct __pyx_obj_21MorphologicalAnalysis_5State_State *, struct __pyx_obj_21MorphologicalAnalysis_5State_State *, PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_21MorphologicalAnalysis_18FiniteStateMachine_18FiniteStateMachine_addTransition *__pyx_optional_args);
  PyObject *(*getTransitions)(struct __pyx_obj_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine *, struct __pyx_obj_21MorphologicalAnalysis_5State_State *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine *__pyx_vtabptr_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine;


/* "MorphologicalAnalysis/MetamorphicParse.pxd":6
 * 
 * 
 * cdef class MetamorphicParse:             # <<<<<<<<<<<<<<
 * 
 *     cdef list __meta_morpheme_list
 */

struct __pyx_vtabstruct_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse {
  PyObject *(*getMetaMorphemeTag)(struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_10Dictionary_4Word_Word *(*getWord)(struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse *, int __pyx_skip_dispatch);
  PyObject *(*getMetaMorphemeTagForParse)(struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse *, struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, PyObject *, int __pyx_skip_dispatch);
  int (*size)(struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse *, int __pyx_skip_dispatch);
  PyObject *(*addMetaMorphemeList)(struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*removeMetaMorphemeFromIndex)(struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse *, int, int __pyx_skip_dispatch);
  PyObject *(*getMetaMorpheme)(struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse *, int, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse *__pyx_vtabptr_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse;


/* "MorphologicalAnalysis/FsmMorphologicalAnalyzer.pxd":13
 * 
 * 
 * cdef class FsmMorphologicalAnalyzer:             # <<<<<<<<<<<<<<
 * 
 *     cdef Trie __dictionary_trie
 */

struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer {
  PyObject *(*addParsedSurfaceForms)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*getPossibleWords)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *, struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse *, int __pyx_skip_dispatch);
  struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *(*getDictionary)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, int __pyx_skip_dispatch);
  struct __pyx_obj_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine *(*getFiniteStateMachine)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, int __pyx_skip_dispatch);
  int (*__pyx___isPossibleSubstring)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, PyObject *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  PyObject *(*__pyx___initializeParseList)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int, int __pyx_skip_dispatch);
  PyObject *(*__pyx___initializeParseListFromRoot)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int, int __pyx_skip_dispatch);
  PyObject *(*__pyx___initializeParseListFromSurfaceForm)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int, int __pyx_skip_dispatch);
  PyObject *(*__pyx___addNewParsesFromCurrentParse)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, PyObject *, PyObject *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  int (*__pyx___parseExists)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*__pyx___parseWord)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*morphologicalAnalysisRoot)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch, struct __pyx_opt_args_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_24FsmMorphologicalAnalyzer_morphologicalAnalysisRoot *__pyx_optional_args);
  PyObject *(*generateAllParses)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int, int __pyx_skip_dispatch);
  PyObject *(*replaceRootWord)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, int __pyx_skip_dispatch);
  struct __pyx_obj_6Corpus_8Sentence_Sentence *(*replaceWord)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  int (*__pyx___analysisExists)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, PyObject *, int, int __pyx_skip_dispatch);
  PyObject *(*__pyx___analysis)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int, int __pyx_skip_dispatch);
  int (*patternMatches)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  int (*isProperNoun)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*morphologicalAnalysis)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*robustMorphologicalAnalysis)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  int (*__pyx___isInteger)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  int (*__pyx___isDouble)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  int (*__pyx___isNumber)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  int (*__pyx___isPercent)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  int (*__pyx___isTime)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  int (*__pyx___isRange)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  int (*__pyx___isDate)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  int (*__pyx___isCode)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*__pyx___toLower)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, PyObject *, int __pyx_skip_dispatch);
  int (*morphologicalAnalysisExists)(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *, struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *__pyx_vtabptr_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer;


/* "SpellChecker/SpellChecker.pxd":4
 * 
 * 
 * cdef class SpellChecker:             # <<<<<<<<<<<<<<
 * 
 *     cpdef Sentence spellCheck(self, Sentence sentence)
 */

struct __pyx_vtabstruct_12SpellChecker_12SpellChecker_SpellChecker {
  struct __pyx_obj_6Corpus_8Sentence_Sentence *(*spellCheck)(struct __pyx_obj_12SpellChecker_12SpellChecker_SpellChecker *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12SpellChecker_12SpellChecker_SpellChecker *__pyx_vtabptr_12SpellChecker_12SpellChecker_SpellChecker;


/* "MorphologicalAnalysis/FsmParseList.pxd":4
 * from MorphologicalAnalysis.FsmParse cimport FsmParse
 * 
 * cdef class FsmParseList:             # <<<<<<<<<<<<<<
 * 
 *     cdef list __fsm_parses
 */

struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList {
  int (*size)(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *, int __pyx_skip_dispatch);
  PyObject *(*addFsmParse)(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *, struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *(*getFsmParse)(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *, int, int __pyx_skip_dispatch);
  PyObject *(*rootWords)(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *, int __pyx_skip_dispatch);
  PyObject *(*reduceToParsesWithSameRootAndPos)(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *(*getParseWithLongestRootWord)(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *, int __pyx_skip_dispatch);
  PyObject *(*reduceToParsesWithSameRoot)(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *, PyObject *, int __pyx_skip_dispatch);
  int (*isLongestRootException)(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *, struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *, int __pyx_skip_dispatch);
  PyObject *(*constructParseListForDifferentRootWithPos)(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *, int __pyx_skip_dispatch);
  PyObject *(*parsesWithoutPrefixAndSuffix)(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList *__pyx_vtabptr_21MorphologicalAnalysis_12FsmParseList_FsmParseList;


/* "SpellChecker/Candidate.pxd":3
 * from Dictionary.Word cimport Word
 * 
 * cdef class Candidate(Word):             # <<<<<<<<<<<<<<
 * 
 *     cdef object __operator
 */

struct __pyx_vtabstruct_12SpellChecker_9Candidate_Candidate {
  struct __pyx_vtabstruct_10Dictionary_4Word_Word __pyx_base;
  PyObject *(*getOperator)(struct __pyx_obj_12SpellChecker_9Candidate_Candidate *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12SpellChecker_9Candidate_Candidate *__pyx_vtabptr_12SpellChecker_9Candidate_Candidate;


/* "SpellChecker/SimpleSpellChecker.pyx":12
 * from SpellChecker.Operator import Operator
 * 
 * cdef class SimpleSpellChecker(SpellChecker):             # <<<<<<<<<<<<<<
 * 
 *     __shortcuts: list = ["cc", "cm2", "cm", "gb", "ghz", "gr", "gram", "hz", "inc", "inch", "in", "kg", "kw", "kva",
 */

struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker {
  struct __pyx_vtabstruct_12SpellChecker_12SpellChecker_SpellChecker __pyx_base;
  PyObject *(*__pyx___generateCandidateList)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*candidateList)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  int (*forcedMisspellCheck)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  int (*forcedBackwardMergeCheck)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  int (*forcedForwardMergeCheck)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  int (*forcedSplitCheck)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  int (*forcedShortcutCheck)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  int (*forcedDeDaSplitCheck)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  int (*forcedSuffixMergeCheck)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  int (*forcedHyphenMergeCheck)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  int (*forcedQuestionSuffixSplitCheck)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch);
  PyObject *(*loadDictionaries)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, int __pyx_skip_dispatch);
  PyObject *(*getCorrectForm)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*mergedCandidatesList)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  PyObject *(*splitCandidatesList)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
  PyObject *(*getSplitPair)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_vtabptr_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyUnicode_Substring.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Substring(
            PyObject* text, Py_ssize_t start, Py_ssize_t stop);

/* GetItemIntUnicode.proto */
#define __Pyx_GetItemInt_Unicode(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Unicode_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "string index out of range"), (Py_UCS4)-1))
static CYTHON_INLINE Py_UCS4 __Pyx_GetItemInt_Unicode_Fast(PyObject* ustring, Py_ssize_t i,
                                                           int wraparound, int boundscheck);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* pop_index.proto */
static PyObject* __Pyx__PyObject_PopNewIndex(PyObject* L, PyObject* py_ix);
static PyObject* __Pyx__PyObject_PopIndex(PyObject* L, PyObject* py_ix);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static PyObject* __Pyx__PyList_PopIndex(PyObject* L, PyObject* py_ix, Py_ssize_t ix);
#define __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    (likely(PyList_CheckExact(L) && __Pyx_fits_Py_ssize_t(ix, type, is_signed))) ?\
        __Pyx__PyList_PopIndex(L, py_ix, ix) : (\
        (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
            __Pyx__PyObject_PopIndex(L, py_ix)))
#define __Pyx_PyList_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    __Pyx_fits_Py_ssize_t(ix, type, is_signed) ?\
        __Pyx__PyList_PopIndex(L, py_ix, ix) : (\
        (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
            __Pyx__PyObject_PopIndex(L, py_ix)))
#else
#define __Pyx_PyList_PopIndex(L, py_ix, ix, is_signed, type, to_py_func)\
    __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func)
#define __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
        __Pyx__PyObject_PopIndex(L, py_ix))
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* unicode_tailmatch.proto */
static int __Pyx_PyUnicode_Tailmatch(
    PyObject* s, PyObject* substr, Py_ssize_t start, Py_ssize_t end, int direction);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* PyUCS4InUnicode.proto */
static CYTHON_INLINE int __Pyx_UnicodeContainsUCS4(PyObject* unicode, Py_UCS4 character);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___generateCandidateList(CYTHON_UNUSED struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v_word, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_candidateList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, CYTHON_UNUSED struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_sentence, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_spellCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_sentence, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedMisspellCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedBackwardMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedForwardMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedShortcutCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedDeDaSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedSuffixMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedHyphenMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedQuestionSuffixSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_loadDictionaries(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_getCorrectForm(CYTHON_UNUSED struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v_wordName, PyObject *__pyx_v_dictionary, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_mergedCandidatesList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_splitCandidatesList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_getSplitPair(CYTHON_UNUSED struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'Dictionary.Word' */
static PyTypeObject *__pyx_ptype_10Dictionary_4Word_Word = 0;

/* Module declarations from 'Corpus.Sentence' */
static PyTypeObject *__pyx_ptype_6Corpus_8Sentence_Sentence = 0;

/* Module declarations from 'Dictionary.Trie.TrieNode' */
static PyTypeObject *__pyx_ptype_10Dictionary_4Trie_8TrieNode_TrieNode = 0;

/* Module declarations from 'Dictionary.TxtWord' */
static PyTypeObject *__pyx_ptype_10Dictionary_7TxtWord_TxtWord = 0;

/* Module declarations from 'Dictionary.Trie.Trie' */
static PyTypeObject *__pyx_ptype_10Dictionary_4Trie_4Trie_Trie = 0;

/* Module declarations from 'Dictionary.Dictionary' */
static PyTypeObject *__pyx_ptype_10Dictionary_10Dictionary_Dictionary = 0;

/* Module declarations from 'Dictionary.TxtDictionary' */
static PyTypeObject *__pyx_ptype_10Dictionary_13TxtDictionary_TxtDictionary = 0;

/* Module declarations from 'DataStructure.Cache.LRUCache' */
static PyTypeObject *__pyx_ptype_13DataStructure_5Cache_8LRUCache_LRUCache = 0;

/* Module declarations from 'MorphologicalAnalysis.InflectionalGroup' */
static PyTypeObject *__pyx_ptype_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup = 0;

/* Module declarations from 'MorphologicalAnalysis.MorphologicalParse' */
static PyTypeObject *__pyx_ptype_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse = 0;

/* Module declarations from 'MorphologicalAnalysis.State' */
static PyTypeObject *__pyx_ptype_21MorphologicalAnalysis_5State_State = 0;

/* Module declarations from 'MorphologicalAnalysis.FsmParse' */
static PyTypeObject *__pyx_ptype_21MorphologicalAnalysis_8FsmParse_FsmParse = 0;

/* Module declarations from 'MorphologicalAnalysis.FiniteStateMachine' */
static PyTypeObject *__pyx_ptype_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine = 0;

/* Module declarations from 'MorphologicalAnalysis.MetamorphicParse' */
static PyTypeObject *__pyx_ptype_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse = 0;

/* Module declarations from 'MorphologicalAnalysis.FsmMorphologicalAnalyzer' */
static PyTypeObject *__pyx_ptype_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer = 0;

/* Module declarations from 'SpellChecker.SpellChecker' */
static PyTypeObject *__pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker = 0;

/* Module declarations from 'Language.Language' */
static PyTypeObject *__pyx_ptype_8Language_8Language_Language = 0;

/* Module declarations from 'Language.TurkishLanguage' */
static PyTypeObject *__pyx_ptype_8Language_15TurkishLanguage_TurkishLanguage = 0;

/* Module declarations from 'MorphologicalAnalysis.FsmParseList' */
static PyTypeObject *__pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList = 0;

/* Module declarations from 'SpellChecker.Candidate' */
static PyTypeObject *__pyx_ptype_12SpellChecker_9Candidate_Candidate = 0;

/* Module declarations from 'SpellChecker.SimpleSpellChecker' */
static PyTypeObject *__pyx_ptype_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker = 0;
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker___pyx_unpickle_SimpleSpellChecker__set_state(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "SpellChecker.SimpleSpellChecker"
extern int __pyx_module_is_main_SpellChecker__SimpleSpellChecker;
int __pyx_module_is_main_SpellChecker__SimpleSpellChecker = 0;

/* Implementation of 'SpellChecker.SimpleSpellChecker' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_open;
static const char __pyx_k_[] = "";
static const char __pyx_k_g[] = "g";
static const char __pyx_k_l[] = "l\304\261";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_w[] = "w";
static const char __pyx_k__2[] = " ";
static const char __pyx_k__3[] = "|";
static const char __pyx_k__4[] = ")";
static const char __pyx_k__5[] = "'";
static const char __pyx_k__8[] = "-";
static const char __pyx_k__9[] = "\342\200\223";
static const char __pyx_k_ah[] = "ah";
static const char __pyx_k_cc[] = "cc";
static const char __pyx_k_cm[] = "cm";
static const char __pyx_k_da[] = "da";
static const char __pyx_k_db[] = "db";
static const char __pyx_k_de[] = "de";
static const char __pyx_k_ft[] = "ft";
static const char __pyx_k_gb[] = "gb";
static const char __pyx_k_gr[] = "gr";
static const char __pyx_k_hp[] = "hp";
static const char __pyx_k_hz[] = "hz";
static const char __pyx_k_in[] = "in\303\247";
static const char __pyx_k_kg[] = "kg";
static const char __pyx_k_kw[] = "kw";
static const char __pyx_k_li[] = "li";
static const char __pyx_k_lk[] = "l\304\261k";
static const char __pyx_k_lt[] = "lt";
static const char __pyx_k_lu[] = "lu";
static const char __pyx_k_m2[] = "m2";
static const char __pyx_k_m3[] = "m3";
static const char __pyx_k_mb[] = "mb";
static const char __pyx_k_mg[] = "mg";
static const char __pyx_k_mi[] = "mi";
static const char __pyx_k_ml[] = "ml";
static const char __pyx_k_mm[] = "mm";
static const char __pyx_k_mp[] = "mp";
static const char __pyx_k_ms[] = "ms";
static const char __pyx_k_mt[] = "mt";
static const char __pyx_k_mu[] = "mu";
static const char __pyx_k_mv[] = "mv";
static const char __pyx_k_oz[] = "oz";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_sn[] = "sn";
static const char __pyx_k_tb[] = "tb";
static const char __pyx_k_tl[] = "tl";
static const char __pyx_k_va[] = "va";
static const char __pyx_k_0_9[] = "[0-9]+";
static const char __pyx_k__10[] = "\342\200\224";
static const char __pyx_k_bit[] = "bit";
static const char __pyx_k_cm2[] = "cm2";
static const char __pyx_k_cm3[] = "cm3";
static const char __pyx_k_dpi[] = "dpi";
static const char __pyx_k_fsm[] = "fsm";
static const char __pyx_k_ghz[] = "ghz";
static const char __pyx_k_inc[] = "inc";
static const char __pyx_k_khz[] = "khz";
static const char __pyx_k_kva[] = "kva";
static const char __pyx_k_kwh[] = "kwh";
static const char __pyx_k_l_2[] = "l\303\274";
static const char __pyx_k_l_3[] = "l";
static const char __pyx_k_lik[] = "lik";
static const char __pyx_k_luk[] = "luk";
static const char __pyx_k_m_2[] = "m\304\261";
static const char __pyx_k_m_3[] = "m\303\274";
static const char __pyx_k_mah[] = "mah";
static const char __pyx_k_mdr[] = "m\304\261d\304\261r";
static const char __pyx_k_mhz[] = "mhz";
static const char __pyx_k_mm2[] = "mm2";
static const char __pyx_k_mm3[] = "mm3";
static const char __pyx_k_msn[] = "m\304\261s\304\261n";
static const char __pyx_k_myd[] = "m\304\261yd\304\261";
static const char __pyx_k_mym[] = "m\304\261y\304\261m";
static const char __pyx_k_myz[] = "m\304\261y\304\261z";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_ohm[] = "ohm";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_ppm[] = "ppm";
static const char __pyx_k_rpm[] = "rpm";
static const char __pyx_k_byte[] = "byte";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_gbit[] = "gbit";
static const char __pyx_k_gbps[] = "gbps";
static const char __pyx_k_gram[] = "gram";
static const char __pyx_k_inch[] = "inch";
static const char __pyx_k_kbit[] = "kbit";
static const char __pyx_k_kcal[] = "kcal";
static const char __pyx_k_lk_2[] = "l\303\274k";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mbit[] = "mbit";
static const char __pyx_k_mbps[] = "mbps";
static const char __pyx_k_msnz[] = "m\304\261s\304\261n\304\261z";
static const char __pyx_k_mydk[] = "m\304\261yd\304\261k";
static const char __pyx_k_mydm[] = "m\304\261yd\304\261m";
static const char __pyx_k_mydn[] = "m\304\261yd\304\261n";
static const char __pyx_k_mymm[] = "m\304\261ym\304\261\305\237\304\261m";
static const char __pyx_k_mymz[] = "m\304\261ym\304\261\305\237\304\261z";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_utf8[] = "utf8";
static const char __pyx_k_volt[] = "volt";
static const char __pyx_k_watt[] = "watt";
static const char __pyx_k_word[] = "word";
static const char __pyx_k_SPLIT[] = "SPLIT";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_litre[] = "litre";
static const char __pyx_k_mdr_2[] = "m\303\274d\303\274r";
static const char __pyx_k_metre[] = "metre";
static const char __pyx_k_midir[] = "midir";
static const char __pyx_k_misin[] = "misin";
static const char __pyx_k_miydi[] = "miydi";
static const char __pyx_k_miyim[] = "miyim";
static const char __pyx_k_miyiz[] = "miyiz";
static const char __pyx_k_miymi[] = "miymi\305\237";
static const char __pyx_k_msn_2[] = "m\303\274s\303\274n";
static const char __pyx_k_msndr[] = "m\304\261s\304\261nd\304\261r";
static const char __pyx_k_mudur[] = "mudur";
static const char __pyx_k_musun[] = "musun";
static const char __pyx_k_muydu[] = "muydu";
static const char __pyx_k_muymu[] = "muymu\305\237";
static const char __pyx_k_muyum[] = "muyum";
static const char __pyx_k_muyuz[] = "muyuz";
static const char __pyx_k_myd_2[] = "m\303\274yd\303\274";
static const char __pyx_k_mydnz[] = "m\304\261yd\304\261n\304\261z";
static const char __pyx_k_mym_2[] = "m\303\274y\303\274m";
static const char __pyx_k_mym_3[] = "m\304\261ym\304\261\305\237";
static const char __pyx_k_mym_4[] = "m\303\274ym\303\274\305\237";
static const char __pyx_k_mymdr[] = "m\304\261y\304\261md\304\261r";
static const char __pyx_k_mymsn[] = "m\304\261ym\304\261\305\237s\304\261n";
static const char __pyx_k_myz_2[] = "m\303\274y\303\274z";
static const char __pyx_k_myzdr[] = "m\304\261y\304\261zd\304\261r";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_a_zA_Z[] = "[a-zA-Z\303\247\303\266\304\237\303\274\305\237\304\261\303\207\303\226\304\236\303\234\305\236\304\260]+";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_miydik[] = "miydik";
static const char __pyx_k_miydim[] = "miydim";
static const char __pyx_k_miydin[] = "miydin";
static const char __pyx_k_msnz_2[] = "m\303\274s\303\274n\303\274z";
static const char __pyx_k_msnzdr[] = "m\304\261s\304\261n\304\261zd\304\261r";
static const char __pyx_k_muyduk[] = "muyduk";
static const char __pyx_k_muydum[] = "muydum";
static const char __pyx_k_muydun[] = "muydun";
static const char __pyx_k_mydk_2[] = "m\303\274yd\303\274k";
static const char __pyx_k_mydlar[] = "m\304\261yd\304\261lar";
static const char __pyx_k_mydler[] = "m\303\274yd\303\274ler";
static const char __pyx_k_mydm_2[] = "m\303\274yd\303\274m";
static const char __pyx_k_mydn_2[] = "m\303\274yd\303\274n";
static const char __pyx_k_mymlar[] = "m\304\261ym\304\261\305\237lar";
static const char __pyx_k_mymler[] = "m\303\274ym\303\274\305\237ler";
static const char __pyx_k_mymm_2[] = "m\303\274ym\303\274\305\237\303\274m";
static const char __pyx_k_mymz_2[] = "m\303\274ym\303\274\305\237\303\274z";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_rindex[] = "rindex";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_LETTERS[] = "LETTERS";
static const char __pyx_k_compile[] = "compile";
static const char __pyx_k_getName[] = "getName";
static const char __pyx_k_misiniz[] = "misiniz";
static const char __pyx_k_miymiim[] = "miymi\305\237im";
static const char __pyx_k_miymiiz[] = "miymi\305\237iz";
static const char __pyx_k_msndr_2[] = "m\303\274s\303\274nd\303\274r";
static const char __pyx_k_musunuz[] = "musunuz";
static const char __pyx_k_muymuum[] = "muymu\305\237um";
static const char __pyx_k_muymuuz[] = "muymu\305\237uz";
static const char __pyx_k_mydnz_2[] = "m\303\274yd\303\274n\303\274z";
static const char __pyx_k_mymdr_2[] = "m\303\274y\303\274md\303\274r";
static const char __pyx_k_mymsn_2[] = "m\303\274ym\303\274\305\237s\303\274n";
static const char __pyx_k_myzdr_2[] = "m\303\274y\303\274zd\303\274r";
static const char __pyx_k_toLower[] = "toLower";
static const char __pyx_k_Operator[] = "Operator";
static const char __pyx_k_encoding[] = "encoding";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_misindir[] = "misindir";
static const char __pyx_k_miydiler[] = "miydiler";
static const char __pyx_k_miydiniz[] = "miydiniz";
static const char __pyx_k_miyimdir[] = "miyimdir";
static const char __pyx_k_miyizdir[] = "miyizdir";
static const char __pyx_k_miymiler[] = "miymi\305\237ler";
static const char __pyx_k_miymisin[] = "miymi\305\237sin";
static const char __pyx_k_msnzdr_2[] = "m\303\274s\303\274n\303\274zd\303\274r";
static const char __pyx_k_musundur[] = "musundur";
static const char __pyx_k_muydular[] = "muydular";
static const char __pyx_k_muydunuz[] = "muydunuz";
static const char __pyx_k_muymular[] = "muymu\305\237lar";
static const char __pyx_k_muymusun[] = "muymu\305\237sun";
static const char __pyx_k_muyumdur[] = "muyumdur";
static const char __pyx_k_muyuzdur[] = "muyuzdur";
static const char __pyx_k_nextWord[] = "nextWord";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_sentence[] = "sentence";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_wordName[] = "wordName";
static const char __pyx_k_fullmatch[] = "fullmatch";
static const char __pyx_k_lastVowel[] = "lastVowel";
static const char __pyx_k_multiWord[] = "multiWord";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_randrange[] = "randrange";
static const char __pyx_k_readlines[] = "readlines";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_shortcuts[] = "__shortcuts";
static const char __pyx_k_toCapital[] = "toCapital";
static const char __pyx_k_dictionary[] = "dictionary";
static const char __pyx_k_misinizdir[] = "misinizdir";
static const char __pyx_k_musunuzdur[] = "musunuzdur";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_spellCheck[] = "spellCheck";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_SPELL_CHECK[] = "SPELL_CHECK";
static const char __pyx_k_getOperator[] = "getOperator";
static const char __pyx_k_isBackVowel[] = "isBackVowel";
static const char __pyx_k_getSplitPair[] = "getSplitPair";
static const char __pyx_k_previousWord[] = "previousWord";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_1_9_0_9_0_0_9[] = "(([1-9][0-9]*)|[0])(([.]|[,])[0-9]*)?(";
static const char __pyx_k_FORWARD_MERGE[] = "FORWARD_MERGE";
static const char __pyx_k_addSplitWords[] = "addSplitWords";
static const char __pyx_k_candidateList[] = "candidateList";
static const char __pyx_k_pkg_resources[] = "pkg_resources";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_BACKWARD_MERGE[] = "BACKWARD_MERGE";
static const char __pyx_k_data_split_txt[] = "data/split.txt";
static const char __pyx_k_getCorrectForm[] = "getCorrectForm";
static const char __pyx_k_data_merged_txt[] = "data/merged.txt";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_forcedSplitCheck[] = "forcedSplitCheck";
static const char __pyx_k_loadDictionaries[] = "loadDictionaries";
static const char __pyx_k_1_9_0_9_0_2_0_0_9[] = "(([1-9][0-9]{0,2})|[0])(([.]|[,])[0-9]*)?(";
static const char __pyx_k_LOWERCASE_LETTERS[] = "LOWERCASE_LETTERS";
static const char __pyx_k_resource_filename[] = "resource_filename";
static const char __pyx_k_MISSPELLED_REPLACE[] = "MISSPELLED_REPLACE";
static const char __pyx_k_SimpleSpellChecker[] = "SimpleSpellChecker";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_questionSuffixList[] = "__questionSuffixList";
static const char __pyx_k_forcedMisspellCheck[] = "forcedMisspellCheck";
static const char __pyx_k_forcedShortcutCheck[] = "forcedShortcutCheck";
static const char __pyx_k_splitCandidatesList[] = "splitCandidatesList";
static const char __pyx_k_conditionalShortcuts[] = "__conditionalShortcuts";
static const char __pyx_k_forcedDeDaSplitCheck[] = "forcedDeDaSplitCheck";
static const char __pyx_k_mergedCandidatesList[] = "mergedCandidatesList";
static const char __pyx_k_SpellChecker_Operator[] = "SpellChecker.Operator";
static const char __pyx_k_generateCandidateList[] = "__generateCandidateList";
static const char __pyx_k_forcedHyphenMergeCheck[] = "forcedHyphenMergeCheck";
static const char __pyx_k_forcedSuffixMergeCheck[] = "forcedSuffixMergeCheck";
static const char __pyx_k_forcedForwardMergeCheck[] = "forcedForwardMergeCheck";
static const char __pyx_k_forcedBackwardMergeCheck[] = "forcedBackwardMergeCheck";
static const char __pyx_k_forcedQuestionSuffixSplitCheck[] = "forcedQuestionSuffixSplitCheck";
static const char __pyx_k_pyx_unpickle_SimpleSpellChecke[] = "__pyx_unpickle_SimpleSpellChecker";
static const char __pyx_k_SpellChecker_SimpleSpellChecker[] = "SpellChecker.SimpleSpellChecker";
static const char __pyx_k_Incompatible_checksums_s_vs_0x16[] = "Incompatible checksums (%s vs 0x162cce2 = (__merged_words, __split_words, fsm))";
static PyObject *__pyx_kp_u_;
static PyObject *__pyx_kp_u_0_9;
static PyObject *__pyx_kp_u_1_9_0_9_0_0_9;
static PyObject *__pyx_kp_u_1_9_0_9_0_2_0_0_9;
static PyObject *__pyx_n_s_BACKWARD_MERGE;
static PyObject *__pyx_n_s_FORWARD_MERGE;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x16;
static PyObject *__pyx_n_s_LETTERS;
static PyObject *__pyx_n_s_LOWERCASE_LETTERS;
static PyObject *__pyx_n_s_MISSPELLED_REPLACE;
static PyObject *__pyx_n_s_Operator;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_SPELL_CHECK;
static PyObject *__pyx_n_s_SPLIT;
static PyObject *__pyx_n_s_SimpleSpellChecker;
static PyObject *__pyx_n_s_SpellChecker_Operator;
static PyObject *__pyx_n_s_SpellChecker_SimpleSpellChecker;
static PyObject *__pyx_kp_u__10;
static PyObject *__pyx_kp_u__2;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_kp_u__4;
static PyObject *__pyx_kp_u__5;
static PyObject *__pyx_kp_u__8;
static PyObject *__pyx_kp_u__9;
static PyObject *__pyx_kp_u_a_zA_Z;
static PyObject *__pyx_n_s_addSplitWords;
static PyObject *__pyx_n_u_ah;
static PyObject *__pyx_n_u_bit;
static PyObject *__pyx_n_u_byte;
static PyObject *__pyx_n_s_candidateList;
static PyObject *__pyx_n_u_cc;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_u_cm;
static PyObject *__pyx_n_u_cm2;
static PyObject *__pyx_n_u_cm3;
static PyObject *__pyx_n_s_compile;
static PyObject *__pyx_n_s_conditionalShortcuts;
static PyObject *__pyx_n_u_da;
static PyObject *__pyx_kp_u_data_merged_txt;
static PyObject *__pyx_kp_u_data_split_txt;
static PyObject *__pyx_n_u_db;
static PyObject *__pyx_n_u_de;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dictionary;
static PyObject *__pyx_n_u_dpi;
static PyObject *__pyx_n_s_encoding;
static PyObject *__pyx_n_s_forcedBackwardMergeCheck;
static PyObject *__pyx_n_s_forcedDeDaSplitCheck;
static PyObject *__pyx_n_s_forcedForwardMergeCheck;
static PyObject *__pyx_n_s_forcedHyphenMergeCheck;
static PyObject *__pyx_n_s_forcedMisspellCheck;
static PyObject *__pyx_n_s_forcedQuestionSuffixSplitCheck;
static PyObject *__pyx_n_s_forcedShortcutCheck;
static PyObject *__pyx_n_s_forcedSplitCheck;
static PyObject *__pyx_n_s_forcedSuffixMergeCheck;
static PyObject *__pyx_n_s_fsm;
static PyObject *__pyx_n_u_ft;
static PyObject *__pyx_n_s_fullmatch;
static PyObject *__pyx_n_u_g;
static PyObject *__pyx_n_u_gb;
static PyObject *__pyx_n_u_gbit;
static PyObject *__pyx_n_u_gbps;
static PyObject *__pyx_n_s_generateCandidateList;
static PyObject *__pyx_n_s_getCorrectForm;
static PyObject *__pyx_n_s_getName;
static PyObject *__pyx_n_s_getOperator;
static PyObject *__pyx_n_s_getSplitPair;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_u_ghz;
static PyObject *__pyx_n_u_gr;
static PyObject *__pyx_n_u_gram;
static PyObject *__pyx_n_u_hp;
static PyObject *__pyx_n_u_hz;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_u_in;
static PyObject *__pyx_n_u_inc;
static PyObject *__pyx_n_u_inch;
static PyObject *__pyx_n_s_isBackVowel;
static PyObject *__pyx_n_u_kbit;
static PyObject *__pyx_n_u_kcal;
static PyObject *__pyx_n_u_kg;
static PyObject *__pyx_n_u_khz;
static PyObject *__pyx_n_u_kva;
static PyObject *__pyx_n_u_kw;
static PyObject *__pyx_n_u_kwh;
static PyObject *__pyx_n_u_l;
static PyObject *__pyx_n_u_l_2;
static PyObject *__pyx_n_u_l_3;
static PyObject *__pyx_n_s_lastVowel;
static PyObject *__pyx_n_u_li;
static PyObject *__pyx_n_u_lik;
static PyObject *__pyx_n_u_litre;
static PyObject *__pyx_n_u_lk;
static PyObject *__pyx_n_u_lk_2;
static PyObject *__pyx_n_s_loadDictionaries;
static PyObject *__pyx_n_u_lt;
static PyObject *__pyx_n_u_lu;
static PyObject *__pyx_n_u_luk;
static PyObject *__pyx_n_u_m;
static PyObject *__pyx_n_u_m2;
static PyObject *__pyx_n_u_m3;
static PyObject *__pyx_n_u_m_2;
static PyObject *__pyx_n_u_m_3;
static PyObject *__pyx_n_u_mah;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_u_mb;
static PyObject *__pyx_n_u_mbit;
static PyObject *__pyx_n_u_mbps;
static PyObject *__pyx_n_u_mdr;
static PyObject *__pyx_n_u_mdr_2;
static PyObject *__pyx_n_s_mergedCandidatesList;
static PyObject *__pyx_n_u_metre;
static PyObject *__pyx_n_u_mg;
static PyObject *__pyx_n_u_mhz;
static PyObject *__pyx_n_u_mi;
static PyObject *__pyx_n_u_midir;
static PyObject *__pyx_n_u_misin;
static PyObject *__pyx_n_u_misindir;
static PyObject *__pyx_n_u_misiniz;
static PyObject *__pyx_n_u_misinizdir;
static PyObject *__pyx_n_u_miydi;
static PyObject *__pyx_n_u_miydik;
static PyObject *__pyx_n_u_miydiler;
static PyObject *__pyx_n_u_miydim;
static PyObject *__pyx_n_u_miydin;
static PyObject *__pyx_n_u_miydiniz;
static PyObject *__pyx_n_u_miyim;
static PyObject *__pyx_n_u_miyimdir;
static PyObject *__pyx_n_u_miyiz;
static PyObject *__pyx_n_u_miyizdir;
static PyObject *__pyx_n_u_miymi;
static PyObject *__pyx_n_u_miymiim;
static PyObject *__pyx_n_u_miymiiz;
static PyObject *__pyx_n_u_miymiler;
static PyObject *__pyx_n_u_miymisin;
static PyObject *__pyx_n_u_ml;
static PyObject *__pyx_n_u_mm;
static PyObject *__pyx_n_u_mm2;
static PyObject *__pyx_n_u_mm3;
static PyObject *__pyx_n_u_mp;
static PyObject *__pyx_n_u_ms;
static PyObject *__pyx_n_u_msn;
static PyObject *__pyx_n_u_msn_2;
static PyObject *__pyx_n_u_msndr;
static PyObject *__pyx_n_u_msndr_2;
static PyObject *__pyx_n_u_msnz;
static PyObject *__pyx_n_u_msnz_2;
static PyObject *__pyx_n_u_msnzdr;
static PyObject *__pyx_n_u_msnzdr_2;
static PyObject *__pyx_n_u_mt;
static PyObject *__pyx_n_u_mu;
static PyObject *__pyx_n_u_mudur;
static PyObject *__pyx_n_s_multiWord;
static PyObject *__pyx_n_u_musun;
static PyObject *__pyx_n_u_musundur;
static PyObject *__pyx_n_u_musunuz;
static PyObject *__pyx_n_u_musunuzdur;
static PyObject *__pyx_n_u_muydu;
static PyObject *__pyx_n_u_muyduk;
static PyObject *__pyx_n_u_muydular;
static PyObject *__pyx_n_u_muydum;
static PyObject *__pyx_n_u_muydun;
static PyObject *__pyx_n_u_muydunuz;
static PyObject *__pyx_n_u_muymu;
static PyObject *__pyx_n_u_muymular;
static PyObject *__pyx_n_u_muymusun;
static PyObject *__pyx_n_u_muymuum;
static PyObject *__pyx_n_u_muymuuz;
static PyObject *__pyx_n_u_muyum;
static PyObject *__pyx_n_u_muyumdur;
static PyObject *__pyx_n_u_muyuz;
static PyObject *__pyx_n_u_muyuzdur;
static PyObject *__pyx_n_u_mv;
static PyObject *__pyx_n_u_myd;
static PyObject *__pyx_n_u_myd_2;
static PyObject *__pyx_n_u_mydk;
static PyObject *__pyx_n_u_mydk_2;
static PyObject *__pyx_n_u_mydlar;
static PyObject *__pyx_n_u_mydler;
static PyObject *__pyx_n_u_mydm;
static PyObject *__pyx_n_u_mydm_2;
static PyObject *__pyx_n_u_mydn;
static PyObject *__pyx_n_u_mydn_2;
static PyObject *__pyx_n_u_mydnz;
static PyObject *__pyx_n_u_mydnz_2;
static PyObject *__pyx_n_u_mym;
static PyObject *__pyx_n_u_mym_2;
static PyObject *__pyx_n_u_mym_3;
static PyObject *__pyx_n_u_mym_4;
static PyObject *__pyx_n_u_mymdr;
static PyObject *__pyx_n_u_mymdr_2;
static PyObject *__pyx_n_u_mymlar;
static PyObject *__pyx_n_u_mymler;
static PyObject *__pyx_n_u_mymm;
static PyObject *__pyx_n_u_mymm_2;
static PyObject *__pyx_n_u_mymsn;
static PyObject *__pyx_n_u_mymsn_2;
static PyObject *__pyx_n_u_mymz;
static PyObject *__pyx_n_u_mymz_2;
static PyObject *__pyx_n_u_myz;
static PyObject *__pyx_n_u_myz_2;
static PyObject *__pyx_n_u_myzdr;
static PyObject *__pyx_n_u_myzdr_2;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_nextWord;
static PyObject *__pyx_n_u_ohm;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_u_oz;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pkg_resources;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_u_ppm;
static PyObject *__pyx_n_s_previousWord;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_SimpleSpellChecke;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_questionSuffixList;
static PyObject *__pyx_n_u_r;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_randrange;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_re;
static PyObject *__pyx_n_s_readlines;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_resource_filename;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_rindex;
static PyObject *__pyx_n_u_rpm;
static PyObject *__pyx_n_u_s;
static PyObject *__pyx_n_s_sentence;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shortcuts;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_u_sn;
static PyObject *__pyx_n_s_spellCheck;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_splitCandidatesList;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_u_tb;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_u_tl;
static PyObject *__pyx_n_s_toCapital;
static PyObject *__pyx_n_s_toLower;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_u_utf8;
static PyObject *__pyx_n_u_v;
static PyObject *__pyx_n_u_va;
static PyObject *__pyx_n_u_volt;
static PyObject *__pyx_n_u_w;
static PyObject *__pyx_n_u_watt;
static PyObject *__pyx_n_s_word;
static PyObject *__pyx_n_s_wordName;
static int __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___init__(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *__pyx_v_fsm); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_2__generateCandidateList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v_word); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_4candidateList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_sentence); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_6spellCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_sentence); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_8forcedMisspellCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_10forcedBackwardMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_12forcedForwardMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_14addSplitWords(CYTHON_UNUSED struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v_multiWord, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_16forcedSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_18forcedShortcutCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_20forcedDeDaSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_22forcedSuffixMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_24forcedHyphenMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_26forcedQuestionSuffixSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_28loadDictionaries(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_30getCorrectForm(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v_wordName, PyObject *__pyx_v_dictionary); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_32mergedCandidatesList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_34splitCandidatesList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_36getSplitPair(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_38__reduce_cython__(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_40__setstate_cython__(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker___pyx_unpickle_SimpleSpellChecker(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type_rindex = {0, &__pyx_n_s_rindex, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type_strip = {0, &__pyx_n_s_strip, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_23252194;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_codeobj__12;
/* Late includes */

/* "SpellChecker/SimpleSpellChecker.pyx":32
 *                                   "mymler", "misinizdir", "msnzdr", "musunuzdur", "msnzdr"]
 * 
 *     def __init__(self, fsm: FsmMorphologicalAnalyzer):             # <<<<<<<<<<<<<<
 *         """
 *         A constructor of SimpleSpellChecker class which takes a FsmMorphologicalAnalyzer as an input and
 */

/* Python wrapper */
static int __pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___init__[] = "\n        A constructor of SimpleSpellChecker class which takes a FsmMorphologicalAnalyzer as an input and\n        assigns it to the fsm variable.\n\n        PARAMETERS\n        ----------\n        fsm : FsmMorphologicalAnalyzer\n            FsmMorphologicalAnalyzer type input.\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___init__;
#endif
static int __pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *__pyx_v_fsm = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fsm,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fsm)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 32, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_fsm = ((struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 32, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fsm), __pyx_ptype_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer, 1, "fsm", 0))) __PYX_ERR(0, 32, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___init__(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_fsm);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___init__(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *__pyx_v_fsm) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "SpellChecker/SimpleSpellChecker.pyx":42
 *             FsmMorphologicalAnalyzer type input.
 *         """
 *         self.fsm = fsm             # <<<<<<<<<<<<<<
 *         self.__merged_words = {}
 *         self.__split_words = {}
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_fsm));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_fsm));
  __Pyx_GOTREF(__pyx_v_self->fsm);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->fsm));
  __pyx_v_self->fsm = __pyx_v_fsm;

  /* "SpellChecker/SimpleSpellChecker.pyx":43
 *         """
 *         self.fsm = fsm
 *         self.__merged_words = {}             # <<<<<<<<<<<<<<
 *         self.__split_words = {}
 *         self.loadDictionaries()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx___merged_words);
  __Pyx_DECREF(__pyx_v_self->__pyx___merged_words);
  __pyx_v_self->__pyx___merged_words = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":44
 *         self.fsm = fsm
 *         self.__merged_words = {}
 *         self.__split_words = {}             # <<<<<<<<<<<<<<
 *         self.loadDictionaries()
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx___split_words);
  __Pyx_DECREF(__pyx_v_self->__pyx___split_words);
  __pyx_v_self->__pyx___split_words = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":45
 *         self.__merged_words = {}
 *         self.__split_words = {}
 *         self.loadDictionaries()             # <<<<<<<<<<<<<<
 * 
 *     cpdef list __generateCandidateList(self, str word):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->loadDictionaries(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":32
 *                                   "mymler", "misinizdir", "msnzdr", "musunuzdur", "msnzdr"]
 * 
 *     def __init__(self, fsm: FsmMorphologicalAnalyzer):             # <<<<<<<<<<<<<<
 *         """
 *         A constructor of SimpleSpellChecker class which takes a FsmMorphologicalAnalyzer as an input and
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":47
 *         self.loadDictionaries()
 * 
 *     cpdef list __generateCandidateList(self, str word):             # <<<<<<<<<<<<<<
 *         """
 *         The generateCandidateList method takes a String as an input. Firstly, it creates a String consists of lowercase
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_3__generateCandidateList(PyObject *__pyx_v_self, PyObject *__pyx_v_word); /*proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___generateCandidateList(CYTHON_UNUSED struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v_word, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_s = 0;
  PyObject *__pyx_v_swapped = 0;
  PyObject *__pyx_v_deleted = 0;
  PyObject *__pyx_v_replaced = 0;
  PyObject *__pyx_v_added = 0;
  PyObject *__pyx_v_candidates = 0;
  int __pyx_v_i;
  int __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  long __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__generateCandidateList", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_generateCandidateList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_3__generateCandidateList)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_word) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_word);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 47, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":72
 *         cdef list candidates
 *         cdef int i, j
 *         s = TurkishLanguage.LOWERCASE_LETTERS             # <<<<<<<<<<<<<<
 *         candidates = []
 *         for i in range(len(word)):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8Language_15TurkishLanguage_TurkishLanguage), __pyx_n_s_LOWERCASE_LETTERS); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 72, __pyx_L1_error)
  __pyx_v_s = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":73
 *         cdef int i, j
 *         s = TurkishLanguage.LOWERCASE_LETTERS
 *         candidates = []             # <<<<<<<<<<<<<<
 *         for i in range(len(word)):
 *             if i < len(word) - 1:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_candidates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":74
 *         s = TurkishLanguage.LOWERCASE_LETTERS
 *         candidates = []
 *         for i in range(len(word)):             # <<<<<<<<<<<<<<
 *             if i < len(word) - 1:
 *                 swapped = word[:i] + word[i + 1] + word[i] + word[i + 2:]
 */
  if (unlikely(__pyx_v_word == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 74, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyUnicode_GET_LENGTH(__pyx_v_word); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 74, __pyx_L1_error)
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "SpellChecker/SimpleSpellChecker.pyx":75
 *         candidates = []
 *         for i in range(len(word)):
 *             if i < len(word) - 1:             # <<<<<<<<<<<<<<
 *                 swapped = word[:i] + word[i + 1] + word[i] + word[i + 2:]
 *                 candidates.append(Candidate(swapped, Operator.SPELL_CHECK))
 */
    if (unlikely(__pyx_v_word == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 75, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyUnicode_GET_LENGTH(__pyx_v_word); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 75, __pyx_L1_error)
    __pyx_t_9 = ((__pyx_v_i < (__pyx_t_8 - 1)) != 0);
    if (__pyx_t_9) {

      /* "SpellChecker/SimpleSpellChecker.pyx":76
 *         for i in range(len(word)):
 *             if i < len(word) - 1:
 *                 swapped = word[:i] + word[i + 1] + word[i] + word[i + 2:]             # <<<<<<<<<<<<<<
 *                 candidates.append(Candidate(swapped, Operator.SPELL_CHECK))
 *             if word[i] in TurkishLanguage.LETTERS or word[i] in "wqx":
 */
      if (unlikely(__pyx_v_word == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 76, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_word, 0, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = (__pyx_v_i + 1);
      __pyx_t_11 = __Pyx_GetItemInt_Unicode(__pyx_v_word, __pyx_t_10, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(__pyx_t_11 == (Py_UCS4)-1)) __PYX_ERR(0, 76, __pyx_L1_error)
      __pyx_t_2 = PyUnicode_FromOrdinal(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = __Pyx_GetItemInt_Unicode(__pyx_v_word, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(__pyx_t_11 == (Py_UCS4)-1)) __PYX_ERR(0, 76, __pyx_L1_error)
      __pyx_t_2 = PyUnicode_FromOrdinal(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(__pyx_v_word == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 76, __pyx_L1_error)
      }
      __pyx_t_2 = __Pyx_PyUnicode_Substring(__pyx_v_word, (__pyx_v_i + 2), PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_swapped, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":77
 *             if i < len(word) - 1:
 *                 swapped = word[:i] + word[i + 1] + word[i] + word[i + 2:]
 *                 candidates.append(Candidate(swapped, Operator.SPELL_CHECK))             # <<<<<<<<<<<<<<
 *             if word[i] in TurkishLanguage.LETTERS or word[i] in "wqx":
 *                 deleted = word[:i] + word[i + 1:]
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Operator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SPELL_CHECK); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_swapped);
      __Pyx_GIVEREF(__pyx_v_swapped);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_swapped);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12SpellChecker_9Candidate_Candidate), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_t_2); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 77, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":75
 *         candidates = []
 *         for i in range(len(word)):
 *             if i < len(word) - 1:             # <<<<<<<<<<<<<<
 *                 swapped = word[:i] + word[i + 1] + word[i] + word[i + 2:]
 *                 candidates.append(Candidate(swapped, Operator.SPELL_CHECK))
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":78
 *                 swapped = word[:i] + word[i + 1] + word[i] + word[i + 2:]
 *                 candidates.append(Candidate(swapped, Operator.SPELL_CHECK))
 *             if word[i] in TurkishLanguage.LETTERS or word[i] in "wqx":             # <<<<<<<<<<<<<<
 *                 deleted = word[:i] + word[i + 1:]
 *                 candidates.append(Candidate(deleted, Operator.SPELL_CHECK))
 */
    __pyx_t_11 = __Pyx_GetItemInt_Unicode(__pyx_v_word, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(__pyx_t_11 == (Py_UCS4)-1)) __PYX_ERR(0, 78, __pyx_L1_error)
    __pyx_t_2 = PyUnicode_FromOrdinal(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8Language_15TurkishLanguage_TurkishLanguage), __pyx_n_s_LETTERS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_t_3, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 78, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_14 = (__pyx_t_13 != 0);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_9 = __pyx_t_14;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_GetItemInt_Unicode(__pyx_v_word, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(__pyx_t_11 == (Py_UCS4)-1)) __PYX_ERR(0, 78, __pyx_L1_error)
    switch (__pyx_t_11) {
      case 0x71:
      case 0x77:
      case 0x78:
      __pyx_t_14 = 1;
      break;
      default:
      __pyx_t_14 = 0;
      break;
    }
    __pyx_t_13 = (__pyx_t_14 != 0);
    __pyx_t_9 = __pyx_t_13;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_9) {

      /* "SpellChecker/SimpleSpellChecker.pyx":79
 *                 candidates.append(Candidate(swapped, Operator.SPELL_CHECK))
 *             if word[i] in TurkishLanguage.LETTERS or word[i] in "wqx":
 *                 deleted = word[:i] + word[i + 1:]             # <<<<<<<<<<<<<<
 *                 candidates.append(Candidate(deleted, Operator.SPELL_CHECK))
 *                 for j in range(len(s)):
 */
      if (unlikely(__pyx_v_word == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 79, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyUnicode_Substring(__pyx_v_word, 0, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__pyx_v_word == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 79, __pyx_L1_error)
      }
      __pyx_t_2 = __Pyx_PyUnicode_Substring(__pyx_v_word, (__pyx_v_i + 1), PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_deleted, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":80
 *             if word[i] in TurkishLanguage.LETTERS or word[i] in "wqx":
 *                 deleted = word[:i] + word[i + 1:]
 *                 candidates.append(Candidate(deleted, Operator.SPELL_CHECK))             # <<<<<<<<<<<<<<
 *                 for j in range(len(s)):
 *                     replaced = word[:i] + s[j] + word[i + 1:]
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Operator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_SPELL_CHECK); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_deleted);
      __Pyx_GIVEREF(__pyx_v_deleted);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_deleted);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12SpellChecker_9Candidate_Candidate), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_t_2); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":81
 *                 deleted = word[:i] + word[i + 1:]
 *                 candidates.append(Candidate(deleted, Operator.SPELL_CHECK))
 *                 for j in range(len(s)):             # <<<<<<<<<<<<<<
 *                     replaced = word[:i] + s[j] + word[i + 1:]
 *                     candidates.append(Candidate(replaced, Operator.SPELL_CHECK))
 */
      if (unlikely(__pyx_v_s == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 81, __pyx_L1_error)
      }
      __pyx_t_8 = __Pyx_PyUnicode_GET_LENGTH(__pyx_v_s); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 81, __pyx_L1_error)
      __pyx_t_15 = __pyx_t_8;
      for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
        __pyx_v_j = __pyx_t_16;

        /* "SpellChecker/SimpleSpellChecker.pyx":82
 *                 candidates.append(Candidate(deleted, Operator.SPELL_CHECK))
 *                 for j in range(len(s)):
 *                     replaced = word[:i] + s[j] + word[i + 1:]             # <<<<<<<<<<<<<<
 *                     candidates.append(Candidate(replaced, Operator.SPELL_CHECK))
 *                 for j in range(len(s)):
 */
        if (unlikely(__pyx_v_word == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 82, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_PyUnicode_Substring(__pyx_v_word, 0, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_11 = __Pyx_GetItemInt_Unicode(__pyx_v_s, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(__pyx_t_11 == (Py_UCS4)-1)) __PYX_ERR(0, 82, __pyx_L1_error)
        __pyx_t_1 = PyUnicode_FromOrdinal(__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(__pyx_v_word == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 82, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_word, (__pyx_v_i + 1), PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_replaced, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":83
 *                 for j in range(len(s)):
 *                     replaced = word[:i] + s[j] + word[i + 1:]
 *                     candidates.append(Candidate(replaced, Operator.SPELL_CHECK))             # <<<<<<<<<<<<<<
 *                 for j in range(len(s)):
 *                     added = word[:i] + s[j] + word[i:]
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Operator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SPELL_CHECK); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_v_replaced);
        __Pyx_GIVEREF(__pyx_v_replaced);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_replaced);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
        __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12SpellChecker_9Candidate_Candidate), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 83, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":84
 *                     replaced = word[:i] + s[j] + word[i + 1:]
 *                     candidates.append(Candidate(replaced, Operator.SPELL_CHECK))
 *                 for j in range(len(s)):             # <<<<<<<<<<<<<<
 *                     added = word[:i] + s[j] + word[i:]
 *                     candidates.append(Candidate(added, Operator.SPELL_CHECK))
 */
      if (unlikely(__pyx_v_s == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 84, __pyx_L1_error)
      }
      __pyx_t_8 = __Pyx_PyUnicode_GET_LENGTH(__pyx_v_s); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 84, __pyx_L1_error)
      __pyx_t_15 = __pyx_t_8;
      for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_15; __pyx_t_16+=1) {
        __pyx_v_j = __pyx_t_16;

        /* "SpellChecker/SimpleSpellChecker.pyx":85
 *                     candidates.append(Candidate(replaced, Operator.SPELL_CHECK))
 *                 for j in range(len(s)):
 *                     added = word[:i] + s[j] + word[i:]             # <<<<<<<<<<<<<<
 *                     candidates.append(Candidate(added, Operator.SPELL_CHECK))
 *         return candidates
 */
        if (unlikely(__pyx_v_word == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 85, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_v_word, 0, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_GetItemInt_Unicode(__pyx_v_s, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(__pyx_t_11 == (Py_UCS4)-1)) __PYX_ERR(0, 85, __pyx_L1_error)
        __pyx_t_2 = PyUnicode_FromOrdinal(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(__pyx_v_word == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 85, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_PyUnicode_Substring(__pyx_v_word, __pyx_v_i, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_added, ((PyObject*)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":86
 *                 for j in range(len(s)):
 *                     added = word[:i] + s[j] + word[i:]
 *                     candidates.append(Candidate(added, Operator.SPELL_CHECK))             # <<<<<<<<<<<<<<
 *         return candidates
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Operator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_SPELL_CHECK); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_v_added);
        __Pyx_GIVEREF(__pyx_v_added);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_added);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12SpellChecker_9Candidate_Candidate), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_t_2); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 86, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":78
 *                 swapped = word[:i] + word[i + 1] + word[i] + word[i + 2:]
 *                 candidates.append(Candidate(swapped, Operator.SPELL_CHECK))
 *             if word[i] in TurkishLanguage.LETTERS or word[i] in "wqx":             # <<<<<<<<<<<<<<
 *                 deleted = word[:i] + word[i + 1:]
 *                 candidates.append(Candidate(deleted, Operator.SPELL_CHECK))
 */
    }
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":87
 *                     added = word[:i] + s[j] + word[i:]
 *                     candidates.append(Candidate(added, Operator.SPELL_CHECK))
 *         return candidates             # <<<<<<<<<<<<<<
 * 
 *     cpdef list candidateList(self, Word word, Sentence sentence):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_candidates);
  __pyx_r = __pyx_v_candidates;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":47
 *         self.loadDictionaries()
 * 
 *     cpdef list __generateCandidateList(self, str word):             # <<<<<<<<<<<<<<
 *         """
 *         The generateCandidateList method takes a String as an input. Firstly, it creates a String consists of lowercase
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.__generateCandidateList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_swapped);
  __Pyx_XDECREF(__pyx_v_deleted);
  __Pyx_XDECREF(__pyx_v_replaced);
  __Pyx_XDECREF(__pyx_v_added);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_3__generateCandidateList(PyObject *__pyx_v_self, PyObject *__pyx_v_word); /*proto*/
static char __pyx_doc_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_2__generateCandidateList[] = "\n        The generateCandidateList method takes a String as an input. Firstly, it creates a String consists of lowercase\n        Turkish letters and an list candidates. Then, it loops i times where i ranges from 0 to the length of given\n        word. It gets substring from 0 to ith index and concatenates it with substring from i+1 to the last index as a\n        new String called deleted. Then, adds this String to the candidates list. Secondly, it loops j times where j\n        ranges from 0 to length of lowercase letters String and adds the jth character of this String between substring\n        of given word from 0 to ith index and the substring from i+1 to the last index, then adds it to the candidates\n        list. Thirdly, it loops j times where j ranges from 0 to length of lowercase letters String and adds the jth\n        character of this String between substring of given word from 0 to ith index and the substring from i to the\n        last index, then adds it to the candidates list.\n\n        PARAMETERS\n        ----------\n        word : str\n            String input.\n\n        RETURNS\n        -------\n        list\n            List candidates.\n        ";
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_3__generateCandidateList(PyObject *__pyx_v_self, PyObject *__pyx_v_word) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__generateCandidateList (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), (&PyUnicode_Type), 1, "word", 1))) __PYX_ERR(0, 47, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_2__generateCandidateList(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), ((PyObject*)__pyx_v_word));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_2__generateCandidateList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v_word) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__generateCandidateList", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___generateCandidateList(__pyx_v_self, __pyx_v_word, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.__generateCandidateList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":89
 *         return candidates
 * 
 *     cpdef list candidateList(self, Word word, Sentence sentence):             # <<<<<<<<<<<<<<
 *         """
 *         The candidateList method takes a Word as an input and creates a candidates list by calling generateCandidateList
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_5candidateList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_candidateList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, CYTHON_UNUSED struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_sentence, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_candidates = 0;
  int __pyx_v_i;
  struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *__pyx_v_fsm_parse_list = 0;
  PyObject *__pyx_v_new_candidate = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("candidateList", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_candidateList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_5candidateList)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_sentence)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_sentence)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 89, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_sentence));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_sentence));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_sentence));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 89, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":112
 *         cdef FsmParseList fsm_parse_list
 *         cdef str new_candidate
 *         candidates = self.__generateCandidateList(word.getName())             # <<<<<<<<<<<<<<
 *         i = 0
 *         while i < len(candidates):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx___generateCandidateList(__pyx_v_self, ((PyObject*)__pyx_t_1), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_candidates = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":113
 *         cdef str new_candidate
 *         candidates = self.__generateCandidateList(word.getName())
 *         i = 0             # <<<<<<<<<<<<<<
 *         while i < len(candidates):
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(candidates[i].getName())
 */
  __pyx_v_i = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":114
 *         candidates = self.__generateCandidateList(word.getName())
 *         i = 0
 *         while i < len(candidates):             # <<<<<<<<<<<<<<
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(candidates[i].getName())
 *             if fsm_parse_list.size() == 0:
 */
  while (1) {
    if (unlikely(__pyx_v_candidates == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 114, __pyx_L1_error)
    }
    __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 114, __pyx_L1_error)
    __pyx_t_8 = ((__pyx_v_i < __pyx_t_7) != 0);
    if (!__pyx_t_8) break;

    /* "SpellChecker/SimpleSpellChecker.pyx":115
 *         i = 0
 *         while i < len(candidates):
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(candidates[i].getName())             # <<<<<<<<<<<<<<
 *             if fsm_parse_list.size() == 0:
 *                 new_candidate = self.fsm.getDictionary().getCorrectForm(candidates[i].getName())
 */
    if (unlikely(__pyx_v_candidates == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 115, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_candidates, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_getName); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList))))) __PYX_ERR(0, 115, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fsm_parse_list, ((struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":116
 *         while i < len(candidates):
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(candidates[i].getName())
 *             if fsm_parse_list.size() == 0:             # <<<<<<<<<<<<<<
 *                 new_candidate = self.fsm.getDictionary().getCorrectForm(candidates[i].getName())
 *                 if new_candidate != "" and self.fsm.morphologicalAnalysis(new_candidate).size() > 0:
 */
    __pyx_t_8 = ((((struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_v_fsm_parse_list->__pyx_vtab)->size(__pyx_v_fsm_parse_list, 0) == 0) != 0);
    if (__pyx_t_8) {

      /* "SpellChecker/SimpleSpellChecker.pyx":117
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(candidates[i].getName())
 *             if fsm_parse_list.size() == 0:
 *                 new_candidate = self.fsm.getDictionary().getCorrectForm(candidates[i].getName())             # <<<<<<<<<<<<<<
 *                 if new_candidate != "" and self.fsm.morphologicalAnalysis(new_candidate).size() > 0:
 *                     candidates[i] = Candidate(new_candidate, Operator.MISSPELLED_REPLACE)
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->getDictionary(__pyx_v_self->fsm, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__pyx_v_candidates == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 117, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_candidates, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_getName); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 117, __pyx_L1_error)
      __pyx_t_6 = ((struct __pyx_vtabstruct_10Dictionary_13TxtDictionary_TxtDictionary *)((struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *)__pyx_t_3)->__pyx_base.__pyx_vtab)->getCorrectForm(((struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *)__pyx_t_3), ((PyObject*)__pyx_t_2), 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_new_candidate, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":118
 *             if fsm_parse_list.size() == 0:
 *                 new_candidate = self.fsm.getDictionary().getCorrectForm(candidates[i].getName())
 *                 if new_candidate != "" and self.fsm.morphologicalAnalysis(new_candidate).size() > 0:             # <<<<<<<<<<<<<<
 *                     candidates[i] = Candidate(new_candidate, Operator.MISSPELLED_REPLACE)
 *                 else:
 */
      __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_v_new_candidate, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 118, __pyx_L1_error)
      __pyx_t_10 = (__pyx_t_9 != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_8 = __pyx_t_10;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_2 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_v_new_candidate, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_6, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_8 = __pyx_t_10;
      __pyx_L7_bool_binop_done:;
      if (__pyx_t_8) {

        /* "SpellChecker/SimpleSpellChecker.pyx":119
 *                 new_candidate = self.fsm.getDictionary().getCorrectForm(candidates[i].getName())
 *                 if new_candidate != "" and self.fsm.morphologicalAnalysis(new_candidate).size() > 0:
 *                     candidates[i] = Candidate(new_candidate, Operator.MISSPELLED_REPLACE)             # <<<<<<<<<<<<<<
 *                 else:
 *                     candidates.pop(i)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Operator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MISSPELLED_REPLACE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_v_new_candidate);
        __Pyx_GIVEREF(__pyx_v_new_candidate);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_new_candidate);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12SpellChecker_9Candidate_Candidate), __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(__pyx_v_candidates == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 119, __pyx_L1_error)
        }
        if (unlikely(__Pyx_SetItemInt(__pyx_v_candidates, __pyx_v_i, __pyx_t_6, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":118
 *             if fsm_parse_list.size() == 0:
 *                 new_candidate = self.fsm.getDictionary().getCorrectForm(candidates[i].getName())
 *                 if new_candidate != "" and self.fsm.morphologicalAnalysis(new_candidate).size() > 0:             # <<<<<<<<<<<<<<
 *                     candidates[i] = Candidate(new_candidate, Operator.MISSPELLED_REPLACE)
 *                 else:
 */
        goto __pyx_L6;
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":121
 *                     candidates[i] = Candidate(new_candidate, Operator.MISSPELLED_REPLACE)
 *                 else:
 *                     candidates.pop(i)             # <<<<<<<<<<<<<<
 *                     i = i - 1
 *             i = i + 1
 */
      /*else*/ {
        if (unlikely(__pyx_v_candidates == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
          __PYX_ERR(0, 121, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_PyList_PopIndex(__pyx_v_candidates, Py_None, __pyx_v_i, 1, int, __Pyx_PyInt_From_int); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":122
 *                 else:
 *                     candidates.pop(i)
 *                     i = i - 1             # <<<<<<<<<<<<<<
 *             i = i + 1
 *         return candidates
 */
        __pyx_v_i = (__pyx_v_i - 1);
      }
      __pyx_L6:;

      /* "SpellChecker/SimpleSpellChecker.pyx":116
 *         while i < len(candidates):
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(candidates[i].getName())
 *             if fsm_parse_list.size() == 0:             # <<<<<<<<<<<<<<
 *                 new_candidate = self.fsm.getDictionary().getCorrectForm(candidates[i].getName())
 *                 if new_candidate != "" and self.fsm.morphologicalAnalysis(new_candidate).size() > 0:
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":123
 *                     candidates.pop(i)
 *                     i = i - 1
 *             i = i + 1             # <<<<<<<<<<<<<<
 *         return candidates
 * 
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":124
 *                     i = i - 1
 *             i = i + 1
 *         return candidates             # <<<<<<<<<<<<<<
 * 
 *     cpdef Sentence spellCheck(self, Sentence sentence):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_candidates);
  __pyx_r = __pyx_v_candidates;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":89
 *         return candidates
 * 
 *     cpdef list candidateList(self, Word word, Sentence sentence):             # <<<<<<<<<<<<<<
 *         """
 *         The candidateList method takes a Word as an input and creates a candidates list by calling generateCandidateList
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.candidateList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF((PyObject *)__pyx_v_fsm_parse_list);
  __Pyx_XDECREF(__pyx_v_new_candidate);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_5candidateList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_4candidateList[] = "\n        The candidateList method takes a Word as an input and creates a candidates list by calling generateCandidateList\n        method with given word. Then, it loop i times where i ranges from 0 to size of candidates list and creates a\n        FsmParseList by calling morphologicalAnalysis with each item of candidates list. If the size of FsmParseList is\n        0, it then removes the ith item.\n\n        PARAMETERS\n        ----------\n        word : Word\n            Word input.\n        sentence: Sentence\n            Sentence input.\n\n        RETURNS\n        -------\n        list\n            candidates list.\n        ";
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_5candidateList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_sentence = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("candidateList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word,&__pyx_n_s_sentence,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sentence)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("candidateList", 1, 2, 2, 1); __PYX_ERR(0, 89, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "candidateList") < 0)) __PYX_ERR(0, 89, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_sentence = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("candidateList", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 89, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.candidateList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 89, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sentence), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "sentence", 0))) __PYX_ERR(0, 89, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_4candidateList(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_word, __pyx_v_sentence);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_4candidateList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_sentence) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("candidateList", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_candidateList(__pyx_v_self, __pyx_v_word, __pyx_v_sentence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.candidateList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":126
 *         return candidates
 * 
 *     cpdef Sentence spellCheck(self, Sentence sentence):             # <<<<<<<<<<<<<<
 *         """
 *         The spellCheck method takes a Sentence as an input and loops i times where i ranges from 0 to size of words in
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_7spellCheck(PyObject *__pyx_v_self, PyObject *__pyx_v_sentence); /*proto*/
static struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_spellCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_sentence, int __pyx_skip_dispatch) {
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  int __pyx_v_i;
  int __pyx_v_random_candidate;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_new_word = 0;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_next_word = 0;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previous_word = 0;
  struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *__pyx_v_fsm_parse_list = 0;
  PyObject *__pyx_v_candidates = 0;
  PyObject *__pyx_v_upper_case_fsm_parse_list = NULL;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("spellCheck", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_spellCheck); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_7spellCheck)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_sentence)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_sentence));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6Corpus_8Sentence_Sentence))))) __PYX_ERR(0, 126, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":150
 *         cdef FsmParseList fsm_parse_list
 *         cdef list candidates
 *         result = Sentence()             # <<<<<<<<<<<<<<
 *         i = 0
 *         while i < sentence.wordCount():
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6Corpus_8Sentence_Sentence)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":151
 *         cdef list candidates
 *         result = Sentence()
 *         i = 0             # <<<<<<<<<<<<<<
 *         while i < sentence.wordCount():
 *             word = sentence.getWord(i)
 */
  __pyx_v_i = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":152
 *         result = Sentence()
 *         i = 0
 *         while i < sentence.wordCount():             # <<<<<<<<<<<<<<
 *             word = sentence.getWord(i)
 *             next_word = None
 */
  while (1) {
    __pyx_t_5 = ((__pyx_v_i < ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_sentence->__pyx_vtab)->wordCount(__pyx_v_sentence, 0)) != 0);
    if (!__pyx_t_5) break;

    /* "SpellChecker/SimpleSpellChecker.pyx":153
 *         i = 0
 *         while i < sentence.wordCount():
 *             word = sentence.getWord(i)             # <<<<<<<<<<<<<<
 *             next_word = None
 *             previous_word = None
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_sentence->__pyx_vtab)->getWord(__pyx_v_sentence, __pyx_v_i, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_word, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":154
 *         while i < sentence.wordCount():
 *             word = sentence.getWord(i)
 *             next_word = None             # <<<<<<<<<<<<<<
 *             previous_word = None
 *             if i > 0:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_next_word, ((struct __pyx_obj_10Dictionary_4Word_Word *)Py_None));

    /* "SpellChecker/SimpleSpellChecker.pyx":155
 *             word = sentence.getWord(i)
 *             next_word = None
 *             previous_word = None             # <<<<<<<<<<<<<<
 *             if i > 0:
 *                 previous_word = sentence.getWord(i - 1)
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_previous_word, ((struct __pyx_obj_10Dictionary_4Word_Word *)Py_None));

    /* "SpellChecker/SimpleSpellChecker.pyx":156
 *             next_word = None
 *             previous_word = None
 *             if i > 0:             # <<<<<<<<<<<<<<
 *                 previous_word = sentence.getWord(i - 1)
 *             if i < sentence.wordCount() - 1:
 */
    __pyx_t_5 = ((__pyx_v_i > 0) != 0);
    if (__pyx_t_5) {

      /* "SpellChecker/SimpleSpellChecker.pyx":157
 *             previous_word = None
 *             if i > 0:
 *                 previous_word = sentence.getWord(i - 1)             # <<<<<<<<<<<<<<
 *             if i < sentence.wordCount() - 1:
 *                 next_word = sentence.getWord(i + 1)
 */
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_sentence->__pyx_vtab)->getWord(__pyx_v_sentence, (__pyx_v_i - 1), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_previous_word, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":156
 *             next_word = None
 *             previous_word = None
 *             if i > 0:             # <<<<<<<<<<<<<<
 *                 previous_word = sentence.getWord(i - 1)
 *             if i < sentence.wordCount() - 1:
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":158
 *             if i > 0:
 *                 previous_word = sentence.getWord(i - 1)
 *             if i < sentence.wordCount() - 1:             # <<<<<<<<<<<<<<
 *                 next_word = sentence.getWord(i + 1)
 *             if self.forcedMisspellCheck(word, result) or \
 */
    __pyx_t_5 = ((__pyx_v_i < (((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_sentence->__pyx_vtab)->wordCount(__pyx_v_sentence, 0) - 1)) != 0);
    if (__pyx_t_5) {

      /* "SpellChecker/SimpleSpellChecker.pyx":159
 *                 previous_word = sentence.getWord(i - 1)
 *             if i < sentence.wordCount() - 1:
 *                 next_word = sentence.getWord(i + 1)             # <<<<<<<<<<<<<<
 *             if self.forcedMisspellCheck(word, result) or \
 *                     self.forcedBackwardMergeCheck(word, result, previous_word) or \
 */
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_sentence->__pyx_vtab)->getWord(__pyx_v_sentence, (__pyx_v_i + 1), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_next_word, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":158
 *             if i > 0:
 *                 previous_word = sentence.getWord(i - 1)
 *             if i < sentence.wordCount() - 1:             # <<<<<<<<<<<<<<
 *                 next_word = sentence.getWord(i + 1)
 *             if self.forcedMisspellCheck(word, result) or \
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":160
 *             if i < sentence.wordCount() - 1:
 *                 next_word = sentence.getWord(i + 1)
 *             if self.forcedMisspellCheck(word, result) or \             # <<<<<<<<<<<<<<
 *                     self.forcedBackwardMergeCheck(word, result, previous_word) or \
 *                     self.forcedSuffixMergeCheck(word, result, previous_word):
 */
    __pyx_t_6 = (((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->forcedMisspellCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, 0) != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L8_bool_binop_done;
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":161
 *                 next_word = sentence.getWord(i + 1)
 *             if self.forcedMisspellCheck(word, result) or \
 *                     self.forcedBackwardMergeCheck(word, result, previous_word) or \             # <<<<<<<<<<<<<<
 *                     self.forcedSuffixMergeCheck(word, result, previous_word):
 *                 i = i + 1
 */
    __pyx_t_6 = (((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->forcedBackwardMergeCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, __pyx_v_previous_word, 0) != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L8_bool_binop_done;
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":162
 *             if self.forcedMisspellCheck(word, result) or \
 *                     self.forcedBackwardMergeCheck(word, result, previous_word) or \
 *                     self.forcedSuffixMergeCheck(word, result, previous_word):             # <<<<<<<<<<<<<<
 *                 i = i + 1
 *                 continue
 */
    __pyx_t_6 = (((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->forcedSuffixMergeCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, __pyx_v_previous_word, 0) != 0);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L8_bool_binop_done:;

    /* "SpellChecker/SimpleSpellChecker.pyx":160
 *             if i < sentence.wordCount() - 1:
 *                 next_word = sentence.getWord(i + 1)
 *             if self.forcedMisspellCheck(word, result) or \             # <<<<<<<<<<<<<<
 *                     self.forcedBackwardMergeCheck(word, result, previous_word) or \
 *                     self.forcedSuffixMergeCheck(word, result, previous_word):
 */
    if (__pyx_t_5) {

      /* "SpellChecker/SimpleSpellChecker.pyx":163
 *                     self.forcedBackwardMergeCheck(word, result, previous_word) or \
 *                     self.forcedSuffixMergeCheck(word, result, previous_word):
 *                 i = i + 1             # <<<<<<<<<<<<<<
 *                 continue
 *             if self.forcedForwardMergeCheck(word, result, next_word) or \
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "SpellChecker/SimpleSpellChecker.pyx":164
 *                     self.forcedSuffixMergeCheck(word, result, previous_word):
 *                 i = i + 1
 *                 continue             # <<<<<<<<<<<<<<
 *             if self.forcedForwardMergeCheck(word, result, next_word) or \
 *                     self.forcedHyphenMergeCheck(word, result, previous_word, next_word):
 */
      goto __pyx_L3_continue;

      /* "SpellChecker/SimpleSpellChecker.pyx":160
 *             if i < sentence.wordCount() - 1:
 *                 next_word = sentence.getWord(i + 1)
 *             if self.forcedMisspellCheck(word, result) or \             # <<<<<<<<<<<<<<
 *                     self.forcedBackwardMergeCheck(word, result, previous_word) or \
 *                     self.forcedSuffixMergeCheck(word, result, previous_word):
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":165
 *                 i = i + 1
 *                 continue
 *             if self.forcedForwardMergeCheck(word, result, next_word) or \             # <<<<<<<<<<<<<<
 *                     self.forcedHyphenMergeCheck(word, result, previous_word, next_word):
 *                 i = i + 2
 */
    __pyx_t_6 = (((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->forcedForwardMergeCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, __pyx_v_next_word, 0) != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L12_bool_binop_done;
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":166
 *                 continue
 *             if self.forcedForwardMergeCheck(word, result, next_word) or \
 *                     self.forcedHyphenMergeCheck(word, result, previous_word, next_word):             # <<<<<<<<<<<<<<
 *                 i = i + 2
 *                 continue
 */
    __pyx_t_6 = (((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->forcedHyphenMergeCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, __pyx_v_previous_word, __pyx_v_next_word, 0) != 0);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L12_bool_binop_done:;

    /* "SpellChecker/SimpleSpellChecker.pyx":165
 *                 i = i + 1
 *                 continue
 *             if self.forcedForwardMergeCheck(word, result, next_word) or \             # <<<<<<<<<<<<<<
 *                     self.forcedHyphenMergeCheck(word, result, previous_word, next_word):
 *                 i = i + 2
 */
    if (__pyx_t_5) {

      /* "SpellChecker/SimpleSpellChecker.pyx":167
 *             if self.forcedForwardMergeCheck(word, result, next_word) or \
 *                     self.forcedHyphenMergeCheck(word, result, previous_word, next_word):
 *                 i = i + 2             # <<<<<<<<<<<<<<
 *                 continue
 *             if self.forcedSplitCheck(word, result) or self.forcedShortcutCheck(word, result) or \
 */
      __pyx_v_i = (__pyx_v_i + 2);

      /* "SpellChecker/SimpleSpellChecker.pyx":168
 *                     self.forcedHyphenMergeCheck(word, result, previous_word, next_word):
 *                 i = i + 2
 *                 continue             # <<<<<<<<<<<<<<
 *             if self.forcedSplitCheck(word, result) or self.forcedShortcutCheck(word, result) or \
 *                     self.forcedDeDaSplitCheck(word, result) or self.forcedQuestionSuffixSplitCheck(word, result):
 */
      goto __pyx_L3_continue;

      /* "SpellChecker/SimpleSpellChecker.pyx":165
 *                 i = i + 1
 *                 continue
 *             if self.forcedForwardMergeCheck(word, result, next_word) or \             # <<<<<<<<<<<<<<
 *                     self.forcedHyphenMergeCheck(word, result, previous_word, next_word):
 *                 i = i + 2
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":169
 *                 i = i + 2
 *                 continue
 *             if self.forcedSplitCheck(word, result) or self.forcedShortcutCheck(word, result) or \             # <<<<<<<<<<<<<<
 *                     self.forcedDeDaSplitCheck(word, result) or self.forcedQuestionSuffixSplitCheck(word, result):
 *                 i = i + 1
 */
    __pyx_t_6 = (((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->forcedSplitCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, 0) != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_6 = (((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->forcedShortcutCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, 0) != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L15_bool_binop_done;
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":170
 *                 continue
 *             if self.forcedSplitCheck(word, result) or self.forcedShortcutCheck(word, result) or \
 *                     self.forcedDeDaSplitCheck(word, result) or self.forcedQuestionSuffixSplitCheck(word, result):             # <<<<<<<<<<<<<<
 *                 i = i + 1
 *                 continue
 */
    __pyx_t_6 = (((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->forcedDeDaSplitCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, 0) != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_6 = (((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->forcedQuestionSuffixSplitCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, 0) != 0);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L15_bool_binop_done:;

    /* "SpellChecker/SimpleSpellChecker.pyx":169
 *                 i = i + 2
 *                 continue
 *             if self.forcedSplitCheck(word, result) or self.forcedShortcutCheck(word, result) or \             # <<<<<<<<<<<<<<
 *                     self.forcedDeDaSplitCheck(word, result) or self.forcedQuestionSuffixSplitCheck(word, result):
 *                 i = i + 1
 */
    if (__pyx_t_5) {

      /* "SpellChecker/SimpleSpellChecker.pyx":171
 *             if self.forcedSplitCheck(word, result) or self.forcedShortcutCheck(word, result) or \
 *                     self.forcedDeDaSplitCheck(word, result) or self.forcedQuestionSuffixSplitCheck(word, result):
 *                 i = i + 1             # <<<<<<<<<<<<<<
 *                 continue
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(word.getName())
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "SpellChecker/SimpleSpellChecker.pyx":172
 *                     self.forcedDeDaSplitCheck(word, result) or self.forcedQuestionSuffixSplitCheck(word, result):
 *                 i = i + 1
 *                 continue             # <<<<<<<<<<<<<<
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(word.getName())
 *             upper_case_fsm_parse_list = self.fsm.morphologicalAnalysis(Word.toCapital(word.getName()))
 */
      goto __pyx_L3_continue;

      /* "SpellChecker/SimpleSpellChecker.pyx":169
 *                 i = i + 2
 *                 continue
 *             if self.forcedSplitCheck(word, result) or self.forcedShortcutCheck(word, result) or \             # <<<<<<<<<<<<<<
 *                     self.forcedDeDaSplitCheck(word, result) or self.forcedQuestionSuffixSplitCheck(word, result):
 *                 i = i + 1
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":173
 *                 i = i + 1
 *                 continue
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(word.getName())             # <<<<<<<<<<<<<<
 *             upper_case_fsm_parse_list = self.fsm.morphologicalAnalysis(Word.toCapital(word.getName()))
 *             if fsm_parse_list.size() == 0 and upper_case_fsm_parse_list.size() == 0:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList))))) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fsm_parse_list, ((struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":174
 *                 continue
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(word.getName())
 *             upper_case_fsm_parse_list = self.fsm.morphologicalAnalysis(Word.toCapital(word.getName()))             # <<<<<<<<<<<<<<
 *             if fsm_parse_list.size() == 0 and upper_case_fsm_parse_list.size() == 0:
 *                 candidates = self.mergedCandidatesList(previous_word, word, next_word)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_n_s_toCapital); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_upper_case_fsm_parse_list, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":175
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(word.getName())
 *             upper_case_fsm_parse_list = self.fsm.morphologicalAnalysis(Word.toCapital(word.getName()))
 *             if fsm_parse_list.size() == 0 and upper_case_fsm_parse_list.size() == 0:             # <<<<<<<<<<<<<<
 *                 candidates = self.mergedCandidatesList(previous_word, word, next_word)
 *                 if len(candidates) < 1:
 */
    __pyx_t_6 = ((((struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_v_fsm_parse_list->__pyx_vtab)->size(__pyx_v_fsm_parse_list, 0) == 0) != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L20_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_upper_case_fsm_parse_list, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __pyx_t_6;
    __pyx_L20_bool_binop_done:;
    if (__pyx_t_5) {

      /* "SpellChecker/SimpleSpellChecker.pyx":176
 *             upper_case_fsm_parse_list = self.fsm.morphologicalAnalysis(Word.toCapital(word.getName()))
 *             if fsm_parse_list.size() == 0 and upper_case_fsm_parse_list.size() == 0:
 *                 candidates = self.mergedCandidatesList(previous_word, word, next_word)             # <<<<<<<<<<<<<<
 *                 if len(candidates) < 1:
 *                     candidates.extend(self.candidateList(word, sentence))
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->mergedCandidatesList(__pyx_v_self, __pyx_v_previous_word, __pyx_v_word, __pyx_v_next_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_candidates, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":177
 *             if fsm_parse_list.size() == 0 and upper_case_fsm_parse_list.size() == 0:
 *                 candidates = self.mergedCandidatesList(previous_word, word, next_word)
 *                 if len(candidates) < 1:             # <<<<<<<<<<<<<<
 *                     candidates.extend(self.candidateList(word, sentence))
 *                 if len(candidates) < 1:
 */
      if (unlikely(__pyx_v_candidates == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 177, __pyx_L1_error)
      }
      __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 177, __pyx_L1_error)
      __pyx_t_5 = ((__pyx_t_7 < 1) != 0);
      if (__pyx_t_5) {

        /* "SpellChecker/SimpleSpellChecker.pyx":178
 *                 candidates = self.mergedCandidatesList(previous_word, word, next_word)
 *                 if len(candidates) < 1:
 *                     candidates.extend(self.candidateList(word, sentence))             # <<<<<<<<<<<<<<
 *                 if len(candidates) < 1:
 *                     candidates.extend(self.splitCandidatesList(word))
 */
        if (unlikely(__pyx_v_candidates == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
          __PYX_ERR(0, 178, __pyx_L1_error)
        }
        __pyx_t_2 = ((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->candidateList(__pyx_v_self, __pyx_v_word, __pyx_v_sentence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = __Pyx_PyList_Extend(__pyx_v_candidates, __pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":177
 *             if fsm_parse_list.size() == 0 and upper_case_fsm_parse_list.size() == 0:
 *                 candidates = self.mergedCandidatesList(previous_word, word, next_word)
 *                 if len(candidates) < 1:             # <<<<<<<<<<<<<<
 *                     candidates.extend(self.candidateList(word, sentence))
 *                 if len(candidates) < 1:
 */
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":179
 *                 if len(candidates) < 1:
 *                     candidates.extend(self.candidateList(word, sentence))
 *                 if len(candidates) < 1:             # <<<<<<<<<<<<<<
 *                     candidates.extend(self.splitCandidatesList(word))
 *                 if len(candidates) > 0:
 */
      if (unlikely(__pyx_v_candidates == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 179, __pyx_L1_error)
      }
      __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 179, __pyx_L1_error)
      __pyx_t_5 = ((__pyx_t_7 < 1) != 0);
      if (__pyx_t_5) {

        /* "SpellChecker/SimpleSpellChecker.pyx":180
 *                     candidates.extend(self.candidateList(word, sentence))
 *                 if len(candidates) < 1:
 *                     candidates.extend(self.splitCandidatesList(word))             # <<<<<<<<<<<<<<
 *                 if len(candidates) > 0:
 *                     random_candidate = randrange(len(candidates))
 */
        if (unlikely(__pyx_v_candidates == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
          __PYX_ERR(0, 180, __pyx_L1_error)
        }
        __pyx_t_2 = ((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->splitCandidatesList(__pyx_v_self, __pyx_v_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = __Pyx_PyList_Extend(__pyx_v_candidates, __pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 180, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":179
 *                 if len(candidates) < 1:
 *                     candidates.extend(self.candidateList(word, sentence))
 *                 if len(candidates) < 1:             # <<<<<<<<<<<<<<
 *                     candidates.extend(self.splitCandidatesList(word))
 *                 if len(candidates) > 0:
 */
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":181
 *                 if len(candidates) < 1:
 *                     candidates.extend(self.splitCandidatesList(word))
 *                 if len(candidates) > 0:             # <<<<<<<<<<<<<<
 *                     random_candidate = randrange(len(candidates))
 *                     new_word = Word(candidates[random_candidate])
 */
      if (unlikely(__pyx_v_candidates == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 181, __pyx_L1_error)
      }
      __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 181, __pyx_L1_error)
      __pyx_t_5 = ((__pyx_t_7 > 0) != 0);
      if (__pyx_t_5) {

        /* "SpellChecker/SimpleSpellChecker.pyx":182
 *                     candidates.extend(self.splitCandidatesList(word))
 *                 if len(candidates) > 0:
 *                     random_candidate = randrange(len(candidates))             # <<<<<<<<<<<<<<
 *                     new_word = Word(candidates[random_candidate])
 *                     if candidates[random_candidate].getOperator() == Operator.BACKWARD_MERGE:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_randrange); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__pyx_v_candidates == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(0, 182, __pyx_L1_error)
        }
        __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 182, __pyx_L1_error)
        __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_random_candidate = __pyx_t_9;

        /* "SpellChecker/SimpleSpellChecker.pyx":183
 *                 if len(candidates) > 0:
 *                     random_candidate = randrange(len(candidates))
 *                     new_word = Word(candidates[random_candidate])             # <<<<<<<<<<<<<<
 *                     if candidates[random_candidate].getOperator() == Operator.BACKWARD_MERGE:
 *                         result.replaceWord(i - 1, new_word)
 */
        if (unlikely(__pyx_v_candidates == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 183, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_candidates, __pyx_v_random_candidate, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 183, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_new_word, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":184
 *                     random_candidate = randrange(len(candidates))
 *                     new_word = Word(candidates[random_candidate])
 *                     if candidates[random_candidate].getOperator() == Operator.BACKWARD_MERGE:             # <<<<<<<<<<<<<<
 *                         result.replaceWord(i - 1, new_word)
 *                         i = i + 1
 */
        if (unlikely(__pyx_v_candidates == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 184, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_candidates, __pyx_v_random_candidate, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_getOperator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Operator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_BACKWARD_MERGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 184, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 184, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_5) {

          /* "SpellChecker/SimpleSpellChecker.pyx":185
 *                     new_word = Word(candidates[random_candidate])
 *                     if candidates[random_candidate].getOperator() == Operator.BACKWARD_MERGE:
 *                         result.replaceWord(i - 1, new_word)             # <<<<<<<<<<<<<<
 *                         i = i + 1
 *                         continue
 */
          __pyx_t_3 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->replaceWord(__pyx_v_result, (__pyx_v_i - 1), __pyx_v_new_word, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "SpellChecker/SimpleSpellChecker.pyx":186
 *                     if candidates[random_candidate].getOperator() == Operator.BACKWARD_MERGE:
 *                         result.replaceWord(i - 1, new_word)
 *                         i = i + 1             # <<<<<<<<<<<<<<
 *                         continue
 *                     if candidates[random_candidate].getOperator() == Operator.FORWARD_MERGE:
 */
          __pyx_v_i = (__pyx_v_i + 1);

          /* "SpellChecker/SimpleSpellChecker.pyx":187
 *                         result.replaceWord(i - 1, new_word)
 *                         i = i + 1
 *                         continue             # <<<<<<<<<<<<<<
 *                     if candidates[random_candidate].getOperator() == Operator.FORWARD_MERGE:
 *                         i = i + 1
 */
          goto __pyx_L3_continue;

          /* "SpellChecker/SimpleSpellChecker.pyx":184
 *                     random_candidate = randrange(len(candidates))
 *                     new_word = Word(candidates[random_candidate])
 *                     if candidates[random_candidate].getOperator() == Operator.BACKWARD_MERGE:             # <<<<<<<<<<<<<<
 *                         result.replaceWord(i - 1, new_word)
 *                         i = i + 1
 */
        }

        /* "SpellChecker/SimpleSpellChecker.pyx":188
 *                         i = i + 1
 *                         continue
 *                     if candidates[random_candidate].getOperator() == Operator.FORWARD_MERGE:             # <<<<<<<<<<<<<<
 *                         i = i + 1
 *                     if candidates[random_candidate].getOperator() == Operator.SPLIT:
 */
        if (unlikely(__pyx_v_candidates == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 188, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_candidates, __pyx_v_random_candidate, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_getOperator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Operator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_FORWARD_MERGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_5) {

          /* "SpellChecker/SimpleSpellChecker.pyx":189
 *                         continue
 *                     if candidates[random_candidate].getOperator() == Operator.FORWARD_MERGE:
 *                         i = i + 1             # <<<<<<<<<<<<<<
 *                     if candidates[random_candidate].getOperator() == Operator.SPLIT:
 *                         self.addSplitWords(candidates[random_candidate].getName(), result)
 */
          __pyx_v_i = (__pyx_v_i + 1);

          /* "SpellChecker/SimpleSpellChecker.pyx":188
 *                         i = i + 1
 *                         continue
 *                     if candidates[random_candidate].getOperator() == Operator.FORWARD_MERGE:             # <<<<<<<<<<<<<<
 *                         i = i + 1
 *                     if candidates[random_candidate].getOperator() == Operator.SPLIT:
 */
        }

        /* "SpellChecker/SimpleSpellChecker.pyx":190
 *                     if candidates[random_candidate].getOperator() == Operator.FORWARD_MERGE:
 *                         i = i + 1
 *                     if candidates[random_candidate].getOperator() == Operator.SPLIT:             # <<<<<<<<<<<<<<
 *                         self.addSplitWords(candidates[random_candidate].getName(), result)
 *                         i = i + 1
 */
        if (unlikely(__pyx_v_candidates == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 190, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_candidates, __pyx_v_random_candidate, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_getOperator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Operator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SPLIT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 190, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 190, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_5) {

          /* "SpellChecker/SimpleSpellChecker.pyx":191
 *                         i = i + 1
 *                     if candidates[random_candidate].getOperator() == Operator.SPLIT:
 *                         self.addSplitWords(candidates[random_candidate].getName(), result)             # <<<<<<<<<<<<<<
 *                         i = i + 1
 *                         continue
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_addSplitWords); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (unlikely(__pyx_v_candidates == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 191, __pyx_L1_error)
          }
          __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_v_candidates, __pyx_v_random_candidate, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 191, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_getName); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 191, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
            }
          }
          __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_1, ((PyObject *)__pyx_v_result)};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_1, ((PyObject *)__pyx_v_result)};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 191, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_10) {
              __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_10); __pyx_t_10 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_9, __pyx_t_1);
            __Pyx_INCREF(((PyObject *)__pyx_v_result));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_9, ((PyObject *)__pyx_v_result));
            __pyx_t_1 = 0;
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "SpellChecker/SimpleSpellChecker.pyx":192
 *                     if candidates[random_candidate].getOperator() == Operator.SPLIT:
 *                         self.addSplitWords(candidates[random_candidate].getName(), result)
 *                         i = i + 1             # <<<<<<<<<<<<<<
 *                         continue
 *                 else:
 */
          __pyx_v_i = (__pyx_v_i + 1);

          /* "SpellChecker/SimpleSpellChecker.pyx":193
 *                         self.addSplitWords(candidates[random_candidate].getName(), result)
 *                         i = i + 1
 *                         continue             # <<<<<<<<<<<<<<
 *                 else:
 *                     new_word = word
 */
          goto __pyx_L3_continue;

          /* "SpellChecker/SimpleSpellChecker.pyx":190
 *                     if candidates[random_candidate].getOperator() == Operator.FORWARD_MERGE:
 *                         i = i + 1
 *                     if candidates[random_candidate].getOperator() == Operator.SPLIT:             # <<<<<<<<<<<<<<
 *                         self.addSplitWords(candidates[random_candidate].getName(), result)
 *                         i = i + 1
 */
        }

        /* "SpellChecker/SimpleSpellChecker.pyx":181
 *                 if len(candidates) < 1:
 *                     candidates.extend(self.splitCandidatesList(word))
 *                 if len(candidates) > 0:             # <<<<<<<<<<<<<<
 *                     random_candidate = randrange(len(candidates))
 *                     new_word = Word(candidates[random_candidate])
 */
        goto __pyx_L24;
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":195
 *                         continue
 *                 else:
 *                     new_word = word             # <<<<<<<<<<<<<<
 *             else:
 *                 new_word = word
 */
      /*else*/ {
        __Pyx_INCREF(((PyObject *)__pyx_v_word));
        __Pyx_XDECREF_SET(__pyx_v_new_word, __pyx_v_word);
      }
      __pyx_L24:;

      /* "SpellChecker/SimpleSpellChecker.pyx":175
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(word.getName())
 *             upper_case_fsm_parse_list = self.fsm.morphologicalAnalysis(Word.toCapital(word.getName()))
 *             if fsm_parse_list.size() == 0 and upper_case_fsm_parse_list.size() == 0:             # <<<<<<<<<<<<<<
 *                 candidates = self.mergedCandidatesList(previous_word, word, next_word)
 *                 if len(candidates) < 1:
 */
      goto __pyx_L19;
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":197
 *                     new_word = word
 *             else:
 *                 new_word = word             # <<<<<<<<<<<<<<
 *             result.addWord(new_word)
 *             i = i + 1
 */
    /*else*/ {
      __Pyx_INCREF(((PyObject *)__pyx_v_word));
      __Pyx_XDECREF_SET(__pyx_v_new_word, __pyx_v_word);
    }
    __pyx_L19:;

    /* "SpellChecker/SimpleSpellChecker.pyx":198
 *             else:
 *                 new_word = word
 *             result.addWord(new_word)             # <<<<<<<<<<<<<<
 *             i = i + 1
 *         return result
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, __pyx_v_new_word, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":199
 *                 new_word = word
 *             result.addWord(new_word)
 *             i = i + 1             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_v_i = (__pyx_v_i + 1);
    __pyx_L3_continue:;
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":200
 *             result.addWord(new_word)
 *             i = i + 1
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint forcedMisspellCheck(self,
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":126
 *         return candidates
 * 
 *     cpdef Sentence spellCheck(self, Sentence sentence):             # <<<<<<<<<<<<<<
 *         """
 *         The spellCheck method takes a Sentence as an input and loops i times where i ranges from 0 to size of words in
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.spellCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_word);
  __Pyx_XDECREF((PyObject *)__pyx_v_new_word);
  __Pyx_XDECREF((PyObject *)__pyx_v_next_word);
  __Pyx_XDECREF((PyObject *)__pyx_v_previous_word);
  __Pyx_XDECREF((PyObject *)__pyx_v_fsm_parse_list);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_upper_case_fsm_parse_list);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_7spellCheck(PyObject *__pyx_v_self, PyObject *__pyx_v_sentence); /*proto*/
static char __pyx_doc_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_6spellCheck[] = "\n        The spellCheck method takes a Sentence as an input and loops i times where i ranges from 0 to size of words in\n        given sentence. Then, it calls morphologicalAnalysis method with each word and assigns it to the FsmParseList,\n        if the size of FsmParseList is equal to the 0, it adds current word to the candidateList and assigns it to the\n        candidates list. If the size of candidates greater than 0, it generates a random number and selects an item from\n        candidates list with this random number and assign it as newWord. If the size of candidates is not greater than\n        0, it directly assigns the current word as newWord. At the end, it adds the newWord to the result Sentence.\n\n        PARAMETERS\n        ----------\n        sentence : Sentence\n            Sentence type input.\n\n        RETURNS\n        -------\n        Sentence\n            Sentence result.\n        ";
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_7spellCheck(PyObject *__pyx_v_self, PyObject *__pyx_v_sentence) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spellCheck (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sentence), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "sentence", 0))) __PYX_ERR(0, 126, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_6spellCheck(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)__pyx_v_sentence));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_6spellCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_sentence) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("spellCheck", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_spellCheck(__pyx_v_self, __pyx_v_sentence, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.spellCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":202
 *         return result
 * 
 *     cpdef bint forcedMisspellCheck(self,             # <<<<<<<<<<<<<<
 *                                    Word word,
 *                                    Sentence result):
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_9forcedMisspellCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedMisspellCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_forced_replacement = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedMisspellCheck", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_forcedMisspellCheck); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_9forcedMisspellCheck)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 202, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_result));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 202, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":206
 *                                    Sentence result):
 *         cdef str forced_replacement
 *         forced_replacement = self.fsm.getDictionary().getCorrectForm(word.getName())             # <<<<<<<<<<<<<<
 *         if forced_replacement != "":
 *             result.addWord(Word(forced_replacement))
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->getDictionary(__pyx_v_self->fsm, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_10Dictionary_13TxtDictionary_TxtDictionary *)((struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *)__pyx_t_1)->__pyx_base.__pyx_vtab)->getCorrectForm(((struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *)__pyx_t_1), ((PyObject*)__pyx_t_2), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_forced_replacement = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":207
 *         cdef str forced_replacement
 *         forced_replacement = self.fsm.getDictionary().getCorrectForm(word.getName())
 *         if forced_replacement != "":             # <<<<<<<<<<<<<<
 *             result.addWord(Word(forced_replacement))
 *             return True
 */
  __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_forced_replacement, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 207, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "SpellChecker/SimpleSpellChecker.pyx":208
 *         forced_replacement = self.fsm.getDictionary().getCorrectForm(word.getName())
 *         if forced_replacement != "":
 *             result.addWord(Word(forced_replacement))             # <<<<<<<<<<<<<<
 *             return True
 *         return False
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_v_forced_replacement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_3), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":209
 *         if forced_replacement != "":
 *             result.addWord(Word(forced_replacement))
 *             return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "SpellChecker/SimpleSpellChecker.pyx":207
 *         cdef str forced_replacement
 *         forced_replacement = self.fsm.getDictionary().getCorrectForm(word.getName())
 *         if forced_replacement != "":             # <<<<<<<<<<<<<<
 *             result.addWord(Word(forced_replacement))
 *             return True
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":210
 *             result.addWord(Word(forced_replacement))
 *             return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint forcedBackwardMergeCheck(self,
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":202
 *         return result
 * 
 *     cpdef bint forcedMisspellCheck(self,             # <<<<<<<<<<<<<<
 *                                    Word word,
 *                                    Sentence result):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedMisspellCheck", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_forced_replacement);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_9forcedMisspellCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_9forcedMisspellCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("forcedMisspellCheck (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word,&__pyx_n_s_result,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedMisspellCheck", 1, 2, 2, 1); __PYX_ERR(0, 202, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "forcedMisspellCheck") < 0)) __PYX_ERR(0, 202, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("forcedMisspellCheck", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 202, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedMisspellCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 203, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "result", 0))) __PYX_ERR(0, 204, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_8forcedMisspellCheck(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_word, __pyx_v_result);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_8forcedMisspellCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedMisspellCheck", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedMisspellCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedMisspellCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":212
 *         return False
 * 
 *     cpdef bint forcedBackwardMergeCheck(self,             # <<<<<<<<<<<<<<
 *                                  Word word,
 *                                  Sentence result,
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_11forcedBackwardMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedBackwardMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_forced_replacement = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedBackwardMergeCheck", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_forcedBackwardMergeCheck); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_11forcedBackwardMergeCheck)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result), ((PyObject *)__pyx_v_previousWord)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result), ((PyObject *)__pyx_v_previousWord)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 212, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_result));
          __Pyx_INCREF(((PyObject *)__pyx_v_previousWord));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_previousWord));
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, ((PyObject *)__pyx_v_previousWord));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 212, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":217
 *                                  Word previousWord):
 *         cdef str forced_replacement
 *         if previousWord is not None:             # <<<<<<<<<<<<<<
 *             forced_replacement = self.getCorrectForm(
 *                 result.getWord(result.wordCount() - 1).getName() + " " + word.getName(),
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_previousWord) != Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "SpellChecker/SimpleSpellChecker.pyx":219
 *         if previousWord is not None:
 *             forced_replacement = self.getCorrectForm(
 *                 result.getWord(result.wordCount() - 1).getName() + " " + word.getName(),             # <<<<<<<<<<<<<<
 *                 self.__merged_words)
 *             if forced_replacement != "":
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->getWord(__pyx_v_result, (((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->wordCount(__pyx_v_result, 0) - 1), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_1)->__pyx_vtab)->getName(((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_1), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_2, __pyx_kp_u__2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":220
 *             forced_replacement = self.getCorrectForm(
 *                 result.getWord(result.wordCount() - 1).getName() + " " + word.getName(),
 *                 self.__merged_words)             # <<<<<<<<<<<<<<
 *             if forced_replacement != "":
 *                 result.replaceWord(result.wordCount() - 1, Word(forced_replacement))
 */
    __pyx_t_2 = __pyx_v_self->__pyx___merged_words;
    __Pyx_INCREF(__pyx_t_2);

    /* "SpellChecker/SimpleSpellChecker.pyx":218
 *         cdef str forced_replacement
 *         if previousWord is not None:
 *             forced_replacement = self.getCorrectForm(             # <<<<<<<<<<<<<<
 *                 result.getWord(result.wordCount() - 1).getName() + " " + word.getName(),
 *                 self.__merged_words)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->getCorrectForm(__pyx_v_self, ((PyObject*)__pyx_t_3), ((PyObject*)__pyx_t_2), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_forced_replacement = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":221
 *                 result.getWord(result.wordCount() - 1).getName() + " " + word.getName(),
 *                 self.__merged_words)
 *             if forced_replacement != "":             # <<<<<<<<<<<<<<
 *                 result.replaceWord(result.wordCount() - 1, Word(forced_replacement))
 *                 return True
 */
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_forced_replacement, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 221, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_8 != 0);
    if (__pyx_t_7) {

      /* "SpellChecker/SimpleSpellChecker.pyx":222
 *                 self.__merged_words)
 *             if forced_replacement != "":
 *                 result.replaceWord(result.wordCount() - 1, Word(forced_replacement))             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_v_forced_replacement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->replaceWord(__pyx_v_result, (((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->wordCount(__pyx_v_result, 0) - 1), ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_1), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":223
 *             if forced_replacement != "":
 *                 result.replaceWord(result.wordCount() - 1, Word(forced_replacement))
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "SpellChecker/SimpleSpellChecker.pyx":221
 *                 result.getWord(result.wordCount() - 1).getName() + " " + word.getName(),
 *                 self.__merged_words)
 *             if forced_replacement != "":             # <<<<<<<<<<<<<<
 *                 result.replaceWord(result.wordCount() - 1, Word(forced_replacement))
 *                 return True
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":217
 *                                  Word previousWord):
 *         cdef str forced_replacement
 *         if previousWord is not None:             # <<<<<<<<<<<<<<
 *             forced_replacement = self.getCorrectForm(
 *                 result.getWord(result.wordCount() - 1).getName() + " " + word.getName(),
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":224
 *                 result.replaceWord(result.wordCount() - 1, Word(forced_replacement))
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint forcedForwardMergeCheck(self,
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":212
 *         return False
 * 
 *     cpdef bint forcedBackwardMergeCheck(self,             # <<<<<<<<<<<<<<
 *                                  Word word,
 *                                  Sentence result,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedBackwardMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_forced_replacement);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_11forcedBackwardMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_11forcedBackwardMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("forcedBackwardMergeCheck (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word,&__pyx_n_s_result,&__pyx_n_s_previousWord,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedBackwardMergeCheck", 1, 3, 3, 1); __PYX_ERR(0, 212, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_previousWord)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedBackwardMergeCheck", 1, 3, 3, 2); __PYX_ERR(0, 212, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "forcedBackwardMergeCheck") < 0)) __PYX_ERR(0, 212, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
    __pyx_v_previousWord = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("forcedBackwardMergeCheck", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 212, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedBackwardMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 213, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "result", 0))) __PYX_ERR(0, 214, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_previousWord), __pyx_ptype_10Dictionary_4Word_Word, 1, "previousWord", 0))) __PYX_ERR(0, 215, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_10forcedBackwardMergeCheck(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_word, __pyx_v_result, __pyx_v_previousWord);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_10forcedBackwardMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedBackwardMergeCheck", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedBackwardMergeCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, __pyx_v_previousWord, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedBackwardMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":226
 *         return False
 * 
 *     cpdef bint forcedForwardMergeCheck(self,             # <<<<<<<<<<<<<<
 *                                 Word word,
 *                                 Sentence result,
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_13forcedForwardMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedForwardMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_forced_replacement = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedForwardMergeCheck", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_forcedForwardMergeCheck); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_13forcedForwardMergeCheck)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result), ((PyObject *)__pyx_v_nextWord)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result), ((PyObject *)__pyx_v_nextWord)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 226, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_result));
          __Pyx_INCREF(((PyObject *)__pyx_v_nextWord));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_nextWord));
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, ((PyObject *)__pyx_v_nextWord));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 226, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":231
 *                                 Word nextWord):
 *         cdef str forced_replacement
 *         if nextWord is not None:             # <<<<<<<<<<<<<<
 *             forced_replacement = self.getCorrectForm(word.getName() + " " + nextWord.getName(),
 *                                                      self.__merged_words)
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_nextWord) != Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "SpellChecker/SimpleSpellChecker.pyx":232
 *         cdef str forced_replacement
 *         if nextWord is not None:
 *             forced_replacement = self.getCorrectForm(word.getName() + " " + nextWord.getName(),             # <<<<<<<<<<<<<<
 *                                                      self.__merged_words)
 *             if forced_replacement != "":
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_1, __pyx_kp_u__2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_nextWord->__pyx_vtab)->getName(__pyx_v_nextWord, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":233
 *         if nextWord is not None:
 *             forced_replacement = self.getCorrectForm(word.getName() + " " + nextWord.getName(),
 *                                                      self.__merged_words)             # <<<<<<<<<<<<<<
 *             if forced_replacement != "":
 *                 result.addWord(Word(forced_replacement))
 */
    __pyx_t_1 = __pyx_v_self->__pyx___merged_words;
    __Pyx_INCREF(__pyx_t_1);

    /* "SpellChecker/SimpleSpellChecker.pyx":232
 *         cdef str forced_replacement
 *         if nextWord is not None:
 *             forced_replacement = self.getCorrectForm(word.getName() + " " + nextWord.getName(),             # <<<<<<<<<<<<<<
 *                                                      self.__merged_words)
 *             if forced_replacement != "":
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->getCorrectForm(__pyx_v_self, ((PyObject*)__pyx_t_3), ((PyObject*)__pyx_t_1), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_forced_replacement = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":234
 *             forced_replacement = self.getCorrectForm(word.getName() + " " + nextWord.getName(),
 *                                                      self.__merged_words)
 *             if forced_replacement != "":             # <<<<<<<<<<<<<<
 *                 result.addWord(Word(forced_replacement))
 *                 return True
 */
    __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_forced_replacement, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 234, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_8 != 0);
    if (__pyx_t_7) {

      /* "SpellChecker/SimpleSpellChecker.pyx":235
 *                                                      self.__merged_words)
 *             if forced_replacement != "":
 *                 result.addWord(Word(forced_replacement))             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_v_forced_replacement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_2), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":236
 *             if forced_replacement != "":
 *                 result.addWord(Word(forced_replacement))
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "SpellChecker/SimpleSpellChecker.pyx":234
 *             forced_replacement = self.getCorrectForm(word.getName() + " " + nextWord.getName(),
 *                                                      self.__merged_words)
 *             if forced_replacement != "":             # <<<<<<<<<<<<<<
 *                 result.addWord(Word(forced_replacement))
 *                 return True
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":231
 *                                 Word nextWord):
 *         cdef str forced_replacement
 *         if nextWord is not None:             # <<<<<<<<<<<<<<
 *             forced_replacement = self.getCorrectForm(word.getName() + " " + nextWord.getName(),
 *                                                      self.__merged_words)
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":237
 *                 result.addWord(Word(forced_replacement))
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def addSplitWords(self,
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":226
 *         return False
 * 
 *     cpdef bint forcedForwardMergeCheck(self,             # <<<<<<<<<<<<<<
 *                                 Word word,
 *                                 Sentence result,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedForwardMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_forced_replacement);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_13forcedForwardMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_13forcedForwardMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("forcedForwardMergeCheck (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word,&__pyx_n_s_result,&__pyx_n_s_nextWord,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedForwardMergeCheck", 1, 3, 3, 1); __PYX_ERR(0, 226, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nextWord)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedForwardMergeCheck", 1, 3, 3, 2); __PYX_ERR(0, 226, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "forcedForwardMergeCheck") < 0)) __PYX_ERR(0, 226, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
    __pyx_v_nextWord = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("forcedForwardMergeCheck", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 226, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedForwardMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 227, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "result", 0))) __PYX_ERR(0, 228, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nextWord), __pyx_ptype_10Dictionary_4Word_Word, 1, "nextWord", 0))) __PYX_ERR(0, 229, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_12forcedForwardMergeCheck(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_word, __pyx_v_result, __pyx_v_nextWord);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_12forcedForwardMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedForwardMergeCheck", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedForwardMergeCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, __pyx_v_nextWord, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedForwardMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":239
 *         return False
 * 
 *     def addSplitWords(self,             # <<<<<<<<<<<<<<
 *                       multiWord: str,
 *                       result: Sentence):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_15addSplitWords(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_15addSplitWords(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_multiWord = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("addSplitWords (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_multiWord,&__pyx_n_s_result,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_multiWord)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("addSplitWords", 1, 2, 2, 1); __PYX_ERR(0, 239, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "addSplitWords") < 0)) __PYX_ERR(0, 239, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_multiWord = ((PyObject*)values[0]);
    __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("addSplitWords", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 239, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.addSplitWords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_multiWord), (&PyUnicode_Type), 1, "multiWord", 1))) __PYX_ERR(0, 240, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "result", 0))) __PYX_ERR(0, 241, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_14addSplitWords(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_multiWord, __pyx_v_result);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_14addSplitWords(CYTHON_UNUSED struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v_multiWord, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result) {
  PyObject *__pyx_v_words = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("addSplitWords", 0);

  /* "SpellChecker/SimpleSpellChecker.pyx":242
 *                       multiWord: str,
 *                       result: Sentence):
 *         words = multiWord.split(" ")             # <<<<<<<<<<<<<<
 *         result.addWord(Word(words[0]))
 *         result.addWord(Word(words[1]))
 */
  if (unlikely(__pyx_v_multiWord == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "split");
    __PYX_ERR(0, 242, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_Split(__pyx_v_multiWord, __pyx_kp_u__2, -1L); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_words = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":243
 *                       result: Sentence):
 *         words = multiWord.split(" ")
 *         result.addWord(Word(words[0]))             # <<<<<<<<<<<<<<
 *         result.addWord(Word(words[1]))
 * 
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_words, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_2), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":244
 *         words = multiWord.split(" ")
 *         result.addWord(Word(words[0]))
 *         result.addWord(Word(words[1]))             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint forcedSplitCheck(self,
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_words, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_2), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":239
 *         return False
 * 
 *     def addSplitWords(self,             # <<<<<<<<<<<<<<
 *                       multiWord: str,
 *                       result: Sentence):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.addSplitWords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_words);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":246
 *         result.addWord(Word(words[1]))
 * 
 *     cpdef bint forcedSplitCheck(self,             # <<<<<<<<<<<<<<
 *                                 Word word,
 *                                 Sentence result):
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_17forcedSplitCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_forced_replacement = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedSplitCheck", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_forcedSplitCheck); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_17forcedSplitCheck)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 246, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_result));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":250
 *                                 Sentence result):
 *         cdef str forced_replacement
 *         forced_replacement = self.getCorrectForm(word.getName(), self.__split_words)             # <<<<<<<<<<<<<<
 *         if forced_replacement != "":
 *             self.addSplitWords(forced_replacement, result)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_self->__pyx___split_words;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->getCorrectForm(__pyx_v_self, ((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_forced_replacement = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":251
 *         cdef str forced_replacement
 *         forced_replacement = self.getCorrectForm(word.getName(), self.__split_words)
 *         if forced_replacement != "":             # <<<<<<<<<<<<<<
 *             self.addSplitWords(forced_replacement, result)
 *             return True
 */
  __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_forced_replacement, __pyx_kp_u_, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 251, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "SpellChecker/SimpleSpellChecker.pyx":252
 *         forced_replacement = self.getCorrectForm(word.getName(), self.__split_words)
 *         if forced_replacement != "":
 *             self.addSplitWords(forced_replacement, result)             # <<<<<<<<<<<<<<
 *             return True
 *         return False
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_addSplitWords); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_forced_replacement, ((PyObject *)__pyx_v_result)};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_forced_replacement, ((PyObject *)__pyx_v_result)};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_forced_replacement);
      __Pyx_GIVEREF(__pyx_v_forced_replacement);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_forced_replacement);
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_result));
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":253
 *         if forced_replacement != "":
 *             self.addSplitWords(forced_replacement, result)
 *             return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "SpellChecker/SimpleSpellChecker.pyx":251
 *         cdef str forced_replacement
 *         forced_replacement = self.getCorrectForm(word.getName(), self.__split_words)
 *         if forced_replacement != "":             # <<<<<<<<<<<<<<
 *             self.addSplitWords(forced_replacement, result)
 *             return True
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":254
 *             self.addSplitWords(forced_replacement, result)
 *             return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint forcedShortcutCheck(self,
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":246
 *         result.addWord(Word(words[1]))
 * 
 *     cpdef bint forcedSplitCheck(self,             # <<<<<<<<<<<<<<
 *                                 Word word,
 *                                 Sentence result):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedSplitCheck", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_forced_replacement);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_17forcedSplitCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_17forcedSplitCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("forcedSplitCheck (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word,&__pyx_n_s_result,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedSplitCheck", 1, 2, 2, 1); __PYX_ERR(0, 246, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "forcedSplitCheck") < 0)) __PYX_ERR(0, 246, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("forcedSplitCheck", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 246, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedSplitCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 247, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "result", 0))) __PYX_ERR(0, 248, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_16forcedSplitCheck(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_word, __pyx_v_result);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_16forcedSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedSplitCheck", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedSplitCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedSplitCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":256
 *         return False
 * 
 *     cpdef bint forcedShortcutCheck(self,             # <<<<<<<<<<<<<<
 *                                    Word word,
 *                                    Sentence result):
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_19forcedShortcutCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedShortcutCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_shortcut_regex = 0;
  PyObject *__pyx_v_conditional_shortcut_regex = 0;
  PyObject *__pyx_v_forced_replacement = 0;
  int __pyx_v_i;
  PyObject *__pyx_v_pair = 0;
  PyObject *__pyx_v_compiled_expression1 = NULL;
  PyObject *__pyx_v_compiled_expression2 = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedShortcutCheck", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_forcedShortcutCheck); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_19forcedShortcutCheck)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 256, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 256, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 256, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_result));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 256, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 256, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":262
 *         cdef int i
 *         cdef tuple pair
 *         shortcut_regex = "(([1-9][0-9]*)|[0])(([.]|[,])[0-9]*)?(" + self.__shortcuts[0]             # <<<<<<<<<<<<<<
 *         for i in range(1, len(self.__shortcuts)):
 *             shortcut_regex = shortcut_regex + "|" + self.__shortcuts[i]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shortcuts); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_kp_u_1_9_0_9_0_0_9, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 262, __pyx_L1_error)
  __pyx_v_shortcut_regex = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":263
 *         cdef tuple pair
 *         shortcut_regex = "(([1-9][0-9]*)|[0])(([.]|[,])[0-9]*)?(" + self.__shortcuts[0]
 *         for i in range(1, len(self.__shortcuts)):             # <<<<<<<<<<<<<<
 *             shortcut_regex = shortcut_regex + "|" + self.__shortcuts[i]
 *         shortcut_regex = shortcut_regex + ")"
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shortcuts); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_5 = 1; __pyx_t_5 < __pyx_t_9; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "SpellChecker/SimpleSpellChecker.pyx":264
 *         shortcut_regex = "(([1-9][0-9]*)|[0])(([.]|[,])[0-9]*)?(" + self.__shortcuts[0]
 *         for i in range(1, len(self.__shortcuts)):
 *             shortcut_regex = shortcut_regex + "|" + self.__shortcuts[i]             # <<<<<<<<<<<<<<
 *         shortcut_regex = shortcut_regex + ")"
 *         conditional_shortcut_regex = "(([1-9][0-9]{0,2})|[0])(([.]|[,])[0-9]*)?(" + self.__conditionalShortcuts[0]
 */
    __pyx_t_1 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_shortcut_regex, __pyx_kp_u__3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_shortcuts); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_shortcut_regex, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":265
 *         for i in range(1, len(self.__shortcuts)):
 *             shortcut_regex = shortcut_regex + "|" + self.__shortcuts[i]
 *         shortcut_regex = shortcut_regex + ")"             # <<<<<<<<<<<<<<
 *         conditional_shortcut_regex = "(([1-9][0-9]{0,2})|[0])(([.]|[,])[0-9]*)?(" + self.__conditionalShortcuts[0]
 *         for i in range(1, len(self.__conditionalShortcuts)):
 */
  __pyx_t_2 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_shortcut_regex, __pyx_kp_u__4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_shortcut_regex, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":266
 *             shortcut_regex = shortcut_regex + "|" + self.__shortcuts[i]
 *         shortcut_regex = shortcut_regex + ")"
 *         conditional_shortcut_regex = "(([1-9][0-9]{0,2})|[0])(([.]|[,])[0-9]*)?(" + self.__conditionalShortcuts[0]             # <<<<<<<<<<<<<<
 *         for i in range(1, len(self.__conditionalShortcuts)):
 *             conditional_shortcut_regex = conditional_shortcut_regex + "|" + self.__conditionalShortcuts[i]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_conditionalShortcuts); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_kp_u_1_9_0_9_0_2_0_0_9, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 266, __pyx_L1_error)
  __pyx_v_conditional_shortcut_regex = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":267
 *         shortcut_regex = shortcut_regex + ")"
 *         conditional_shortcut_regex = "(([1-9][0-9]{0,2})|[0])(([.]|[,])[0-9]*)?(" + self.__conditionalShortcuts[0]
 *         for i in range(1, len(self.__conditionalShortcuts)):             # <<<<<<<<<<<<<<
 *             conditional_shortcut_regex = conditional_shortcut_regex + "|" + self.__conditionalShortcuts[i]
 *         conditional_shortcut_regex = conditional_shortcut_regex + ")"
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_conditionalShortcuts); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_5 = 1; __pyx_t_5 < __pyx_t_9; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "SpellChecker/SimpleSpellChecker.pyx":268
 *         conditional_shortcut_regex = "(([1-9][0-9]{0,2})|[0])(([.]|[,])[0-9]*)?(" + self.__conditionalShortcuts[0]
 *         for i in range(1, len(self.__conditionalShortcuts)):
 *             conditional_shortcut_regex = conditional_shortcut_regex + "|" + self.__conditionalShortcuts[i]             # <<<<<<<<<<<<<<
 *         conditional_shortcut_regex = conditional_shortcut_regex + ")"
 *         compiled_expression1 = re.compile(shortcut_regex)
 */
    __pyx_t_2 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_conditional_shortcut_regex, __pyx_kp_u__3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_conditionalShortcuts); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_conditional_shortcut_regex, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":269
 *         for i in range(1, len(self.__conditionalShortcuts)):
 *             conditional_shortcut_regex = conditional_shortcut_regex + "|" + self.__conditionalShortcuts[i]
 *         conditional_shortcut_regex = conditional_shortcut_regex + ")"             # <<<<<<<<<<<<<<
 *         compiled_expression1 = re.compile(shortcut_regex)
 *         compiled_expression2 = re.compile(conditional_shortcut_regex)
 */
  __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_v_conditional_shortcut_regex, __pyx_kp_u__4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF_SET(__pyx_v_conditional_shortcut_regex, ((PyObject*)__pyx_t_3));
  __pyx_t_3 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":270
 *             conditional_shortcut_regex = conditional_shortcut_regex + "|" + self.__conditionalShortcuts[i]
 *         conditional_shortcut_regex = conditional_shortcut_regex + ")"
 *         compiled_expression1 = re.compile(shortcut_regex)             # <<<<<<<<<<<<<<
 *         compiled_expression2 = re.compile(conditional_shortcut_regex)
 *         if compiled_expression1.fullmatch(word.getName()) or compiled_expression2.fullmatch(word.getName()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_re); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_compile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_shortcut_regex) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_shortcut_regex);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_compiled_expression1 = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":271
 *         conditional_shortcut_regex = conditional_shortcut_regex + ")"
 *         compiled_expression1 = re.compile(shortcut_regex)
 *         compiled_expression2 = re.compile(conditional_shortcut_regex)             # <<<<<<<<<<<<<<
 *         if compiled_expression1.fullmatch(word.getName()) or compiled_expression2.fullmatch(word.getName()):
 *             pair = self.getSplitPair(word)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_conditional_shortcut_regex) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_conditional_shortcut_regex);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_compiled_expression2 = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":272
 *         compiled_expression1 = re.compile(shortcut_regex)
 *         compiled_expression2 = re.compile(conditional_shortcut_regex)
 *         if compiled_expression1.fullmatch(word.getName()) or compiled_expression2.fullmatch(word.getName()):             # <<<<<<<<<<<<<<
 *             pair = self.getSplitPair(word)
 *             forced_replacement = pair[0] + " " + pair[1]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiled_expression1, __pyx_n_s_fullmatch); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_10) {
  } else {
    __pyx_t_7 = __pyx_t_10;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiled_expression2, __pyx_n_s_fullmatch); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __pyx_t_10;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_7) {

    /* "SpellChecker/SimpleSpellChecker.pyx":273
 *         compiled_expression2 = re.compile(conditional_shortcut_regex)
 *         if compiled_expression1.fullmatch(word.getName()) or compiled_expression2.fullmatch(word.getName()):
 *             pair = self.getSplitPair(word)             # <<<<<<<<<<<<<<
 *             forced_replacement = pair[0] + " " + pair[1]
 *             result.addWord(Word(forced_replacement))
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self->__pyx_base.__pyx_vtab)->getSplitPair(__pyx_v_self, __pyx_v_word, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_pair = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":274
 *         if compiled_expression1.fullmatch(word.getName()) or compiled_expression2.fullmatch(word.getName()):
 *             pair = self.getSplitPair(word)
 *             forced_replacement = pair[0] + " " + pair[1]             # <<<<<<<<<<<<<<
 *             result.addWord(Word(forced_replacement))
 *             return True
 */
    if (unlikely(__pyx_v_pair == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 274, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_pair, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_kp_u__2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_v_pair == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 274, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_pair, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 274, __pyx_L1_error)
    __pyx_v_forced_replacement = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":275
 *             pair = self.getSplitPair(word)
 *             forced_replacement = pair[0] + " " + pair[1]
 *             result.addWord(Word(forced_replacement))             # <<<<<<<<<<<<<<
 *             return True
 *         return False
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_v_forced_replacement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_2), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":276
 *             forced_replacement = pair[0] + " " + pair[1]
 *             result.addWord(Word(forced_replacement))
 *             return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "SpellChecker/SimpleSpellChecker.pyx":272
 *         compiled_expression1 = re.compile(shortcut_regex)
 *         compiled_expression2 = re.compile(conditional_shortcut_regex)
 *         if compiled_expression1.fullmatch(word.getName()) or compiled_expression2.fullmatch(word.getName()):             # <<<<<<<<<<<<<<
 *             pair = self.getSplitPair(word)
 *             forced_replacement = pair[0] + " " + pair[1]
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":277
 *             result.addWord(Word(forced_replacement))
 *             return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint forcedDeDaSplitCheck(self,
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":256
 *         return False
 * 
 *     cpdef bint forcedShortcutCheck(self,             # <<<<<<<<<<<<<<
 *                                    Word word,
 *                                    Sentence result):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedShortcutCheck", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shortcut_regex);
  __Pyx_XDECREF(__pyx_v_conditional_shortcut_regex);
  __Pyx_XDECREF(__pyx_v_forced_replacement);
  __Pyx_XDECREF(__pyx_v_pair);
  __Pyx_XDECREF(__pyx_v_compiled_expression1);
  __Pyx_XDECREF(__pyx_v_compiled_expression2);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_19forcedShortcutCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_19forcedShortcutCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("forcedShortcutCheck (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word,&__pyx_n_s_result,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedShortcutCheck", 1, 2, 2, 1); __PYX_ERR(0, 256, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "forcedShortcutCheck") < 0)) __PYX_ERR(0, 256, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("forcedShortcutCheck", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 256, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedShortcutCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 257, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "result", 0))) __PYX_ERR(0, 258, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_18forcedShortcutCheck(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_word, __pyx_v_result);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_18forcedShortcutCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedShortcutCheck", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedShortcutCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedShortcutCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":279
 *         return False
 * 
 *     cpdef bint forcedDeDaSplitCheck(self,             # <<<<<<<<<<<<<<
 *                              Word word,
 *                              Sentence result):
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_21forcedDeDaSplitCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedDeDaSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_word_name = 0;
  PyObject *__pyx_v_capitalized_word_name = 0;
  PyObject *__pyx_v_new_word_name = 0;
  struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *__pyx_v_txt_word = 0;
  struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *__pyx_v_txt_new_word = 0;
  struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *__pyx_v_fsm_parse_list = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedDeDaSplitCheck", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_forcedDeDaSplitCheck); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_21forcedDeDaSplitCheck)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 279, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_result));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":285
 *         cdef TxtWord txt_word, txt_new_word
 *         cdef FsmParseList fsm_parse_list
 *         word_name = word.getName()             # <<<<<<<<<<<<<<
 *         capitalized_word_name = Word.toCapital(word_name)
 *         txt_word = None
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_word_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":286
 *         cdef FsmParseList fsm_parse_list
 *         word_name = word.getName()
 *         capitalized_word_name = Word.toCapital(word_name)             # <<<<<<<<<<<<<<
 *         txt_word = None
 *         if word_name.endswith("da") or word_name.endswith("de"):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_n_s_toCapital); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_word_name) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_word_name);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 286, __pyx_L1_error)
  __pyx_v_capitalized_word_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":287
 *         word_name = word.getName()
 *         capitalized_word_name = Word.toCapital(word_name)
 *         txt_word = None             # <<<<<<<<<<<<<<
 *         if word_name.endswith("da") or word_name.endswith("de"):
 *             if self.fsm.morphologicalAnalysis(word_name).size() == 0 and \
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_txt_word = ((struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *)Py_None);

  /* "SpellChecker/SimpleSpellChecker.pyx":288
 *         capitalized_word_name = Word.toCapital(word_name)
 *         txt_word = None
 *         if word_name.endswith("da") or word_name.endswith("de"):             # <<<<<<<<<<<<<<
 *             if self.fsm.morphologicalAnalysis(word_name).size() == 0 and \
 *                     self.fsm.morphologicalAnalysis(capitalized_word_name).size() == 0:
 */
  if (unlikely(__pyx_v_word_name == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "endswith");
    __PYX_ERR(0, 288, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyUnicode_Tailmatch(__pyx_v_word_name, __pyx_n_u_da, 0, PY_SSIZE_T_MAX, 1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 288, __pyx_L1_error)
  if (!(__pyx_t_8 != 0)) {
  } else {
    __pyx_t_7 = (__pyx_t_8 != 0);
    goto __pyx_L4_bool_binop_done;
  }
  if (unlikely(__pyx_v_word_name == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "endswith");
    __PYX_ERR(0, 288, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyUnicode_Tailmatch(__pyx_v_word_name, __pyx_n_u_de, 0, PY_SSIZE_T_MAX, 1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 288, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_8 != 0);
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_7) {

    /* "SpellChecker/SimpleSpellChecker.pyx":289
 *         txt_word = None
 *         if word_name.endswith("da") or word_name.endswith("de"):
 *             if self.fsm.morphologicalAnalysis(word_name).size() == 0 and \             # <<<<<<<<<<<<<<
 *                     self.fsm.morphologicalAnalysis(capitalized_word_name).size() == 0:
 *                 new_word_name = word_name[0:len(word_name) - 2]
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_v_word_name, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L7_bool_binop_done;
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":290
 *         if word_name.endswith("da") or word_name.endswith("de"):
 *             if self.fsm.morphologicalAnalysis(word_name).size() == 0 and \
 *                     self.fsm.morphologicalAnalysis(capitalized_word_name).size() == 0:             # <<<<<<<<<<<<<<
 *                 new_word_name = word_name[0:len(word_name) - 2]
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(new_word_name)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_v_capitalized_word_name, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __pyx_t_8;
    __pyx_L7_bool_binop_done:;

    /* "SpellChecker/SimpleSpellChecker.pyx":289
 *         txt_word = None
 *         if word_name.endswith("da") or word_name.endswith("de"):
 *             if self.fsm.morphologicalAnalysis(word_name).size() == 0 and \             # <<<<<<<<<<<<<<
 *                     self.fsm.morphologicalAnalysis(capitalized_word_name).size() == 0:
 *                 new_word_name = word_name[0:len(word_name) - 2]
 */
    if (__pyx_t_7) {

      /* "SpellChecker/SimpleSpellChecker.pyx":291
 *             if self.fsm.morphologicalAnalysis(word_name).size() == 0 and \
 *                     self.fsm.morphologicalAnalysis(capitalized_word_name).size() == 0:
 *                 new_word_name = word_name[0:len(word_name) - 2]             # <<<<<<<<<<<<<<
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(new_word_name)
 *                 txt_new_word = self.fsm.getDictionary().getWord(Word.toLower(new_word_name))
 */
      if (unlikely(__pyx_v_word_name == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 291, __pyx_L1_error)
      }
      if (unlikely(__pyx_v_word_name == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 291, __pyx_L1_error)
      }
      __pyx_t_9 = __Pyx_PyUnicode_GET_LENGTH(__pyx_v_word_name); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 291, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyUnicode_Substring(__pyx_v_word_name, 0, (__pyx_t_9 - 2)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_new_word_name = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":292
 *                     self.fsm.morphologicalAnalysis(capitalized_word_name).size() == 0:
 *                 new_word_name = word_name[0:len(word_name) - 2]
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(new_word_name)             # <<<<<<<<<<<<<<
 *                 txt_new_word = self.fsm.getDictionary().getWord(Word.toLower(new_word_name))
 *                 if txt_new_word is not None and isinstance(txt_new_word, TxtWord) and txt_new_word.isProperNoun():
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_v_new_word_name, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList))))) __PYX_ERR(0, 292, __pyx_L1_error)
      __pyx_v_fsm_parse_list = ((struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":293
 *                 new_word_name = word_name[0:len(word_name) - 2]
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(new_word_name)
 *                 txt_new_word = self.fsm.getDictionary().getWord(Word.toLower(new_word_name))             # <<<<<<<<<<<<<<
 *                 if txt_new_word is not None and isinstance(txt_new_word, TxtWord) and txt_new_word.isProperNoun():
 *                     if self.fsm.morphologicalAnalysis(new_word_name + "'" + "da").size() > 0:
 */
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->getDictionary(__pyx_v_self->fsm, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_n_s_toLower); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_new_word_name) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_new_word_name);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 293, __pyx_L1_error)
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_10Dictionary_13TxtDictionary_TxtDictionary *)((struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *)__pyx_t_2)->__pyx_base.__pyx_vtab)->__pyx_base.getWord(((struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *)__pyx_t_2), ((PyObject*)__pyx_t_3), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10Dictionary_7TxtWord_TxtWord))))) __PYX_ERR(0, 293, __pyx_L1_error)
      __pyx_v_txt_new_word = ((struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":294
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(new_word_name)
 *                 txt_new_word = self.fsm.getDictionary().getWord(Word.toLower(new_word_name))
 *                 if txt_new_word is not None and isinstance(txt_new_word, TxtWord) and txt_new_word.isProperNoun():             # <<<<<<<<<<<<<<
 *                     if self.fsm.morphologicalAnalysis(new_word_name + "'" + "da").size() > 0:
 *                         result.addWord(Word(new_word_name + "'" + "da"))
 */
      __pyx_t_8 = (((PyObject *)__pyx_v_txt_new_word) != Py_None);
      __pyx_t_10 = (__pyx_t_8 != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_7 = __pyx_t_10;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_10 = __Pyx_TypeCheck(((PyObject *)__pyx_v_txt_new_word), __pyx_ptype_10Dictionary_7TxtWord_TxtWord); 
      __pyx_t_8 = (__pyx_t_10 != 0);
      if (__pyx_t_8) {
      } else {
        __pyx_t_7 = __pyx_t_8;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_8 = (((struct __pyx_vtabstruct_10Dictionary_7TxtWord_TxtWord *)__pyx_v_txt_new_word->__pyx_base.__pyx_vtab)->isProperNoun(__pyx_v_txt_new_word, 0) != 0);
      __pyx_t_7 = __pyx_t_8;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_7) {

        /* "SpellChecker/SimpleSpellChecker.pyx":295
 *                 txt_new_word = self.fsm.getDictionary().getWord(Word.toLower(new_word_name))
 *                 if txt_new_word is not None and isinstance(txt_new_word, TxtWord) and txt_new_word.isProperNoun():
 *                     if self.fsm.morphologicalAnalysis(new_word_name + "'" + "da").size() > 0:             # <<<<<<<<<<<<<<
 *                         result.addWord(Word(new_word_name + "'" + "da"))
 *                     else:
 */
        __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_v_new_word_name, __pyx_kp_u__5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 295, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_t_3, __pyx_n_u_da); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 295, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 295, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_7) {

          /* "SpellChecker/SimpleSpellChecker.pyx":296
 *                 if txt_new_word is not None and isinstance(txt_new_word, TxtWord) and txt_new_word.isProperNoun():
 *                     if self.fsm.morphologicalAnalysis(new_word_name + "'" + "da").size() > 0:
 *                         result.addWord(Word(new_word_name + "'" + "da"))             # <<<<<<<<<<<<<<
 *                     else:
 *                         result.addWord(Word(new_word_name + "'" + "de"))
 */
          __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_v_new_word_name, __pyx_kp_u__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_t_2, __pyx_n_u_da); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_2), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "SpellChecker/SimpleSpellChecker.pyx":295
 *                 txt_new_word = self.fsm.getDictionary().getWord(Word.toLower(new_word_name))
 *                 if txt_new_word is not None and isinstance(txt_new_word, TxtWord) and txt_new_word.isProperNoun():
 *                     if self.fsm.morphologicalAnalysis(new_word_name + "'" + "da").size() > 0:             # <<<<<<<<<<<<<<
 *                         result.addWord(Word(new_word_name + "'" + "da"))
 *                     else:
 */
          goto __pyx_L13;
        }

        /* "SpellChecker/SimpleSpellChecker.pyx":298
 *                         result.addWord(Word(new_word_name + "'" + "da"))
 *                     else:
 *                         result.addWord(Word(new_word_name + "'" + "de"))             # <<<<<<<<<<<<<<
 *                     return True
 *                 if fsm_parse_list.size() > 0:
 */
        /*else*/ {
          __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_v_new_word_name, __pyx_kp_u__5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_t_1, __pyx_n_u_de); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 298, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_1), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 298, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __pyx_L13:;

        /* "SpellChecker/SimpleSpellChecker.pyx":299
 *                     else:
 *                         result.addWord(Word(new_word_name + "'" + "de"))
 *                     return True             # <<<<<<<<<<<<<<
 *                 if fsm_parse_list.size() > 0:
 *                     txt_word = self.fsm.getDictionary().getWord(
 */
        __pyx_r = 1;
        goto __pyx_L0;

        /* "SpellChecker/SimpleSpellChecker.pyx":294
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(new_word_name)
 *                 txt_new_word = self.fsm.getDictionary().getWord(Word.toLower(new_word_name))
 *                 if txt_new_word is not None and isinstance(txt_new_word, TxtWord) and txt_new_word.isProperNoun():             # <<<<<<<<<<<<<<
 *                     if self.fsm.morphologicalAnalysis(new_word_name + "'" + "da").size() > 0:
 *                         result.addWord(Word(new_word_name + "'" + "da"))
 */
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":300
 *                         result.addWord(Word(new_word_name + "'" + "de"))
 *                     return True
 *                 if fsm_parse_list.size() > 0:             # <<<<<<<<<<<<<<
 *                     txt_word = self.fsm.getDictionary().getWord(
 *                         fsm_parse_list.getParseWithLongestRootWord().getWord().getName())
 */
      __pyx_t_7 = ((((struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_v_fsm_parse_list->__pyx_vtab)->size(__pyx_v_fsm_parse_list, 0) > 0) != 0);
      if (__pyx_t_7) {

        /* "SpellChecker/SimpleSpellChecker.pyx":301
 *                     return True
 *                 if fsm_parse_list.size() > 0:
 *                     txt_word = self.fsm.getDictionary().getWord(             # <<<<<<<<<<<<<<
 *                         fsm_parse_list.getParseWithLongestRootWord().getWord().getName())
 *                 if txt_word is not None and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 */
        __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->getDictionary(__pyx_v_self->fsm, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);

        /* "SpellChecker/SimpleSpellChecker.pyx":302
 *                 if fsm_parse_list.size() > 0:
 *                     txt_word = self.fsm.getDictionary().getWord(
 *                         fsm_parse_list.getParseWithLongestRootWord().getWord().getName())             # <<<<<<<<<<<<<<
 *                 if txt_word is not None and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 *                     result.addWord(Word(new_word_name))
 */
        __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_v_fsm_parse_list->__pyx_vtab)->getParseWithLongestRootWord(__pyx_v_fsm_parse_list, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_21MorphologicalAnalysis_8FsmParse_FsmParse *)((struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse *)__pyx_t_1)->__pyx_base.__pyx_vtab)->__pyx_base.getWord(((struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse *)__pyx_t_1), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 302, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_3)->__pyx_vtab)->getName(((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_3), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":301
 *                     return True
 *                 if fsm_parse_list.size() > 0:
 *                     txt_word = self.fsm.getDictionary().getWord(             # <<<<<<<<<<<<<<
 *                         fsm_parse_list.getParseWithLongestRootWord().getWord().getName())
 *                 if txt_word is not None and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 */
        __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_10Dictionary_13TxtDictionary_TxtDictionary *)((struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *)__pyx_t_2)->__pyx_base.__pyx_vtab)->__pyx_base.getWord(((struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *)__pyx_t_2), ((PyObject*)__pyx_t_1), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_10Dictionary_7TxtWord_TxtWord))))) __PYX_ERR(0, 301, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_txt_word, ((struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *)__pyx_t_3));
        __pyx_t_3 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":300
 *                         result.addWord(Word(new_word_name + "'" + "de"))
 *                     return True
 *                 if fsm_parse_list.size() > 0:             # <<<<<<<<<<<<<<
 *                     txt_word = self.fsm.getDictionary().getWord(
 *                         fsm_parse_list.getParseWithLongestRootWord().getWord().getName())
 */
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":303
 *                     txt_word = self.fsm.getDictionary().getWord(
 *                         fsm_parse_list.getParseWithLongestRootWord().getWord().getName())
 *                 if txt_word is not None and isinstance(txt_word, TxtWord) and not txt_word.isCode():             # <<<<<<<<<<<<<<
 *                     result.addWord(Word(new_word_name))
 *                     if TurkishLanguage.isBackVowel(Word.lastVowel(new_word_name)):
 */
      __pyx_t_8 = (((PyObject *)__pyx_v_txt_word) != Py_None);
      __pyx_t_10 = (__pyx_t_8 != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_7 = __pyx_t_10;
        goto __pyx_L16_bool_binop_done;
      }
      __pyx_t_10 = __Pyx_TypeCheck(((PyObject *)__pyx_v_txt_word), __pyx_ptype_10Dictionary_7TxtWord_TxtWord); 
      __pyx_t_8 = (__pyx_t_10 != 0);
      if (__pyx_t_8) {
      } else {
        __pyx_t_7 = __pyx_t_8;
        goto __pyx_L16_bool_binop_done;
      }
      __pyx_t_8 = ((!(((struct __pyx_vtabstruct_10Dictionary_7TxtWord_TxtWord *)__pyx_v_txt_word->__pyx_base.__pyx_vtab)->isCode(__pyx_v_txt_word, 0) != 0)) != 0);
      __pyx_t_7 = __pyx_t_8;
      __pyx_L16_bool_binop_done:;
      if (__pyx_t_7) {

        /* "SpellChecker/SimpleSpellChecker.pyx":304
 *                         fsm_parse_list.getParseWithLongestRootWord().getWord().getName())
 *                 if txt_word is not None and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 *                     result.addWord(Word(new_word_name))             # <<<<<<<<<<<<<<
 *                     if TurkishLanguage.isBackVowel(Word.lastVowel(new_word_name)):
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():
 */
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_v_new_word_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 304, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_3), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":305
 *                 if txt_word is not None and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 *                     result.addWord(Word(new_word_name))
 *                     if TurkishLanguage.isBackVowel(Word.lastVowel(new_word_name)):             # <<<<<<<<<<<<<<
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():
 *                             result.addWord(Word("de"))
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8Language_15TurkishLanguage_TurkishLanguage), __pyx_n_s_isBackVowel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_n_s_lastVowel); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_new_word_name) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_new_word_name);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_7) {

          /* "SpellChecker/SimpleSpellChecker.pyx":306
 *                     result.addWord(Word(new_word_name))
 *                     if TurkishLanguage.isBackVowel(Word.lastVowel(new_word_name)):
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():             # <<<<<<<<<<<<<<
 *                             result.addWord(Word("de"))
 *                         else:
 */
          __pyx_t_7 = (((struct __pyx_vtabstruct_10Dictionary_7TxtWord_TxtWord *)__pyx_v_txt_word->__pyx_base.__pyx_vtab)->notObeysVowelHarmonyDuringAgglutination(__pyx_v_txt_word, 0) != 0);
          if (__pyx_t_7) {

            /* "SpellChecker/SimpleSpellChecker.pyx":307
 *                     if TurkishLanguage.isBackVowel(Word.lastVowel(new_word_name)):
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():
 *                             result.addWord(Word("de"))             # <<<<<<<<<<<<<<
 *                         else:
 *                             result.addWord(Word("da"))
 */
            __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_1), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 307, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "SpellChecker/SimpleSpellChecker.pyx":306
 *                     result.addWord(Word(new_word_name))
 *                     if TurkishLanguage.isBackVowel(Word.lastVowel(new_word_name)):
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():             # <<<<<<<<<<<<<<
 *                             result.addWord(Word("de"))
 *                         else:
 */
            goto __pyx_L20;
          }

          /* "SpellChecker/SimpleSpellChecker.pyx":309
 *                             result.addWord(Word("de"))
 *                         else:
 *                             result.addWord(Word("da"))             # <<<<<<<<<<<<<<
 *                     else:
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():
 */
          /*else*/ {
            __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 309, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_1 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_3), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __pyx_L20:;

          /* "SpellChecker/SimpleSpellChecker.pyx":305
 *                 if txt_word is not None and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 *                     result.addWord(Word(new_word_name))
 *                     if TurkishLanguage.isBackVowel(Word.lastVowel(new_word_name)):             # <<<<<<<<<<<<<<
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():
 *                             result.addWord(Word("de"))
 */
          goto __pyx_L19;
        }

        /* "SpellChecker/SimpleSpellChecker.pyx":311
 *                             result.addWord(Word("da"))
 *                     else:
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():             # <<<<<<<<<<<<<<
 *                             result.addWord(Word("da"))
 *                         else:
 */
        /*else*/ {
          __pyx_t_7 = (((struct __pyx_vtabstruct_10Dictionary_7TxtWord_TxtWord *)__pyx_v_txt_word->__pyx_base.__pyx_vtab)->notObeysVowelHarmonyDuringAgglutination(__pyx_v_txt_word, 0) != 0);
          if (__pyx_t_7) {

            /* "SpellChecker/SimpleSpellChecker.pyx":312
 *                     else:
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():
 *                             result.addWord(Word("da"))             # <<<<<<<<<<<<<<
 *                         else:
 *                             result.addWord(Word("de"))
 */
            __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_1), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 312, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "SpellChecker/SimpleSpellChecker.pyx":311
 *                             result.addWord(Word("da"))
 *                     else:
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():             # <<<<<<<<<<<<<<
 *                             result.addWord(Word("da"))
 *                         else:
 */
            goto __pyx_L21;
          }

          /* "SpellChecker/SimpleSpellChecker.pyx":314
 *                             result.addWord(Word("da"))
 *                         else:
 *                             result.addWord(Word("de"))             # <<<<<<<<<<<<<<
 *                     return True
 *         return False
 */
          /*else*/ {
            __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 314, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_1 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_3), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __pyx_L21:;
        }
        __pyx_L19:;

        /* "SpellChecker/SimpleSpellChecker.pyx":315
 *                         else:
 *                             result.addWord(Word("de"))
 *                     return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
        __pyx_r = 1;
        goto __pyx_L0;

        /* "SpellChecker/SimpleSpellChecker.pyx":303
 *                     txt_word = self.fsm.getDictionary().getWord(
 *                         fsm_parse_list.getParseWithLongestRootWord().getWord().getName())
 *                 if txt_word is not None and isinstance(txt_word, TxtWord) and not txt_word.isCode():             # <<<<<<<<<<<<<<
 *                     result.addWord(Word(new_word_name))
 *                     if TurkishLanguage.isBackVowel(Word.lastVowel(new_word_name)):
 */
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":289
 *         txt_word = None
 *         if word_name.endswith("da") or word_name.endswith("de"):
 *             if self.fsm.morphologicalAnalysis(word_name).size() == 0 and \             # <<<<<<<<<<<<<<
 *                     self.fsm.morphologicalAnalysis(capitalized_word_name).size() == 0:
 *                 new_word_name = word_name[0:len(word_name) - 2]
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":288
 *         capitalized_word_name = Word.toCapital(word_name)
 *         txt_word = None
 *         if word_name.endswith("da") or word_name.endswith("de"):             # <<<<<<<<<<<<<<
 *             if self.fsm.morphologicalAnalysis(word_name).size() == 0 and \
 *                     self.fsm.morphologicalAnalysis(capitalized_word_name).size() == 0:
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":316
 *                             result.addWord(Word("de"))
 *                     return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint forcedSuffixMergeCheck(self,
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":279
 *         return False
 * 
 *     cpdef bint forcedDeDaSplitCheck(self,             # <<<<<<<<<<<<<<
 *                              Word word,
 *                              Sentence result):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedDeDaSplitCheck", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_word_name);
  __Pyx_XDECREF(__pyx_v_capitalized_word_name);
  __Pyx_XDECREF(__pyx_v_new_word_name);
  __Pyx_XDECREF((PyObject *)__pyx_v_txt_word);
  __Pyx_XDECREF((PyObject *)__pyx_v_txt_new_word);
  __Pyx_XDECREF((PyObject *)__pyx_v_fsm_parse_list);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_21forcedDeDaSplitCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_21forcedDeDaSplitCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("forcedDeDaSplitCheck (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word,&__pyx_n_s_result,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedDeDaSplitCheck", 1, 2, 2, 1); __PYX_ERR(0, 279, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "forcedDeDaSplitCheck") < 0)) __PYX_ERR(0, 279, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("forcedDeDaSplitCheck", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 279, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedDeDaSplitCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 280, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "result", 0))) __PYX_ERR(0, 281, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_20forcedDeDaSplitCheck(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_word, __pyx_v_result);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_20forcedDeDaSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedDeDaSplitCheck", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedDeDaSplitCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedDeDaSplitCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":318
 *         return False
 * 
 *     cpdef bint forcedSuffixMergeCheck(self,             # <<<<<<<<<<<<<<
 *                                Word word,
 *                                Sentence result,
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_23forcedSuffixMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedSuffixMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_li_list = 0;
  PyObject *__pyx_v_lik_list = 0;
  PyObject *__pyx_v_suffix = 0;
  PyObject *__pyx_v_compiled_expression = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedSuffixMergeCheck", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_forcedSuffixMergeCheck); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_23forcedSuffixMergeCheck)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result), ((PyObject *)__pyx_v_previousWord)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result), ((PyObject *)__pyx_v_previousWord)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 318, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_result));
          __Pyx_INCREF(((PyObject *)__pyx_v_previousWord));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_previousWord));
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, ((PyObject *)__pyx_v_previousWord));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":324
 *         cdef list li_list, lik_list
 *         cdef str suffix
 *         li_list = ["li", "l", "lu", "l"]             # <<<<<<<<<<<<<<
 *         lik_list = ["lik", "lk", "luk", "lk"]
 *         compiled_expression = re.compile("[0-9]+")
 */
  __pyx_t_1 = PyList_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_li);
  __Pyx_GIVEREF(__pyx_n_u_li);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_li);
  __Pyx_INCREF(__pyx_n_u_l);
  __Pyx_GIVEREF(__pyx_n_u_l);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_l);
  __Pyx_INCREF(__pyx_n_u_lu);
  __Pyx_GIVEREF(__pyx_n_u_lu);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_lu);
  __Pyx_INCREF(__pyx_n_u_l_2);
  __Pyx_GIVEREF(__pyx_n_u_l_2);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_l_2);
  __pyx_v_li_list = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":325
 *         cdef str suffix
 *         li_list = ["li", "l", "lu", "l"]
 *         lik_list = ["lik", "lk", "luk", "lk"]             # <<<<<<<<<<<<<<
 *         compiled_expression = re.compile("[0-9]+")
 *         if word.getName() in li_list or word.getName() in lik_list:
 */
  __pyx_t_1 = PyList_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_lik);
  __Pyx_GIVEREF(__pyx_n_u_lik);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_lik);
  __Pyx_INCREF(__pyx_n_u_lk);
  __Pyx_GIVEREF(__pyx_n_u_lk);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_lk);
  __Pyx_INCREF(__pyx_n_u_luk);
  __Pyx_GIVEREF(__pyx_n_u_luk);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_luk);
  __Pyx_INCREF(__pyx_n_u_lk_2);
  __Pyx_GIVEREF(__pyx_n_u_lk_2);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_lk_2);
  __pyx_v_lik_list = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":326
 *         li_list = ["li", "l", "lu", "l"]
 *         lik_list = ["lik", "lk", "luk", "lk"]
 *         compiled_expression = re.compile("[0-9]+")             # <<<<<<<<<<<<<<
 *         if word.getName() in li_list or word.getName() in lik_list:
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_u_0_9) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_0_9);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_compiled_expression = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":327
 *         lik_list = ["lik", "lk", "luk", "lk"]
 *         compiled_expression = re.compile("[0-9]+")
 *         if word.getName() in li_list or word.getName() in lik_list:             # <<<<<<<<<<<<<<
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):
 *                 for suffix in li_list:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_li_list, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (!__pyx_t_9) {
  } else {
    __pyx_t_7 = __pyx_t_9;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_lik_list, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = (__pyx_t_9 != 0);
  __pyx_t_7 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_7) {

    /* "SpellChecker/SimpleSpellChecker.pyx":328
 *         compiled_expression = re.compile("[0-9]+")
 *         if word.getName() in li_list or word.getName() in lik_list:
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):             # <<<<<<<<<<<<<<
 *                 for suffix in li_list:
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(
 */
    __pyx_t_8 = (((PyObject *)__pyx_v_previousWord) != Py_None);
    __pyx_t_9 = (__pyx_t_8 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_7 = __pyx_t_9;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiled_expression, __pyx_n_s_fullmatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_previousWord->__pyx_vtab)->getName(__pyx_v_previousWord, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __pyx_t_9;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_7) {

      /* "SpellChecker/SimpleSpellChecker.pyx":329
 *         if word.getName() in li_list or word.getName() in lik_list:
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):
 *                 for suffix in li_list:             # <<<<<<<<<<<<<<
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 */
      __pyx_t_1 = __pyx_v_li_list; __Pyx_INCREF(__pyx_t_1); __pyx_t_10 = 0;
      for (;;) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 329, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
        if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 329, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_suffix, ((PyObject*)__pyx_t_3));
        __pyx_t_3 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":330
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):
 *                 for suffix in li_list:
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(             # <<<<<<<<<<<<<<
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 */
        __pyx_t_3 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (unlikely(__pyx_t_3 == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(0, 330, __pyx_L1_error)
        }
        __pyx_t_11 = __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 330, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_9 = ((__pyx_t_11 == 2) != 0);
        if (__pyx_t_9) {
        } else {
          __pyx_t_7 = __pyx_t_9;
          goto __pyx_L12_bool_binop_done;
        }

        /* "SpellChecker/SimpleSpellChecker.pyx":331
 *                 for suffix in li_list:
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(
 *                             previousWord.getName() + "'" + suffix).size() > 0:             # <<<<<<<<<<<<<<
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 *                         return True
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_previousWord->__pyx_vtab)->getName(__pyx_v_previousWord, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 331, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_2, __pyx_kp_u__5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 331, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_6, __pyx_v_suffix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 331, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":330
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):
 *                 for suffix in li_list:
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(             # <<<<<<<<<<<<<<
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 */
        __pyx_t_6 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_t_2, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":331
 *                 for suffix in li_list:
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(
 *                             previousWord.getName() + "'" + suffix).size() > 0:             # <<<<<<<<<<<<<<
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 *                         return True
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 331, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 331, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 331, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 331, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_7 = __pyx_t_9;
        __pyx_L12_bool_binop_done:;

        /* "SpellChecker/SimpleSpellChecker.pyx":330
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):
 *                 for suffix in li_list:
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(             # <<<<<<<<<<<<<<
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 */
        if (__pyx_t_7) {

          /* "SpellChecker/SimpleSpellChecker.pyx":332
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))             # <<<<<<<<<<<<<<
 *                         return True
 *                 for suffix in lik_list:
 */
          __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_previousWord->__pyx_vtab)->getName(__pyx_v_previousWord, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 332, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_2, __pyx_kp_u__5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 332, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_3, __pyx_v_suffix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 332, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 332, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->replaceWord(__pyx_v_result, (((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->wordCount(__pyx_v_result, 0) - 1), ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_3), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 332, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "SpellChecker/SimpleSpellChecker.pyx":333
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 *                         return True             # <<<<<<<<<<<<<<
 *                 for suffix in lik_list:
 *                     if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(
 */
          __pyx_r = 1;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L0;

          /* "SpellChecker/SimpleSpellChecker.pyx":330
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):
 *                 for suffix in li_list:
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(             # <<<<<<<<<<<<<<
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 */
        }

        /* "SpellChecker/SimpleSpellChecker.pyx":329
 *         if word.getName() in li_list or word.getName() in lik_list:
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):
 *                 for suffix in li_list:             # <<<<<<<<<<<<<<
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":334
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 *                         return True
 *                 for suffix in lik_list:             # <<<<<<<<<<<<<<
 *                     if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 */
      __pyx_t_1 = __pyx_v_lik_list; __Pyx_INCREF(__pyx_t_1); __pyx_t_10 = 0;
      for (;;) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 334, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
        if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 334, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_suffix, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":335
 *                         return True
 *                 for suffix in lik_list:
 *                     if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(             # <<<<<<<<<<<<<<
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(__pyx_t_2 == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(0, 335, __pyx_L1_error)
        }
        __pyx_t_11 = __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_9 = ((__pyx_t_11 == 3) != 0);
        if (__pyx_t_9) {
        } else {
          __pyx_t_7 = __pyx_t_9;
          goto __pyx_L17_bool_binop_done;
        }

        /* "SpellChecker/SimpleSpellChecker.pyx":336
 *                 for suffix in lik_list:
 *                     if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(
 *                             previousWord.getName() + "'" + suffix).size() > 0:             # <<<<<<<<<<<<<<
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 *                         return True
 */
        __pyx_t_3 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_previousWord->__pyx_vtab)->getName(__pyx_v_previousWord, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_3, __pyx_kp_u__5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_6, __pyx_v_suffix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":335
 *                         return True
 *                 for suffix in lik_list:
 *                     if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(             # <<<<<<<<<<<<<<
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 */
        __pyx_t_6 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_t_3, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":336
 *                 for suffix in lik_list:
 *                     if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(
 *                             previousWord.getName() + "'" + suffix).size() > 0:             # <<<<<<<<<<<<<<
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 *                         return True
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 336, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 336, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __pyx_t_9;
        __pyx_L17_bool_binop_done:;

        /* "SpellChecker/SimpleSpellChecker.pyx":335
 *                         return True
 *                 for suffix in lik_list:
 *                     if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(             # <<<<<<<<<<<<<<
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 */
        if (__pyx_t_7) {

          /* "SpellChecker/SimpleSpellChecker.pyx":337
 *                     if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))             # <<<<<<<<<<<<<<
 *                         return True
 *         return False
 */
          __pyx_t_3 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_previousWord->__pyx_vtab)->getName(__pyx_v_previousWord, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 337, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_3, __pyx_kp_u__5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_2, __pyx_v_suffix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 337, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->replaceWord(__pyx_v_result, (((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->wordCount(__pyx_v_result, 0) - 1), ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_2), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 337, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "SpellChecker/SimpleSpellChecker.pyx":338
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 *                         return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
          __pyx_r = 1;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L0;

          /* "SpellChecker/SimpleSpellChecker.pyx":335
 *                         return True
 *                 for suffix in lik_list:
 *                     if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(             # <<<<<<<<<<<<<<
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 */
        }

        /* "SpellChecker/SimpleSpellChecker.pyx":334
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 *                         return True
 *                 for suffix in lik_list:             # <<<<<<<<<<<<<<
 *                     if len(word.getName()) == 3 and self.fsm.morphologicalAnalysis(
 *                             previousWord.getName() + "'" + suffix).size() > 0:
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":328
 *         compiled_expression = re.compile("[0-9]+")
 *         if word.getName() in li_list or word.getName() in lik_list:
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):             # <<<<<<<<<<<<<<
 *                 for suffix in li_list:
 *                     if len(word.getName()) == 2 and self.fsm.morphologicalAnalysis(
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":327
 *         lik_list = ["lik", "lk", "luk", "lk"]
 *         compiled_expression = re.compile("[0-9]+")
 *         if word.getName() in li_list or word.getName() in lik_list:             # <<<<<<<<<<<<<<
 *             if previousWord is not None and compiled_expression.fullmatch(previousWord.getName()):
 *                 for suffix in li_list:
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":339
 *                         result.replaceWord(result.wordCount() - 1, Word(previousWord.getName() + "'" + suffix))
 *                         return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint forcedHyphenMergeCheck(self,
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":318
 *         return False
 * 
 *     cpdef bint forcedSuffixMergeCheck(self,             # <<<<<<<<<<<<<<
 *                                Word word,
 *                                Sentence result,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedSuffixMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_li_list);
  __Pyx_XDECREF(__pyx_v_lik_list);
  __Pyx_XDECREF(__pyx_v_suffix);
  __Pyx_XDECREF(__pyx_v_compiled_expression);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_23forcedSuffixMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_23forcedSuffixMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("forcedSuffixMergeCheck (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word,&__pyx_n_s_result,&__pyx_n_s_previousWord,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedSuffixMergeCheck", 1, 3, 3, 1); __PYX_ERR(0, 318, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_previousWord)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedSuffixMergeCheck", 1, 3, 3, 2); __PYX_ERR(0, 318, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "forcedSuffixMergeCheck") < 0)) __PYX_ERR(0, 318, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
    __pyx_v_previousWord = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("forcedSuffixMergeCheck", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 318, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedSuffixMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 319, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "result", 0))) __PYX_ERR(0, 320, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_previousWord), __pyx_ptype_10Dictionary_4Word_Word, 1, "previousWord", 0))) __PYX_ERR(0, 321, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_22forcedSuffixMergeCheck(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_word, __pyx_v_result, __pyx_v_previousWord);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_22forcedSuffixMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedSuffixMergeCheck", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedSuffixMergeCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, __pyx_v_previousWord, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedSuffixMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":341
 *         return False
 * 
 *     cpdef bint forcedHyphenMergeCheck(self,             # <<<<<<<<<<<<<<
 *                                Word word,
 *                                Sentence result,
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_25forcedHyphenMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedHyphenMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_new_word_name = 0;
  PyObject *__pyx_v_compiled_expression = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedHyphenMergeCheck", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_forcedHyphenMergeCheck); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_25forcedHyphenMergeCheck)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[5] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result), ((PyObject *)__pyx_v_previousWord), ((PyObject *)__pyx_v_nextWord)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 341, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[5] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result), ((PyObject *)__pyx_v_previousWord), ((PyObject *)__pyx_v_nextWord)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 341, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(4+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 341, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_result));
          __Pyx_INCREF(((PyObject *)__pyx_v_previousWord));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_previousWord));
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, ((PyObject *)__pyx_v_previousWord));
          __Pyx_INCREF(((PyObject *)__pyx_v_nextWord));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_nextWord));
          PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_5, ((PyObject *)__pyx_v_nextWord));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 341, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 341, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":347
 *                                Word nextWord):
 *         cdef str new_word_name
 *         if word.getName() == "-" or word.getName() == "" or word.getName() == "":             # <<<<<<<<<<<<<<
 *             compiled_expression = re.compile("[a-zA-Z]+")
 *             if previousWord is not None and nextWord is not None and \
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_kp_u__8, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (!__pyx_t_9) {
  } else {
    __pyx_t_7 = __pyx_t_9;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_kp_u__9, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = (__pyx_t_9 != 0);
  if (!__pyx_t_8) {
  } else {
    __pyx_t_7 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_kp_u__10, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_7 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_7) {

    /* "SpellChecker/SimpleSpellChecker.pyx":348
 *         cdef str new_word_name
 *         if word.getName() == "-" or word.getName() == "" or word.getName() == "":
 *             compiled_expression = re.compile("[a-zA-Z]+")             # <<<<<<<<<<<<<<
 *             if previousWord is not None and nextWord is not None and \
 *                     compiled_expression.fullmatch(previousWord.getName()) and \
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 348, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 348, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_u_a_zA_Z) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_a_zA_Z);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_compiled_expression = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":349
 *         if word.getName() == "-" or word.getName() == "" or word.getName() == "":
 *             compiled_expression = re.compile("[a-zA-Z]+")
 *             if previousWord is not None and nextWord is not None and \             # <<<<<<<<<<<<<<
 *                     compiled_expression.fullmatch(previousWord.getName()) and \
 *                     compiled_expression.fullmatch(nextWord.getName()):
 */
    __pyx_t_9 = (((PyObject *)__pyx_v_previousWord) != Py_None);
    __pyx_t_8 = (__pyx_t_9 != 0);
    if (__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_nextWord) != Py_None);
    __pyx_t_9 = (__pyx_t_8 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_7 = __pyx_t_9;
      goto __pyx_L8_bool_binop_done;
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":350
 *             compiled_expression = re.compile("[a-zA-Z]+")
 *             if previousWord is not None and nextWord is not None and \
 *                     compiled_expression.fullmatch(previousWord.getName()) and \             # <<<<<<<<<<<<<<
 *                     compiled_expression.fullmatch(nextWord.getName()):
 *                 new_word_name = previousWord.getName() + "-" + nextWord.getName()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiled_expression, __pyx_n_s_fullmatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_previousWord->__pyx_vtab)->getName(__pyx_v_previousWord, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_9) {
    } else {
      __pyx_t_7 = __pyx_t_9;
      goto __pyx_L8_bool_binop_done;
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":351
 *             if previousWord is not None and nextWord is not None and \
 *                     compiled_expression.fullmatch(previousWord.getName()) and \
 *                     compiled_expression.fullmatch(nextWord.getName()):             # <<<<<<<<<<<<<<
 *                 new_word_name = previousWord.getName() + "-" + nextWord.getName()
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_compiled_expression, __pyx_n_s_fullmatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_nextWord->__pyx_vtab)->getName(__pyx_v_nextWord, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 351, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __pyx_t_9;
    __pyx_L8_bool_binop_done:;

    /* "SpellChecker/SimpleSpellChecker.pyx":349
 *         if word.getName() == "-" or word.getName() == "" or word.getName() == "":
 *             compiled_expression = re.compile("[a-zA-Z]+")
 *             if previousWord is not None and nextWord is not None and \             # <<<<<<<<<<<<<<
 *                     compiled_expression.fullmatch(previousWord.getName()) and \
 *                     compiled_expression.fullmatch(nextWord.getName()):
 */
    if (__pyx_t_7) {

      /* "SpellChecker/SimpleSpellChecker.pyx":352
 *                     compiled_expression.fullmatch(previousWord.getName()) and \
 *                     compiled_expression.fullmatch(nextWord.getName()):
 *                 new_word_name = previousWord.getName() + "-" + nextWord.getName()             # <<<<<<<<<<<<<<
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0:
 *                     result.replaceWord(result.wordCount() - 1, Word(new_word_name))
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_previousWord->__pyx_vtab)->getName(__pyx_v_previousWord, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_1, __pyx_kp_u__8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_nextWord->__pyx_vtab)->getName(__pyx_v_nextWord, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_new_word_name = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":353
 *                     compiled_expression.fullmatch(nextWord.getName()):
 *                 new_word_name = previousWord.getName() + "-" + nextWord.getName()
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0:             # <<<<<<<<<<<<<<
 *                     result.replaceWord(result.wordCount() - 1, Word(new_word_name))
 *                     return True
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_v_new_word_name, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {

        /* "SpellChecker/SimpleSpellChecker.pyx":354
 *                 new_word_name = previousWord.getName() + "-" + nextWord.getName()
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0:
 *                     result.replaceWord(result.wordCount() - 1, Word(new_word_name))             # <<<<<<<<<<<<<<
 *                     return True
 *         return False
 */
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_v_new_word_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->replaceWord(__pyx_v_result, (((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->wordCount(__pyx_v_result, 0) - 1), ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_3), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":355
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0:
 *                     result.replaceWord(result.wordCount() - 1, Word(new_word_name))
 *                     return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
        __pyx_r = 1;
        goto __pyx_L0;

        /* "SpellChecker/SimpleSpellChecker.pyx":353
 *                     compiled_expression.fullmatch(nextWord.getName()):
 *                 new_word_name = previousWord.getName() + "-" + nextWord.getName()
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0:             # <<<<<<<<<<<<<<
 *                     result.replaceWord(result.wordCount() - 1, Word(new_word_name))
 *                     return True
 */
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":349
 *         if word.getName() == "-" or word.getName() == "" or word.getName() == "":
 *             compiled_expression = re.compile("[a-zA-Z]+")
 *             if previousWord is not None and nextWord is not None and \             # <<<<<<<<<<<<<<
 *                     compiled_expression.fullmatch(previousWord.getName()) and \
 *                     compiled_expression.fullmatch(nextWord.getName()):
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":347
 *                                Word nextWord):
 *         cdef str new_word_name
 *         if word.getName() == "-" or word.getName() == "" or word.getName() == "":             # <<<<<<<<<<<<<<
 *             compiled_expression = re.compile("[a-zA-Z]+")
 *             if previousWord is not None and nextWord is not None and \
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":356
 *                     result.replaceWord(result.wordCount() - 1, Word(new_word_name))
 *                     return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint forcedQuestionSuffixSplitCheck(self,
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":341
 *         return False
 * 
 *     cpdef bint forcedHyphenMergeCheck(self,             # <<<<<<<<<<<<<<
 *                                Word word,
 *                                Sentence result,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedHyphenMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_new_word_name);
  __Pyx_XDECREF(__pyx_v_compiled_expression);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_25forcedHyphenMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_25forcedHyphenMergeCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord = 0;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("forcedHyphenMergeCheck (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word,&__pyx_n_s_result,&__pyx_n_s_previousWord,&__pyx_n_s_nextWord,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedHyphenMergeCheck", 1, 4, 4, 1); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_previousWord)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedHyphenMergeCheck", 1, 4, 4, 2); __PYX_ERR(0, 341, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nextWord)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedHyphenMergeCheck", 1, 4, 4, 3); __PYX_ERR(0, 341, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "forcedHyphenMergeCheck") < 0)) __PYX_ERR(0, 341, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
    __pyx_v_previousWord = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[2]);
    __pyx_v_nextWord = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[3]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("forcedHyphenMergeCheck", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 341, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedHyphenMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 342, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "result", 0))) __PYX_ERR(0, 343, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_previousWord), __pyx_ptype_10Dictionary_4Word_Word, 1, "previousWord", 0))) __PYX_ERR(0, 344, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nextWord), __pyx_ptype_10Dictionary_4Word_Word, 1, "nextWord", 0))) __PYX_ERR(0, 345, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_24forcedHyphenMergeCheck(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_word, __pyx_v_result, __pyx_v_previousWord, __pyx_v_nextWord);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_24forcedHyphenMergeCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedHyphenMergeCheck", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedHyphenMergeCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, __pyx_v_previousWord, __pyx_v_nextWord, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedHyphenMergeCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":358
 *         return False
 * 
 *     cpdef bint forcedQuestionSuffixSplitCheck(self,             # <<<<<<<<<<<<<<
 *                                        Word word,
 *                                        Sentence result):
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_27forcedQuestionSuffixSplitCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedQuestionSuffixSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_word_name = 0;
  PyObject *__pyx_v_new_word_name = 0;
  struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *__pyx_v_txt_word = 0;
  PyObject *__pyx_v_question_suffix = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedQuestionSuffixSplitCheck", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_forcedQuestionSuffixSplitCheck); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_27forcedQuestionSuffixSplitCheck)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_result)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_result));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":363
 *         cdef str word_name, new_word_name
 *         cdef TxtWord txt_word
 *         word_name = word.getName()             # <<<<<<<<<<<<<<
 *         if self.fsm.morphologicalAnalysis(word_name).size() > 0:
 *             return False
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_word_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":364
 *         cdef TxtWord txt_word
 *         word_name = word.getName()
 *         if self.fsm.morphologicalAnalysis(word_name).size() > 0:             # <<<<<<<<<<<<<<
 *             return False
 *         for question_suffix in self.__questionSuffixList:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_v_word_name, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_7) {

    /* "SpellChecker/SimpleSpellChecker.pyx":365
 *         word_name = word.getName()
 *         if self.fsm.morphologicalAnalysis(word_name).size() > 0:
 *             return False             # <<<<<<<<<<<<<<
 *         for question_suffix in self.__questionSuffixList:
 *             if word_name.endswith(question_suffix):
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "SpellChecker/SimpleSpellChecker.pyx":364
 *         cdef TxtWord txt_word
 *         word_name = word.getName()
 *         if self.fsm.morphologicalAnalysis(word_name).size() > 0:             # <<<<<<<<<<<<<<
 *             return False
 *         for question_suffix in self.__questionSuffixList:
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":366
 *         if self.fsm.morphologicalAnalysis(word_name).size() > 0:
 *             return False
 *         for question_suffix in self.__questionSuffixList:             # <<<<<<<<<<<<<<
 *             if word_name.endswith(question_suffix):
 *                 new_word_name = word_name[0:word_name.rindex(question_suffix)]
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_questionSuffixList); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_1 = __pyx_t_3; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 366, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 366, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 366, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 366, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_9(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 366, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_question_suffix, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":367
 *             return False
 *         for question_suffix in self.__questionSuffixList:
 *             if word_name.endswith(question_suffix):             # <<<<<<<<<<<<<<
 *                 new_word_name = word_name[0:word_name.rindex(question_suffix)]
 *                 txt_word = self.fsm.getDictionary().getWord(new_word_name)
 */
    if (unlikely(__pyx_v_word_name == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "endswith");
      __PYX_ERR(0, 367, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_PyUnicode_Tailmatch(__pyx_v_word_name, __pyx_v_question_suffix, 0, PY_SSIZE_T_MAX, 1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 367, __pyx_L1_error)
    if ((__pyx_t_7 != 0)) {

      /* "SpellChecker/SimpleSpellChecker.pyx":368
 *         for question_suffix in self.__questionSuffixList:
 *             if word_name.endswith(question_suffix):
 *                 new_word_name = word_name[0:word_name.rindex(question_suffix)]             # <<<<<<<<<<<<<<
 *                 txt_word = self.fsm.getDictionary().getWord(new_word_name)
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0 and txt_word is not None \
 */
      if (unlikely(__pyx_v_word_name == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 368, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyUnicode_Type_rindex, __pyx_v_word_name, __pyx_v_question_suffix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = (__pyx_t_3 == Py_None);
      if (__pyx_t_7) {
        __pyx_t_10 = PY_SSIZE_T_MAX;
      } else {
        __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 368, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_11;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyUnicode_Substring(__pyx_v_word_name, 0, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 368, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_new_word_name, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":369
 *             if word_name.endswith(question_suffix):
 *                 new_word_name = word_name[0:word_name.rindex(question_suffix)]
 *                 txt_word = self.fsm.getDictionary().getWord(new_word_name)             # <<<<<<<<<<<<<<
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0 and txt_word is not None \
 *                         and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->getDictionary(__pyx_v_self->fsm, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 369, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_10Dictionary_13TxtDictionary_TxtDictionary *)((struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary *)__pyx_t_3)->__pyx_base.__pyx_vtab)->__pyx_base.getWord(((struct __pyx_obj_10Dictionary_10Dictionary_Dictionary *)__pyx_t_3), __pyx_v_new_word_name, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 369, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10Dictionary_7TxtWord_TxtWord))))) __PYX_ERR(0, 369, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_txt_word, ((struct __pyx_obj_10Dictionary_7TxtWord_TxtWord *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":370
 *                 new_word_name = word_name[0:word_name.rindex(question_suffix)]
 *                 txt_word = self.fsm.getDictionary().getWord(new_word_name)
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0 and txt_word is not None \             # <<<<<<<<<<<<<<
 *                         and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 *                     result.addWord(Word(new_word_name))
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_v_new_word_name, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 370, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 370, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 370, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 370, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 370, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_12) {
      } else {
        __pyx_t_7 = __pyx_t_12;
        goto __pyx_L8_bool_binop_done;
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":371
 *                 txt_word = self.fsm.getDictionary().getWord(new_word_name)
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0 and txt_word is not None \
 *                         and isinstance(txt_word, TxtWord) and not txt_word.isCode():             # <<<<<<<<<<<<<<
 *                     result.addWord(Word(new_word_name))
 *                     result.addWord(Word(question_suffix))
 */
      __pyx_t_12 = (((PyObject *)__pyx_v_txt_word) != Py_None);

      /* "SpellChecker/SimpleSpellChecker.pyx":370
 *                 new_word_name = word_name[0:word_name.rindex(question_suffix)]
 *                 txt_word = self.fsm.getDictionary().getWord(new_word_name)
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0 and txt_word is not None \             # <<<<<<<<<<<<<<
 *                         and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 *                     result.addWord(Word(new_word_name))
 */
      __pyx_t_13 = (__pyx_t_12 != 0);
      if (__pyx_t_13) {
      } else {
        __pyx_t_7 = __pyx_t_13;
        goto __pyx_L8_bool_binop_done;
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":371
 *                 txt_word = self.fsm.getDictionary().getWord(new_word_name)
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0 and txt_word is not None \
 *                         and isinstance(txt_word, TxtWord) and not txt_word.isCode():             # <<<<<<<<<<<<<<
 *                     result.addWord(Word(new_word_name))
 *                     result.addWord(Word(question_suffix))
 */
      __pyx_t_13 = __Pyx_TypeCheck(((PyObject *)__pyx_v_txt_word), __pyx_ptype_10Dictionary_7TxtWord_TxtWord); 
      __pyx_t_12 = (__pyx_t_13 != 0);
      if (__pyx_t_12) {
      } else {
        __pyx_t_7 = __pyx_t_12;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_12 = ((!(((struct __pyx_vtabstruct_10Dictionary_7TxtWord_TxtWord *)__pyx_v_txt_word->__pyx_base.__pyx_vtab)->isCode(__pyx_v_txt_word, 0) != 0)) != 0);
      __pyx_t_7 = __pyx_t_12;
      __pyx_L8_bool_binop_done:;

      /* "SpellChecker/SimpleSpellChecker.pyx":370
 *                 new_word_name = word_name[0:word_name.rindex(question_suffix)]
 *                 txt_word = self.fsm.getDictionary().getWord(new_word_name)
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0 and txt_word is not None \             # <<<<<<<<<<<<<<
 *                         and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 *                     result.addWord(Word(new_word_name))
 */
      if (__pyx_t_7) {

        /* "SpellChecker/SimpleSpellChecker.pyx":372
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0 and txt_word is not None \
 *                         and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 *                     result.addWord(Word(new_word_name))             # <<<<<<<<<<<<<<
 *                     result.addWord(Word(question_suffix))
 *                     return True
 */
        __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_v_new_word_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 372, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_6), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 372, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":373
 *                         and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 *                     result.addWord(Word(new_word_name))
 *                     result.addWord(Word(question_suffix))             # <<<<<<<<<<<<<<
 *                     return True
 *         return False
 */
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10Dictionary_4Word_Word), __pyx_v_question_suffix); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 373, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = ((struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence *)__pyx_v_result->__pyx_vtab)->addWord(__pyx_v_result, ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_t_2), 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 373, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "SpellChecker/SimpleSpellChecker.pyx":374
 *                     result.addWord(Word(new_word_name))
 *                     result.addWord(Word(question_suffix))
 *                     return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
        __pyx_r = 1;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;

        /* "SpellChecker/SimpleSpellChecker.pyx":370
 *                 new_word_name = word_name[0:word_name.rindex(question_suffix)]
 *                 txt_word = self.fsm.getDictionary().getWord(new_word_name)
 *                 if self.fsm.morphologicalAnalysis(new_word_name).size() > 0 and txt_word is not None \             # <<<<<<<<<<<<<<
 *                         and isinstance(txt_word, TxtWord) and not txt_word.isCode():
 *                     result.addWord(Word(new_word_name))
 */
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":367
 *             return False
 *         for question_suffix in self.__questionSuffixList:
 *             if word_name.endswith(question_suffix):             # <<<<<<<<<<<<<<
 *                 new_word_name = word_name[0:word_name.rindex(question_suffix)]
 *                 txt_word = self.fsm.getDictionary().getWord(new_word_name)
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":366
 *         if self.fsm.morphologicalAnalysis(word_name).size() > 0:
 *             return False
 *         for question_suffix in self.__questionSuffixList:             # <<<<<<<<<<<<<<
 *             if word_name.endswith(question_suffix):
 *                 new_word_name = word_name[0:word_name.rindex(question_suffix)]
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":375
 *                     result.addWord(Word(question_suffix))
 *                     return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cpdef loadDictionaries(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":358
 *         return False
 * 
 *     cpdef bint forcedQuestionSuffixSplitCheck(self,             # <<<<<<<<<<<<<<
 *                                        Word word,
 *                                        Sentence result):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedQuestionSuffixSplitCheck", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_word_name);
  __Pyx_XDECREF(__pyx_v_new_word_name);
  __Pyx_XDECREF((PyObject *)__pyx_v_txt_word);
  __Pyx_XDECREF(__pyx_v_question_suffix);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_27forcedQuestionSuffixSplitCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_27forcedQuestionSuffixSplitCheck(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("forcedQuestionSuffixSplitCheck (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word,&__pyx_n_s_result,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("forcedQuestionSuffixSplitCheck", 1, 2, 2, 1); __PYX_ERR(0, 358, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "forcedQuestionSuffixSplitCheck") < 0)) __PYX_ERR(0, 358, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_result = ((struct __pyx_obj_6Corpus_8Sentence_Sentence *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("forcedQuestionSuffixSplitCheck", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 358, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedQuestionSuffixSplitCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 359, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6Corpus_8Sentence_Sentence, 1, "result", 0))) __PYX_ERR(0, 360, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_26forcedQuestionSuffixSplitCheck(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_word, __pyx_v_result);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_26forcedQuestionSuffixSplitCheck(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_6Corpus_8Sentence_Sentence *__pyx_v_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("forcedQuestionSuffixSplitCheck", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedQuestionSuffixSplitCheck(__pyx_v_self, __pyx_v_word, __pyx_v_result, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.forcedQuestionSuffixSplitCheck", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":377
 *         return False
 * 
 *     cpdef loadDictionaries(self):             # <<<<<<<<<<<<<<
 *         cdef str line
 *         cdef list items, lines
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_29loadDictionaries(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_loadDictionaries(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_line = 0;
  PyObject *__pyx_v_items = 0;
  PyObject *__pyx_v_lines = 0;
  PyObject *__pyx_v_input_file = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("loadDictionaries", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_loadDictionaries); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_29loadDictionaries)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":380
 *         cdef str line
 *         cdef list items, lines
 *         input_file = open(pkg_resources.resource_filename(__name__, 'data/merged.txt'), "r", encoding="utf8")             # <<<<<<<<<<<<<<
 *         lines = input_file.readlines()
 *         for line in lines:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pkg_resources); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_resource_filename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_kp_u_data_merged_txt};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_kp_u_data_merged_txt};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_2);
    __Pyx_INCREF(__pyx_kp_u_data_merged_txt);
    __Pyx_GIVEREF(__pyx_kp_u_data_merged_txt);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_kp_u_data_merged_txt);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_r);
  __Pyx_GIVEREF(__pyx_n_u_r);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_u_r);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_encoding, __pyx_n_u_utf8) < 0) __PYX_ERR(0, 380, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_input_file = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":381
 *         cdef list items, lines
 *         input_file = open(pkg_resources.resource_filename(__name__, 'data/merged.txt'), "r", encoding="utf8")
 *         lines = input_file.readlines()             # <<<<<<<<<<<<<<
 *         for line in lines:
 *             items = line.strip().split(" ")
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_input_file, __pyx_n_s_readlines); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_6 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 381, __pyx_L1_error)
  __pyx_v_lines = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":382
 *         input_file = open(pkg_resources.resource_filename(__name__, 'data/merged.txt'), "r", encoding="utf8")
 *         lines = input_file.readlines()
 *         for line in lines:             # <<<<<<<<<<<<<<
 *             items = line.strip().split(" ")
 *             self.__merged_words[items[0] + " " + items[1]] = items[2]
 */
  if (unlikely(__pyx_v_lines == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 382, __pyx_L1_error)
  }
  __pyx_t_6 = __pyx_v_lines; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
  for (;;) {
    if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 382, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_line, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":383
 *         lines = input_file.readlines()
 *         for line in lines:
 *             items = line.strip().split(" ")             # <<<<<<<<<<<<<<
 *             self.__merged_words[items[0] + " " + items[1]] = items[2]
 *         input_file.close()
 */
    __pyx_t_3 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_strip, __pyx_v_line); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_split); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u__2) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u__2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_items, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":384
 *         for line in lines:
 *             items = line.strip().split(" ")
 *             self.__merged_words[items[0] + " " + items[1]] = items[2]             # <<<<<<<<<<<<<<
 *         input_file.close()
 *         input_file = open(pkg_resources.resource_filename(__name__, 'data/split.txt'), "r", encoding="utf8")
 */
    if (unlikely(__pyx_v_items == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 384, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_items, 2, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_self->__pyx___merged_words == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 384, __pyx_L1_error)
    }
    if (unlikely(__pyx_v_items == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 384, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_items, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_kp_u__2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_v_items == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 384, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_items, 1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(PyDict_SetItem(__pyx_v_self->__pyx___merged_words, __pyx_t_4, __pyx_t_1) < 0)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":382
 *         input_file = open(pkg_resources.resource_filename(__name__, 'data/merged.txt'), "r", encoding="utf8")
 *         lines = input_file.readlines()
 *         for line in lines:             # <<<<<<<<<<<<<<
 *             items = line.strip().split(" ")
 *             self.__merged_words[items[0] + " " + items[1]] = items[2]
 */
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":385
 *             items = line.strip().split(" ")
 *             self.__merged_words[items[0] + " " + items[1]] = items[2]
 *         input_file.close()             # <<<<<<<<<<<<<<
 *         input_file = open(pkg_resources.resource_filename(__name__, 'data/split.txt'), "r", encoding="utf8")
 *         lines = input_file.readlines()
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_input_file, __pyx_n_s_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":386
 *             self.__merged_words[items[0] + " " + items[1]] = items[2]
 *         input_file.close()
 *         input_file = open(pkg_resources.resource_filename(__name__, 'data/split.txt'), "r", encoding="utf8")             # <<<<<<<<<<<<<<
 *         lines = input_file.readlines()
 *         for line in lines:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pkg_resources); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_resource_filename); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_1, __pyx_kp_u_data_split_txt};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_1, __pyx_kp_u_data_split_txt};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_5, __pyx_t_1);
    __Pyx_INCREF(__pyx_kp_u_data_split_txt);
    __Pyx_GIVEREF(__pyx_kp_u_data_split_txt);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_5, __pyx_kp_u_data_split_txt);
    __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
  __Pyx_INCREF(__pyx_n_u_r);
  __Pyx_GIVEREF(__pyx_n_u_r);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_n_u_r);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_encoding, __pyx_n_u_utf8) < 0) __PYX_ERR(0, 386, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF_SET(__pyx_v_input_file, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":387
 *         input_file.close()
 *         input_file = open(pkg_resources.resource_filename(__name__, 'data/split.txt'), "r", encoding="utf8")
 *         lines = input_file.readlines()             # <<<<<<<<<<<<<<
 *         for line in lines:
 *             items = line.strip().split(" ")
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_input_file, __pyx_n_s_readlines); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_lines, ((PyObject*)__pyx_t_3));
  __pyx_t_3 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":388
 *         input_file = open(pkg_resources.resource_filename(__name__, 'data/split.txt'), "r", encoding="utf8")
 *         lines = input_file.readlines()
 *         for line in lines:             # <<<<<<<<<<<<<<
 *             items = line.strip().split(" ")
 *             self.__split_words[items[0]] = items[1] + " " + items[2]
 */
  if (unlikely(__pyx_v_lines == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 388, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_v_lines; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
  for (;;) {
    if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 388, __pyx_L1_error)
    #else
    __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_line, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":389
 *         lines = input_file.readlines()
 *         for line in lines:
 *             items = line.strip().split(" ")             # <<<<<<<<<<<<<<
 *             self.__split_words[items[0]] = items[1] + " " + items[2]
 *         input_file.close()
 */
    __pyx_t_4 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_strip, __pyx_v_line); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_split); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_kp_u__2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_u__2);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_items, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":390
 *         for line in lines:
 *             items = line.strip().split(" ")
 *             self.__split_words[items[0]] = items[1] + " " + items[2]             # <<<<<<<<<<<<<<
 *         input_file.close()
 * 
 */
    if (unlikely(__pyx_v_items == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 390, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_items, 1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyNumber_Add(__pyx_t_6, __pyx_kp_u__2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v_items == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 390, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_items, 2, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v_self->__pyx___split_words == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 390, __pyx_L1_error)
    }
    if (unlikely(__pyx_v_items == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 390, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_items, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(PyDict_SetItem(__pyx_v_self->__pyx___split_words, __pyx_t_6, __pyx_t_4) < 0)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":388
 *         input_file = open(pkg_resources.resource_filename(__name__, 'data/split.txt'), "r", encoding="utf8")
 *         lines = input_file.readlines()
 *         for line in lines:             # <<<<<<<<<<<<<<
 *             items = line.strip().split(" ")
 *             self.__split_words[items[0]] = items[1] + " " + items[2]
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":391
 *             items = line.strip().split(" ")
 *             self.__split_words[items[0]] = items[1] + " " + items[2]
 *         input_file.close()             # <<<<<<<<<<<<<<
 * 
 *     cpdef str getCorrectForm(self, str wordName, dict dictionary):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_input_file, __pyx_n_s_close); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":377
 *         return False
 * 
 *     cpdef loadDictionaries(self):             # <<<<<<<<<<<<<<
 *         cdef str line
 *         cdef list items, lines
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.loadDictionaries", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_lines);
  __Pyx_XDECREF(__pyx_v_input_file);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_29loadDictionaries(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_29loadDictionaries(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("loadDictionaries (wrapper)", 0);
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_28loadDictionaries(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_28loadDictionaries(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("loadDictionaries", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_loadDictionaries(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.loadDictionaries", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":393
 *         input_file.close()
 * 
 *     cpdef str getCorrectForm(self, str wordName, dict dictionary):             # <<<<<<<<<<<<<<
 *         if wordName in dictionary:
 *             return dictionary[wordName]
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_31getCorrectForm(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_getCorrectForm(CYTHON_UNUSED struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v_wordName, PyObject *__pyx_v_dictionary, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getCorrectForm", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getCorrectForm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_31getCorrectForm)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_wordName, __pyx_v_dictionary};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_wordName, __pyx_v_dictionary};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 393, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_wordName);
          __Pyx_GIVEREF(__pyx_v_wordName);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_wordName);
          __Pyx_INCREF(__pyx_v_dictionary);
          __Pyx_GIVEREF(__pyx_v_dictionary);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_dictionary);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 393, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":394
 * 
 *     cpdef str getCorrectForm(self, str wordName, dict dictionary):
 *         if wordName in dictionary:             # <<<<<<<<<<<<<<
 *             return dictionary[wordName]
 *         return ""
 */
  if (unlikely(__pyx_v_dictionary == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 394, __pyx_L1_error)
  }
  __pyx_t_7 = (__Pyx_PyDict_ContainsTF(__pyx_v_wordName, __pyx_v_dictionary, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 394, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "SpellChecker/SimpleSpellChecker.pyx":395
 *     cpdef str getCorrectForm(self, str wordName, dict dictionary):
 *         if wordName in dictionary:
 *             return dictionary[wordName]             # <<<<<<<<<<<<<<
 *         return ""
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_dictionary == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 395, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_dictionary, __pyx_v_wordName); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 395, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "SpellChecker/SimpleSpellChecker.pyx":394
 * 
 *     cpdef str getCorrectForm(self, str wordName, dict dictionary):
 *         if wordName in dictionary:             # <<<<<<<<<<<<<<
 *             return dictionary[wordName]
 *         return ""
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":396
 *         if wordName in dictionary:
 *             return dictionary[wordName]
 *         return ""             # <<<<<<<<<<<<<<
 * 
 *     cpdef list mergedCandidatesList(self, Word previousWord, Word word, Word nextWord):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_r = __pyx_kp_u_;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":393
 *         input_file.close()
 * 
 *     cpdef str getCorrectForm(self, str wordName, dict dictionary):             # <<<<<<<<<<<<<<
 *         if wordName in dictionary:
 *             return dictionary[wordName]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.getCorrectForm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_31getCorrectForm(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_31getCorrectForm(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_wordName = 0;
  PyObject *__pyx_v_dictionary = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getCorrectForm (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_wordName,&__pyx_n_s_dictionary,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_wordName)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dictionary)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("getCorrectForm", 1, 2, 2, 1); __PYX_ERR(0, 393, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "getCorrectForm") < 0)) __PYX_ERR(0, 393, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_wordName = ((PyObject*)values[0]);
    __pyx_v_dictionary = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getCorrectForm", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 393, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.getCorrectForm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wordName), (&PyUnicode_Type), 1, "wordName", 1))) __PYX_ERR(0, 393, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dictionary), (&PyDict_Type), 1, "dictionary", 1))) __PYX_ERR(0, 393, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_30getCorrectForm(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_wordName, __pyx_v_dictionary);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_30getCorrectForm(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v_wordName, PyObject *__pyx_v_dictionary) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getCorrectForm", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_getCorrectForm(__pyx_v_self, __pyx_v_wordName, __pyx_v_dictionary, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.getCorrectForm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":398
 *         return ""
 * 
 *     cpdef list mergedCandidatesList(self, Word previousWord, Word word, Word nextWord):             # <<<<<<<<<<<<<<
 *         cdef list merged_candidates
 *         cdef Candidate backward_merge_candidate, forward_merge_candidate
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_33mergedCandidatesList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_mergedCandidatesList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_merged_candidates = 0;
  struct __pyx_obj_12SpellChecker_9Candidate_Candidate *__pyx_v_backward_merge_candidate = 0;
  struct __pyx_obj_12SpellChecker_9Candidate_Candidate *__pyx_v_forward_merge_candidate = 0;
  struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *__pyx_v_fsm_parse_list = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mergedCandidatesList", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mergedCandidatesList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_33mergedCandidatesList)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_previousWord), ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_nextWord)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_4, ((PyObject *)__pyx_v_previousWord), ((PyObject *)__pyx_v_word), ((PyObject *)__pyx_v_nextWord)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 398, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_previousWord));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_previousWord));
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_previousWord));
          __Pyx_INCREF(((PyObject *)__pyx_v_word));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_word));
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_word));
          __Pyx_INCREF(((PyObject *)__pyx_v_nextWord));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_nextWord));
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, ((PyObject *)__pyx_v_nextWord));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 398, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":402
 *         cdef Candidate backward_merge_candidate, forward_merge_candidate
 *         cdef FsmParseList fsm_parse_list
 *         merged_candidates = []             # <<<<<<<<<<<<<<
 *         backward_merge_candidate = None
 *         if previousWord is not None:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_merged_candidates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":403
 *         cdef FsmParseList fsm_parse_list
 *         merged_candidates = []
 *         backward_merge_candidate = None             # <<<<<<<<<<<<<<
 *         if previousWord is not None:
 *             backward_merge_candidate = Candidate(previousWord.getName() + word.getName(), Operator.BACKWARD_MERGE)
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_backward_merge_candidate = ((struct __pyx_obj_12SpellChecker_9Candidate_Candidate *)Py_None);

  /* "SpellChecker/SimpleSpellChecker.pyx":404
 *         merged_candidates = []
 *         backward_merge_candidate = None
 *         if previousWord is not None:             # <<<<<<<<<<<<<<
 *             backward_merge_candidate = Candidate(previousWord.getName() + word.getName(), Operator.BACKWARD_MERGE)
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(backward_merge_candidate.getName())
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_previousWord) != Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "SpellChecker/SimpleSpellChecker.pyx":405
 *         backward_merge_candidate = None
 *         if previousWord is not None:
 *             backward_merge_candidate = Candidate(previousWord.getName() + word.getName(), Operator.BACKWARD_MERGE)             # <<<<<<<<<<<<<<
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(backward_merge_candidate.getName())
 *             if fsm_parse_list.size() != 0:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_previousWord->__pyx_vtab)->getName(__pyx_v_previousWord, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Operator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BACKWARD_MERGE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12SpellChecker_9Candidate_Candidate), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_backward_merge_candidate, ((struct __pyx_obj_12SpellChecker_9Candidate_Candidate *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":406
 *         if previousWord is not None:
 *             backward_merge_candidate = Candidate(previousWord.getName() + word.getName(), Operator.BACKWARD_MERGE)
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(backward_merge_candidate.getName())             # <<<<<<<<<<<<<<
 *             if fsm_parse_list.size() != 0:
 *                 merged_candidates.append(backward_merge_candidate)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_12SpellChecker_9Candidate_Candidate *)__pyx_v_backward_merge_candidate->__pyx_base.__pyx_vtab)->__pyx_base.getName(((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_v_backward_merge_candidate), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList))))) __PYX_ERR(0, 406, __pyx_L1_error)
    __pyx_v_fsm_parse_list = ((struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":407
 *             backward_merge_candidate = Candidate(previousWord.getName() + word.getName(), Operator.BACKWARD_MERGE)
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(backward_merge_candidate.getName())
 *             if fsm_parse_list.size() != 0:             # <<<<<<<<<<<<<<
 *                 merged_candidates.append(backward_merge_candidate)
 *         if nextWord is not None:
 */
    __pyx_t_8 = ((((struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_v_fsm_parse_list->__pyx_vtab)->size(__pyx_v_fsm_parse_list, 0) != 0) != 0);
    if (__pyx_t_8) {

      /* "SpellChecker/SimpleSpellChecker.pyx":408
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(backward_merge_candidate.getName())
 *             if fsm_parse_list.size() != 0:
 *                 merged_candidates.append(backward_merge_candidate)             # <<<<<<<<<<<<<<
 *         if nextWord is not None:
 *             forward_merge_candidate = Candidate(word.getName() + nextWord.getName(), Operator.FORWARD_MERGE)
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_merged_candidates, ((PyObject *)__pyx_v_backward_merge_candidate)); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 408, __pyx_L1_error)

      /* "SpellChecker/SimpleSpellChecker.pyx":407
 *             backward_merge_candidate = Candidate(previousWord.getName() + word.getName(), Operator.BACKWARD_MERGE)
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(backward_merge_candidate.getName())
 *             if fsm_parse_list.size() != 0:             # <<<<<<<<<<<<<<
 *                 merged_candidates.append(backward_merge_candidate)
 *         if nextWord is not None:
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":404
 *         merged_candidates = []
 *         backward_merge_candidate = None
 *         if previousWord is not None:             # <<<<<<<<<<<<<<
 *             backward_merge_candidate = Candidate(previousWord.getName() + word.getName(), Operator.BACKWARD_MERGE)
 *             fsm_parse_list = self.fsm.morphologicalAnalysis(backward_merge_candidate.getName())
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":409
 *             if fsm_parse_list.size() != 0:
 *                 merged_candidates.append(backward_merge_candidate)
 *         if nextWord is not None:             # <<<<<<<<<<<<<<
 *             forward_merge_candidate = Candidate(word.getName() + nextWord.getName(), Operator.FORWARD_MERGE)
 *             if backward_merge_candidate is None or backward_merge_candidate.getName() != forward_merge_candidate.getName():
 */
  __pyx_t_8 = (((PyObject *)__pyx_v_nextWord) != Py_None);
  __pyx_t_7 = (__pyx_t_8 != 0);
  if (__pyx_t_7) {

    /* "SpellChecker/SimpleSpellChecker.pyx":410
 *                 merged_candidates.append(backward_merge_candidate)
 *         if nextWord is not None:
 *             forward_merge_candidate = Candidate(word.getName() + nextWord.getName(), Operator.FORWARD_MERGE)             # <<<<<<<<<<<<<<
 *             if backward_merge_candidate is None or backward_merge_candidate.getName() != forward_merge_candidate.getName():
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(forward_merge_candidate.getName())
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_nextWord->__pyx_vtab)->getName(__pyx_v_nextWord, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Operator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_FORWARD_MERGE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12SpellChecker_9Candidate_Candidate), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_forward_merge_candidate = ((struct __pyx_obj_12SpellChecker_9Candidate_Candidate *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":411
 *         if nextWord is not None:
 *             forward_merge_candidate = Candidate(word.getName() + nextWord.getName(), Operator.FORWARD_MERGE)
 *             if backward_merge_candidate is None or backward_merge_candidate.getName() != forward_merge_candidate.getName():             # <<<<<<<<<<<<<<
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(forward_merge_candidate.getName())
 *                 if fsm_parse_list.size() != 0:
 */
    __pyx_t_8 = (((PyObject *)__pyx_v_backward_merge_candidate) == Py_None);
    __pyx_t_10 = (__pyx_t_8 != 0);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_7 = __pyx_t_10;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_2 = ((struct __pyx_vtabstruct_12SpellChecker_9Candidate_Candidate *)__pyx_v_backward_merge_candidate->__pyx_base.__pyx_vtab)->__pyx_base.getName(((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_v_backward_merge_candidate), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = ((struct __pyx_vtabstruct_12SpellChecker_9Candidate_Candidate *)__pyx_v_forward_merge_candidate->__pyx_base.__pyx_vtab)->__pyx_base.getName(((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_v_forward_merge_candidate), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_t_1, Py_NE)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = (__pyx_t_10 != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_7) {

      /* "SpellChecker/SimpleSpellChecker.pyx":412
 *             forward_merge_candidate = Candidate(word.getName() + nextWord.getName(), Operator.FORWARD_MERGE)
 *             if backward_merge_candidate is None or backward_merge_candidate.getName() != forward_merge_candidate.getName():
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(forward_merge_candidate.getName())             # <<<<<<<<<<<<<<
 *                 if fsm_parse_list.size() != 0:
 *                     merged_candidates.append(forward_merge_candidate)
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_12SpellChecker_9Candidate_Candidate *)__pyx_v_forward_merge_candidate->__pyx_base.__pyx_vtab)->__pyx_base.getName(((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_v_forward_merge_candidate), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList))))) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_fsm_parse_list, ((struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":413
 *             if backward_merge_candidate is None or backward_merge_candidate.getName() != forward_merge_candidate.getName():
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(forward_merge_candidate.getName())
 *                 if fsm_parse_list.size() != 0:             # <<<<<<<<<<<<<<
 *                     merged_candidates.append(forward_merge_candidate)
 *         return merged_candidates
 */
      __pyx_t_7 = ((((struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_v_fsm_parse_list->__pyx_vtab)->size(__pyx_v_fsm_parse_list, 0) != 0) != 0);
      if (__pyx_t_7) {

        /* "SpellChecker/SimpleSpellChecker.pyx":414
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(forward_merge_candidate.getName())
 *                 if fsm_parse_list.size() != 0:
 *                     merged_candidates.append(forward_merge_candidate)             # <<<<<<<<<<<<<<
 *         return merged_candidates
 * 
 */
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_merged_candidates, ((PyObject *)__pyx_v_forward_merge_candidate)); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 414, __pyx_L1_error)

        /* "SpellChecker/SimpleSpellChecker.pyx":413
 *             if backward_merge_candidate is None or backward_merge_candidate.getName() != forward_merge_candidate.getName():
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(forward_merge_candidate.getName())
 *                 if fsm_parse_list.size() != 0:             # <<<<<<<<<<<<<<
 *                     merged_candidates.append(forward_merge_candidate)
 *         return merged_candidates
 */
      }

      /* "SpellChecker/SimpleSpellChecker.pyx":411
 *         if nextWord is not None:
 *             forward_merge_candidate = Candidate(word.getName() + nextWord.getName(), Operator.FORWARD_MERGE)
 *             if backward_merge_candidate is None or backward_merge_candidate.getName() != forward_merge_candidate.getName():             # <<<<<<<<<<<<<<
 *                 fsm_parse_list = self.fsm.morphologicalAnalysis(forward_merge_candidate.getName())
 *                 if fsm_parse_list.size() != 0:
 */
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":409
 *             if fsm_parse_list.size() != 0:
 *                 merged_candidates.append(backward_merge_candidate)
 *         if nextWord is not None:             # <<<<<<<<<<<<<<
 *             forward_merge_candidate = Candidate(word.getName() + nextWord.getName(), Operator.FORWARD_MERGE)
 *             if backward_merge_candidate is None or backward_merge_candidate.getName() != forward_merge_candidate.getName():
 */
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":415
 *                 if fsm_parse_list.size() != 0:
 *                     merged_candidates.append(forward_merge_candidate)
 *         return merged_candidates             # <<<<<<<<<<<<<<
 * 
 *     cpdef list splitCandidatesList(self, Word word):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_merged_candidates);
  __pyx_r = __pyx_v_merged_candidates;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":398
 *         return ""
 * 
 *     cpdef list mergedCandidatesList(self, Word previousWord, Word word, Word nextWord):             # <<<<<<<<<<<<<<
 *         cdef list merged_candidates
 *         cdef Candidate backward_merge_candidate, forward_merge_candidate
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.mergedCandidatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_merged_candidates);
  __Pyx_XDECREF((PyObject *)__pyx_v_backward_merge_candidate);
  __Pyx_XDECREF((PyObject *)__pyx_v_forward_merge_candidate);
  __Pyx_XDECREF((PyObject *)__pyx_v_fsm_parse_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_33mergedCandidatesList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_33mergedCandidatesList(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord = 0;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word = 0;
  struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mergedCandidatesList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_previousWord,&__pyx_n_s_word,&__pyx_n_s_nextWord,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_previousWord)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_word)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mergedCandidatesList", 1, 3, 3, 1); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nextWord)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mergedCandidatesList", 1, 3, 3, 2); __PYX_ERR(0, 398, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mergedCandidatesList") < 0)) __PYX_ERR(0, 398, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_previousWord = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[0]);
    __pyx_v_word = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[1]);
    __pyx_v_nextWord = ((struct __pyx_obj_10Dictionary_4Word_Word *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mergedCandidatesList", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 398, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.mergedCandidatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_previousWord), __pyx_ptype_10Dictionary_4Word_Word, 1, "previousWord", 0))) __PYX_ERR(0, 398, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 398, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_nextWord), __pyx_ptype_10Dictionary_4Word_Word, 1, "nextWord", 0))) __PYX_ERR(0, 398, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_32mergedCandidatesList(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), __pyx_v_previousWord, __pyx_v_word, __pyx_v_nextWord);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_32mergedCandidatesList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_previousWord, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_nextWord) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mergedCandidatesList", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_mergedCandidatesList(__pyx_v_self, __pyx_v_previousWord, __pyx_v_word, __pyx_v_nextWord, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.mergedCandidatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":417
 *         return merged_candidates
 * 
 *     cpdef list splitCandidatesList(self, Word word):             # <<<<<<<<<<<<<<
 *         cdef list split_candidates
 *         cdef int i
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_35splitCandidatesList(PyObject *__pyx_v_self, PyObject *__pyx_v_word); /*proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_splitCandidatesList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_split_candidates = 0;
  int __pyx_v_i;
  PyObject *__pyx_v_first_part = 0;
  PyObject *__pyx_v_second_part = 0;
  struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *__pyx_v_fsm_parse_list_first = 0;
  struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *__pyx_v_fsm_parse_list_second = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("splitCandidatesList", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_splitCandidatesList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_35splitCandidatesList)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_word)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_word));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 417, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 417, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":422
 *         cdef str first_part, second_part
 *         cdef FsmParseList fsm_parse_list_first, fsm_parse_list_second
 *         split_candidates = []             # <<<<<<<<<<<<<<
 *         for i in range(4, len(word.getName()) - 3):
 *             first_part = word.getName()[0:i]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_split_candidates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":423
 *         cdef FsmParseList fsm_parse_list_first, fsm_parse_list_second
 *         split_candidates = []
 *         for i in range(4, len(word.getName()) - 3):             # <<<<<<<<<<<<<<
 *             first_part = word.getName()[0:i]
 *             second_part = word.getName()[i:]
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 423, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (__pyx_t_5 - 3);
  __pyx_t_5 = __pyx_t_6;
  for (__pyx_t_7 = 4; __pyx_t_7 < __pyx_t_5; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "SpellChecker/SimpleSpellChecker.pyx":424
 *         split_candidates = []
 *         for i in range(4, len(word.getName()) - 3):
 *             first_part = word.getName()[0:i]             # <<<<<<<<<<<<<<
 *             second_part = word.getName()[i:]
 *             fsm_parse_list_first = self.fsm.morphologicalAnalysis(first_part)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_t_1 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 424, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyUnicode_Substring(__pyx_t_1, 0, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_first_part, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":425
 *         for i in range(4, len(word.getName()) - 3):
 *             first_part = word.getName()[0:i]
 *             second_part = word.getName()[i:]             # <<<<<<<<<<<<<<
 *             fsm_parse_list_first = self.fsm.morphologicalAnalysis(first_part)
 *             fsm_parse_list_second = self.fsm.morphologicalAnalysis(second_part)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_t_2 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 425, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_t_2, __pyx_v_i, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_second_part, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":426
 *             first_part = word.getName()[0:i]
 *             second_part = word.getName()[i:]
 *             fsm_parse_list_first = self.fsm.morphologicalAnalysis(first_part)             # <<<<<<<<<<<<<<
 *             fsm_parse_list_second = self.fsm.morphologicalAnalysis(second_part)
 *             if fsm_parse_list_first.size() > 0 and fsm_parse_list_second.size() > 0:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_v_first_part, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList))))) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fsm_parse_list_first, ((struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":427
 *             second_part = word.getName()[i:]
 *             fsm_parse_list_first = self.fsm.morphologicalAnalysis(first_part)
 *             fsm_parse_list_second = self.fsm.morphologicalAnalysis(second_part)             # <<<<<<<<<<<<<<
 *             if fsm_parse_list_first.size() > 0 and fsm_parse_list_second.size() > 0:
 *                 split_candidates.append(Candidate(first_part + " " + second_part, Operator.SPLIT))
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_v_self->fsm->__pyx_vtab)->morphologicalAnalysis(__pyx_v_self->fsm, __pyx_v_second_part, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList))))) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fsm_parse_list_second, ((struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "SpellChecker/SimpleSpellChecker.pyx":428
 *             fsm_parse_list_first = self.fsm.morphologicalAnalysis(first_part)
 *             fsm_parse_list_second = self.fsm.morphologicalAnalysis(second_part)
 *             if fsm_parse_list_first.size() > 0 and fsm_parse_list_second.size() > 0:             # <<<<<<<<<<<<<<
 *                 split_candidates.append(Candidate(first_part + " " + second_part, Operator.SPLIT))
 *         return split_candidates
 */
    __pyx_t_9 = ((((struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_v_fsm_parse_list_first->__pyx_vtab)->size(__pyx_v_fsm_parse_list_first, 0) > 0) != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_9 = ((((struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList *)__pyx_v_fsm_parse_list_second->__pyx_vtab)->size(__pyx_v_fsm_parse_list_second, 0) > 0) != 0);
    __pyx_t_8 = __pyx_t_9;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_8) {

      /* "SpellChecker/SimpleSpellChecker.pyx":429
 *             fsm_parse_list_second = self.fsm.morphologicalAnalysis(second_part)
 *             if fsm_parse_list_first.size() > 0 and fsm_parse_list_second.size() > 0:
 *                 split_candidates.append(Candidate(first_part + " " + second_part, Operator.SPLIT))             # <<<<<<<<<<<<<<
 *         return split_candidates
 * 
 */
      __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_v_first_part, __pyx_kp_u__2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_t_1, __pyx_v_second_part); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Operator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_SPLIT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
      __pyx_t_2 = 0;
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_12SpellChecker_9Candidate_Candidate), __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_split_candidates, __pyx_t_3); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 429, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":428
 *             fsm_parse_list_first = self.fsm.morphologicalAnalysis(first_part)
 *             fsm_parse_list_second = self.fsm.morphologicalAnalysis(second_part)
 *             if fsm_parse_list_first.size() > 0 and fsm_parse_list_second.size() > 0:             # <<<<<<<<<<<<<<
 *                 split_candidates.append(Candidate(first_part + " " + second_part, Operator.SPLIT))
 *         return split_candidates
 */
    }
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":430
 *             if fsm_parse_list_first.size() > 0 and fsm_parse_list_second.size() > 0:
 *                 split_candidates.append(Candidate(first_part + " " + second_part, Operator.SPLIT))
 *         return split_candidates             # <<<<<<<<<<<<<<
 * 
 *     cpdef tuple getSplitPair(self, Word word):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_split_candidates);
  __pyx_r = __pyx_v_split_candidates;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":417
 *         return merged_candidates
 * 
 *     cpdef list splitCandidatesList(self, Word word):             # <<<<<<<<<<<<<<
 *         cdef list split_candidates
 *         cdef int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.splitCandidatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_split_candidates);
  __Pyx_XDECREF(__pyx_v_first_part);
  __Pyx_XDECREF(__pyx_v_second_part);
  __Pyx_XDECREF((PyObject *)__pyx_v_fsm_parse_list_first);
  __Pyx_XDECREF((PyObject *)__pyx_v_fsm_parse_list_second);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_35splitCandidatesList(PyObject *__pyx_v_self, PyObject *__pyx_v_word); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_35splitCandidatesList(PyObject *__pyx_v_self, PyObject *__pyx_v_word) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("splitCandidatesList (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 417, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_34splitCandidatesList(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_v_word));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_34splitCandidatesList(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("splitCandidatesList", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_splitCandidatesList(__pyx_v_self, __pyx_v_word, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.splitCandidatesList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "SpellChecker/SimpleSpellChecker.pyx":432
 *         return split_candidates
 * 
 *     cpdef tuple getSplitPair(self, Word word):             # <<<<<<<<<<<<<<
 *         cdef str key, value
 *         cdef int j
 */

static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_37getSplitPair(PyObject *__pyx_v_self, PyObject *__pyx_v_word); /*proto*/
static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_getSplitPair(CYTHON_UNUSED struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getSplitPair", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getSplitPair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_37getSplitPair)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_word)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_word));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyTuple_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 432, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "SpellChecker/SimpleSpellChecker.pyx":435
 *         cdef str key, value
 *         cdef int j
 *         key = ""             # <<<<<<<<<<<<<<
 *         j = 0
 *         while j < len(word.getName()):
 */
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_v_key = __pyx_kp_u_;

  /* "SpellChecker/SimpleSpellChecker.pyx":436
 *         cdef int j
 *         key = ""
 *         j = 0             # <<<<<<<<<<<<<<
 *         while j < len(word.getName()):
 *             if "0" <= word.getName()[j] <= "9":
 */
  __pyx_v_j = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":437
 *         key = ""
 *         j = 0
 *         while j < len(word.getName()):             # <<<<<<<<<<<<<<
 *             if "0" <= word.getName()[j] <= "9":
 *                 key = key + word.getName()[j]
 */
  while (1) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_t_1 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 437, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = ((__pyx_v_j < __pyx_t_5) != 0);
    if (!__pyx_t_6) break;

    /* "SpellChecker/SimpleSpellChecker.pyx":438
 *         j = 0
 *         while j < len(word.getName()):
 *             if "0" <= word.getName()[j] <= "9":             # <<<<<<<<<<<<<<
 *                 key = key + word.getName()[j]
 *             else:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_GetItemInt_Unicode(__pyx_t_1, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(__pyx_t_7 == (Py_UCS4)-1)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (48 <= __pyx_t_7);
    if (__pyx_t_6) {
      __pyx_t_6 = (__pyx_t_7 <= 57);
    }
    __pyx_t_8 = (__pyx_t_6 != 0);
    if (__pyx_t_8) {

      /* "SpellChecker/SimpleSpellChecker.pyx":439
 *         while j < len(word.getName()):
 *             if "0" <= word.getName()[j] <= "9":
 *                 key = key + word.getName()[j]             # <<<<<<<<<<<<<<
 *             else:
 *                 break
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 439, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_GetItemInt_Unicode(__pyx_t_1, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(__pyx_t_7 == (Py_UCS4)-1)) __PYX_ERR(0, 439, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyUnicode_FromOrdinal(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 439, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_v_key, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 439, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "SpellChecker/SimpleSpellChecker.pyx":438
 *         j = 0
 *         while j < len(word.getName()):
 *             if "0" <= word.getName()[j] <= "9":             # <<<<<<<<<<<<<<
 *                 key = key + word.getName()[j]
 *             else:
 */
      goto __pyx_L5;
    }

    /* "SpellChecker/SimpleSpellChecker.pyx":441
 *                 key = key + word.getName()[j]
 *             else:
 *                 break             # <<<<<<<<<<<<<<
 *             j = j + 1
 *         value = word.getName()[j:]
 */
    /*else*/ {
      goto __pyx_L4_break;
    }
    __pyx_L5:;

    /* "SpellChecker/SimpleSpellChecker.pyx":442
 *             else:
 *                 break
 *             j = j + 1             # <<<<<<<<<<<<<<
 *         value = word.getName()[j:]
 *         return key, value
 */
    __pyx_v_j = (__pyx_v_j + 1);
  }
  __pyx_L4_break:;

  /* "SpellChecker/SimpleSpellChecker.pyx":443
 *                 break
 *             j = j + 1
 *         value = word.getName()[j:]             # <<<<<<<<<<<<<<
 *         return key, value
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_10Dictionary_4Word_Word *)__pyx_v_word->__pyx_vtab)->getName(__pyx_v_word, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 443, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyUnicode_Substring(__pyx_t_2, __pyx_v_j, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_value = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":444
 *             j = j + 1
 *         value = word.getName()[j:]
 *         return key, value             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_key);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_value);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "SpellChecker/SimpleSpellChecker.pyx":432
 *         return split_candidates
 * 
 *     cpdef tuple getSplitPair(self, Word word):             # <<<<<<<<<<<<<<
 *         cdef str key, value
 *         cdef int j
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.getSplitPair", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_37getSplitPair(PyObject *__pyx_v_self, PyObject *__pyx_v_word); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_37getSplitPair(PyObject *__pyx_v_self, PyObject *__pyx_v_word) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getSplitPair (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_word), __pyx_ptype_10Dictionary_4Word_Word, 1, "word", 0))) __PYX_ERR(0, 432, __pyx_L1_error)
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_36getSplitPair(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), ((struct __pyx_obj_10Dictionary_4Word_Word *)__pyx_v_word));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_36getSplitPair(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, struct __pyx_obj_10Dictionary_4Word_Word *__pyx_v_word) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getSplitPair", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_getSplitPair(__pyx_v_self, __pyx_v_word, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.getSplitPair", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_39__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_39__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_38__reduce_cython__(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_38__reduce_cython__(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.__merged_words, self.__split_words, self.fsm)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx___merged_words);
  __Pyx_GIVEREF(__pyx_v_self->__pyx___merged_words);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx___merged_words);
  __Pyx_INCREF(__pyx_v_self->__pyx___split_words);
  __Pyx_GIVEREF(__pyx_v_self->__pyx___split_words);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->__pyx___split_words);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->fsm));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->fsm));
  PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->fsm));
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.__merged_words, self.__split_words, self.fsm)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.__merged_words, self.__split_words, self.fsm)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.__merged_words is not None or self.__split_words is not None or self.fsm is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.__merged_words, self.__split_words, self.fsm)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.__merged_words is not None or self.__split_words is not None or self.fsm is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_SimpleSpellChecker, (type(self), 0x162cce2, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->__pyx___merged_words != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx___split_words != ((PyObject*)Py_None));
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (((PyObject *)__pyx_v_self->fsm) != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    __pyx_t_3 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.__merged_words is not None or self.__split_words is not None or self.fsm is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SimpleSpellChecker, (type(self), 0x162cce2, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.__merged_words is not None or self.__split_words is not None or self.fsm is not None
 *     if use_setstate:
 *         return __pyx_unpickle_SimpleSpellChecker, (type(self), 0x162cce2, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_SimpleSpellChecker, (type(self), 0x162cce2, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_SimpleSpellChecke); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_23252194);
    __Pyx_GIVEREF(__pyx_int_23252194);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_23252194);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.__merged_words is not None or self.__split_words is not None or self.fsm is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_SimpleSpellChecker, (type(self), 0x162cce2, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_SimpleSpellChecker, (type(self), 0x162cce2, None), state
 *     else:
 *         return __pyx_unpickle_SimpleSpellChecker, (type(self), 0x162cce2, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SimpleSpellChecker__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_SimpleSpellChecke); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_23252194);
    __Pyx_GIVEREF(__pyx_int_23252194);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_23252194);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SimpleSpellChecker, (type(self), 0x162cce2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SimpleSpellChecker__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_41__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_41__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_40__setstate_cython__(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_40__setstate_cython__(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_SimpleSpellChecker, (type(self), 0x162cce2, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_SimpleSpellChecker__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_12SpellChecker_18SimpleSpellChecker___pyx_unpickle_SimpleSpellChecker__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_SimpleSpellChecker, (type(self), 0x162cce2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_SimpleSpellChecker__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.SimpleSpellChecker.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_SimpleSpellChecker(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_1__pyx_unpickle_SimpleSpellChecker(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_12SpellChecker_18SimpleSpellChecker_1__pyx_unpickle_SimpleSpellChecker = {"__pyx_unpickle_SimpleSpellChecker", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_1__pyx_unpickle_SimpleSpellChecker, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12SpellChecker_18SimpleSpellChecker_1__pyx_unpickle_SimpleSpellChecker(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_SimpleSpellChecker (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SimpleSpellChecker", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SimpleSpellChecker", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_SimpleSpellChecker") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_SimpleSpellChecker", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.__pyx_unpickle_SimpleSpellChecker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12SpellChecker_18SimpleSpellChecker___pyx_unpickle_SimpleSpellChecker(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12SpellChecker_18SimpleSpellChecker___pyx_unpickle_SimpleSpellChecker(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_SimpleSpellChecker", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x162cce2:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x162cce2 = (__merged_words, __split_words, fsm))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x162cce2) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x162cce2:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x162cce2 = (__merged_words, __split_words, fsm))" % __pyx_checksum)
 *     __pyx_result = SimpleSpellChecker.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x162cce2:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x162cce2 = (__merged_words, __split_words, fsm))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = SimpleSpellChecker.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x16, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x162cce2:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x162cce2 = (__merged_words, __split_words, fsm))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x162cce2 = (__merged_words, __split_words, fsm))" % __pyx_checksum)
 *     __pyx_result = SimpleSpellChecker.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SimpleSpellChecker__set_state(<SimpleSpellChecker> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x162cce2 = (__merged_words, __split_words, fsm))" % __pyx_checksum)
 *     __pyx_result = SimpleSpellChecker.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SimpleSpellChecker__set_state(<SimpleSpellChecker> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = SimpleSpellChecker.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SimpleSpellChecker__set_state(<SimpleSpellChecker> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_SimpleSpellChecker__set_state(SimpleSpellChecker __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_12SpellChecker_18SimpleSpellChecker___pyx_unpickle_SimpleSpellChecker__set_state(((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x162cce2 = (__merged_words, __split_words, fsm))" % __pyx_checksum)
 *     __pyx_result = SimpleSpellChecker.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_SimpleSpellChecker__set_state(<SimpleSpellChecker> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_SimpleSpellChecker__set_state(<SimpleSpellChecker> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_SimpleSpellChecker__set_state(SimpleSpellChecker __pyx_result, tuple __pyx_state):
 *     __pyx_result.__merged_words = __pyx_state[0]; __pyx_result.__split_words = __pyx_state[1]; __pyx_result.fsm = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_SimpleSpellChecker(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.__pyx_unpickle_SimpleSpellChecker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_SimpleSpellChecker__set_state(<SimpleSpellChecker> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SimpleSpellChecker__set_state(SimpleSpellChecker __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.__merged_words = __pyx_state[0]; __pyx_result.__split_words = __pyx_state[1]; __pyx_result.fsm = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_12SpellChecker_18SimpleSpellChecker___pyx_unpickle_SimpleSpellChecker__set_state(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_SimpleSpellChecker__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_SimpleSpellChecker__set_state(SimpleSpellChecker __pyx_result, tuple __pyx_state):
 *     __pyx_result.__merged_words = __pyx_state[0]; __pyx_result.__split_words = __pyx_state[1]; __pyx_result.fsm = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx___merged_words);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx___merged_words);
  __pyx_v___pyx_result->__pyx___merged_words = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx___split_words);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx___split_words);
  __pyx_v___pyx_result->__pyx___split_words = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fsm);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->fsm));
  __pyx_v___pyx_result->fsm = ((struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_SimpleSpellChecker__set_state(SimpleSpellChecker __pyx_result, tuple __pyx_state):
 *     __pyx_result.__merged_words = __pyx_state[0]; __pyx_result.__split_words = __pyx_state[1]; __pyx_result.fsm = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 3) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.__merged_words = __pyx_state[0]; __pyx_result.__split_words = __pyx_state[1]; __pyx_result.fsm = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_SimpleSpellChecker__set_state(SimpleSpellChecker __pyx_result, tuple __pyx_state):
 *     __pyx_result.__merged_words = __pyx_state[0]; __pyx_result.__split_words = __pyx_state[1]; __pyx_result.fsm = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_SimpleSpellChecker__set_state(<SimpleSpellChecker> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_SimpleSpellChecker__set_state(SimpleSpellChecker __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.__merged_words = __pyx_state[0]; __pyx_result.__split_words = __pyx_state[1]; __pyx_result.fsm = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("SpellChecker.SimpleSpellChecker.__pyx_unpickle_SimpleSpellChecker__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker;

static PyObject *__pyx_tp_new_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *p;
  PyObject *o = __pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_12SpellChecker_12SpellChecker_SpellChecker*)__pyx_vtabptr_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker;
  p->fsm = ((struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)Py_None); Py_INCREF(Py_None);
  p->__pyx___merged_words = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->__pyx___split_words = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker(PyObject *o) {
  struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *p = (struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->fsm);
  Py_CLEAR(p->__pyx___merged_words);
  Py_CLEAR(p->__pyx___split_words);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker)) __pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker);
}

static int __pyx_tp_traverse_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *p = (struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)o;
  e = ((likely(__pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker)) ? ((__pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker->tp_traverse) ? __pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker)); if (e) return e;
  if (p->fsm) {
    e = (*v)(((PyObject *)p->fsm), a); if (e) return e;
  }
  if (p->__pyx___merged_words) {
    e = (*v)(p->__pyx___merged_words, a); if (e) return e;
  }
  if (p->__pyx___split_words) {
    e = (*v)(p->__pyx___split_words, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *p = (struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *)o;
  if (likely(__pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker)) { if (__pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker->tp_clear) __pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker);
  tmp = ((PyObject*)p->fsm);
  p->fsm = ((struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx___merged_words);
  p->__pyx___merged_words = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx___split_words);
  p->__pyx___split_words = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker[] = {
  {"__generateCandidateList", (PyCFunction)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_3__generateCandidateList, METH_O, __pyx_doc_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_2__generateCandidateList},
  {"candidateList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_5candidateList, METH_VARARGS|METH_KEYWORDS, __pyx_doc_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_4candidateList},
  {"spellCheck", (PyCFunction)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_7spellCheck, METH_O, __pyx_doc_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_6spellCheck},
  {"forcedMisspellCheck", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_9forcedMisspellCheck, METH_VARARGS|METH_KEYWORDS, 0},
  {"forcedBackwardMergeCheck", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_11forcedBackwardMergeCheck, METH_VARARGS|METH_KEYWORDS, 0},
  {"forcedForwardMergeCheck", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_13forcedForwardMergeCheck, METH_VARARGS|METH_KEYWORDS, 0},
  {"addSplitWords", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_15addSplitWords, METH_VARARGS|METH_KEYWORDS, 0},
  {"forcedSplitCheck", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_17forcedSplitCheck, METH_VARARGS|METH_KEYWORDS, 0},
  {"forcedShortcutCheck", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_19forcedShortcutCheck, METH_VARARGS|METH_KEYWORDS, 0},
  {"forcedDeDaSplitCheck", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_21forcedDeDaSplitCheck, METH_VARARGS|METH_KEYWORDS, 0},
  {"forcedSuffixMergeCheck", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_23forcedSuffixMergeCheck, METH_VARARGS|METH_KEYWORDS, 0},
  {"forcedHyphenMergeCheck", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_25forcedHyphenMergeCheck, METH_VARARGS|METH_KEYWORDS, 0},
  {"forcedQuestionSuffixSplitCheck", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_27forcedQuestionSuffixSplitCheck, METH_VARARGS|METH_KEYWORDS, 0},
  {"loadDictionaries", (PyCFunction)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_29loadDictionaries, METH_NOARGS, 0},
  {"getCorrectForm", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_31getCorrectForm, METH_VARARGS|METH_KEYWORDS, 0},
  {"mergedCandidatesList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_33mergedCandidatesList, METH_VARARGS|METH_KEYWORDS, 0},
  {"splitCandidatesList", (PyCFunction)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_35splitCandidatesList, METH_O, 0},
  {"getSplitPair", (PyCFunction)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_37getSplitPair, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_39__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_41__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker = {
  PyVarObject_HEAD_INIT(0, 0)
  "SpellChecker.SimpleSpellChecker.SimpleSpellChecker", /*tp_name*/
  sizeof(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker, /*tp_traverse*/
  __pyx_tp_clear_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_SimpleSpellChecker(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_SimpleSpellChecker},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "SimpleSpellChecker",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
  {&__pyx_kp_u_0_9, __pyx_k_0_9, sizeof(__pyx_k_0_9), 0, 1, 0, 0},
  {&__pyx_kp_u_1_9_0_9_0_0_9, __pyx_k_1_9_0_9_0_0_9, sizeof(__pyx_k_1_9_0_9_0_0_9), 0, 1, 0, 0},
  {&__pyx_kp_u_1_9_0_9_0_2_0_0_9, __pyx_k_1_9_0_9_0_2_0_0_9, sizeof(__pyx_k_1_9_0_9_0_2_0_0_9), 0, 1, 0, 0},
  {&__pyx_n_s_BACKWARD_MERGE, __pyx_k_BACKWARD_MERGE, sizeof(__pyx_k_BACKWARD_MERGE), 0, 0, 1, 1},
  {&__pyx_n_s_FORWARD_MERGE, __pyx_k_FORWARD_MERGE, sizeof(__pyx_k_FORWARD_MERGE), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x16, __pyx_k_Incompatible_checksums_s_vs_0x16, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x16), 0, 0, 1, 0},
  {&__pyx_n_s_LETTERS, __pyx_k_LETTERS, sizeof(__pyx_k_LETTERS), 0, 0, 1, 1},
  {&__pyx_n_s_LOWERCASE_LETTERS, __pyx_k_LOWERCASE_LETTERS, sizeof(__pyx_k_LOWERCASE_LETTERS), 0, 0, 1, 1},
  {&__pyx_n_s_MISSPELLED_REPLACE, __pyx_k_MISSPELLED_REPLACE, sizeof(__pyx_k_MISSPELLED_REPLACE), 0, 0, 1, 1},
  {&__pyx_n_s_Operator, __pyx_k_Operator, sizeof(__pyx_k_Operator), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_SPELL_CHECK, __pyx_k_SPELL_CHECK, sizeof(__pyx_k_SPELL_CHECK), 0, 0, 1, 1},
  {&__pyx_n_s_SPLIT, __pyx_k_SPLIT, sizeof(__pyx_k_SPLIT), 0, 0, 1, 1},
  {&__pyx_n_s_SimpleSpellChecker, __pyx_k_SimpleSpellChecker, sizeof(__pyx_k_SimpleSpellChecker), 0, 0, 1, 1},
  {&__pyx_n_s_SpellChecker_Operator, __pyx_k_SpellChecker_Operator, sizeof(__pyx_k_SpellChecker_Operator), 0, 0, 1, 1},
  {&__pyx_n_s_SpellChecker_SimpleSpellChecker, __pyx_k_SpellChecker_SimpleSpellChecker, sizeof(__pyx_k_SpellChecker_SimpleSpellChecker), 0, 0, 1, 1},
  {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
  {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
  {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
  {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
  {&__pyx_kp_u_a_zA_Z, __pyx_k_a_zA_Z, sizeof(__pyx_k_a_zA_Z), 0, 1, 0, 0},
  {&__pyx_n_s_addSplitWords, __pyx_k_addSplitWords, sizeof(__pyx_k_addSplitWords), 0, 0, 1, 1},
  {&__pyx_n_u_ah, __pyx_k_ah, sizeof(__pyx_k_ah), 0, 1, 0, 1},
  {&__pyx_n_u_bit, __pyx_k_bit, sizeof(__pyx_k_bit), 0, 1, 0, 1},
  {&__pyx_n_u_byte, __pyx_k_byte, sizeof(__pyx_k_byte), 0, 1, 0, 1},
  {&__pyx_n_s_candidateList, __pyx_k_candidateList, sizeof(__pyx_k_candidateList), 0, 0, 1, 1},
  {&__pyx_n_u_cc, __pyx_k_cc, sizeof(__pyx_k_cc), 0, 1, 0, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_u_cm, __pyx_k_cm, sizeof(__pyx_k_cm), 0, 1, 0, 1},
  {&__pyx_n_u_cm2, __pyx_k_cm2, sizeof(__pyx_k_cm2), 0, 1, 0, 1},
  {&__pyx_n_u_cm3, __pyx_k_cm3, sizeof(__pyx_k_cm3), 0, 1, 0, 1},
  {&__pyx_n_s_compile, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
  {&__pyx_n_s_conditionalShortcuts, __pyx_k_conditionalShortcuts, sizeof(__pyx_k_conditionalShortcuts), 0, 0, 1, 1},
  {&__pyx_n_u_da, __pyx_k_da, sizeof(__pyx_k_da), 0, 1, 0, 1},
  {&__pyx_kp_u_data_merged_txt, __pyx_k_data_merged_txt, sizeof(__pyx_k_data_merged_txt), 0, 1, 0, 0},
  {&__pyx_kp_u_data_split_txt, __pyx_k_data_split_txt, sizeof(__pyx_k_data_split_txt), 0, 1, 0, 0},
  {&__pyx_n_u_db, __pyx_k_db, sizeof(__pyx_k_db), 0, 1, 0, 1},
  {&__pyx_n_u_de, __pyx_k_de, sizeof(__pyx_k_de), 0, 1, 0, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dictionary, __pyx_k_dictionary, sizeof(__pyx_k_dictionary), 0, 0, 1, 1},
  {&__pyx_n_u_dpi, __pyx_k_dpi, sizeof(__pyx_k_dpi), 0, 1, 0, 1},
  {&__pyx_n_s_encoding, __pyx_k_encoding, sizeof(__pyx_k_encoding), 0, 0, 1, 1},
  {&__pyx_n_s_forcedBackwardMergeCheck, __pyx_k_forcedBackwardMergeCheck, sizeof(__pyx_k_forcedBackwardMergeCheck), 0, 0, 1, 1},
  {&__pyx_n_s_forcedDeDaSplitCheck, __pyx_k_forcedDeDaSplitCheck, sizeof(__pyx_k_forcedDeDaSplitCheck), 0, 0, 1, 1},
  {&__pyx_n_s_forcedForwardMergeCheck, __pyx_k_forcedForwardMergeCheck, sizeof(__pyx_k_forcedForwardMergeCheck), 0, 0, 1, 1},
  {&__pyx_n_s_forcedHyphenMergeCheck, __pyx_k_forcedHyphenMergeCheck, sizeof(__pyx_k_forcedHyphenMergeCheck), 0, 0, 1, 1},
  {&__pyx_n_s_forcedMisspellCheck, __pyx_k_forcedMisspellCheck, sizeof(__pyx_k_forcedMisspellCheck), 0, 0, 1, 1},
  {&__pyx_n_s_forcedQuestionSuffixSplitCheck, __pyx_k_forcedQuestionSuffixSplitCheck, sizeof(__pyx_k_forcedQuestionSuffixSplitCheck), 0, 0, 1, 1},
  {&__pyx_n_s_forcedShortcutCheck, __pyx_k_forcedShortcutCheck, sizeof(__pyx_k_forcedShortcutCheck), 0, 0, 1, 1},
  {&__pyx_n_s_forcedSplitCheck, __pyx_k_forcedSplitCheck, sizeof(__pyx_k_forcedSplitCheck), 0, 0, 1, 1},
  {&__pyx_n_s_forcedSuffixMergeCheck, __pyx_k_forcedSuffixMergeCheck, sizeof(__pyx_k_forcedSuffixMergeCheck), 0, 0, 1, 1},
  {&__pyx_n_s_fsm, __pyx_k_fsm, sizeof(__pyx_k_fsm), 0, 0, 1, 1},
  {&__pyx_n_u_ft, __pyx_k_ft, sizeof(__pyx_k_ft), 0, 1, 0, 1},
  {&__pyx_n_s_fullmatch, __pyx_k_fullmatch, sizeof(__pyx_k_fullmatch), 0, 0, 1, 1},
  {&__pyx_n_u_g, __pyx_k_g, sizeof(__pyx_k_g), 0, 1, 0, 1},
  {&__pyx_n_u_gb, __pyx_k_gb, sizeof(__pyx_k_gb), 0, 1, 0, 1},
  {&__pyx_n_u_gbit, __pyx_k_gbit, sizeof(__pyx_k_gbit), 0, 1, 0, 1},
  {&__pyx_n_u_gbps, __pyx_k_gbps, sizeof(__pyx_k_gbps), 0, 1, 0, 1},
  {&__pyx_n_s_generateCandidateList, __pyx_k_generateCandidateList, sizeof(__pyx_k_generateCandidateList), 0, 0, 1, 1},
  {&__pyx_n_s_getCorrectForm, __pyx_k_getCorrectForm, sizeof(__pyx_k_getCorrectForm), 0, 0, 1, 1},
  {&__pyx_n_s_getName, __pyx_k_getName, sizeof(__pyx_k_getName), 0, 0, 1, 1},
  {&__pyx_n_s_getOperator, __pyx_k_getOperator, sizeof(__pyx_k_getOperator), 0, 0, 1, 1},
  {&__pyx_n_s_getSplitPair, __pyx_k_getSplitPair, sizeof(__pyx_k_getSplitPair), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_u_ghz, __pyx_k_ghz, sizeof(__pyx_k_ghz), 0, 1, 0, 1},
  {&__pyx_n_u_gr, __pyx_k_gr, sizeof(__pyx_k_gr), 0, 1, 0, 1},
  {&__pyx_n_u_gram, __pyx_k_gram, sizeof(__pyx_k_gram), 0, 1, 0, 1},
  {&__pyx_n_u_hp, __pyx_k_hp, sizeof(__pyx_k_hp), 0, 1, 0, 1},
  {&__pyx_n_u_hz, __pyx_k_hz, sizeof(__pyx_k_hz), 0, 1, 0, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_u_in, __pyx_k_in, sizeof(__pyx_k_in), 0, 1, 0, 1},
  {&__pyx_n_u_inc, __pyx_k_inc, sizeof(__pyx_k_inc), 0, 1, 0, 1},
  {&__pyx_n_u_inch, __pyx_k_inch, sizeof(__pyx_k_inch), 0, 1, 0, 1},
  {&__pyx_n_s_isBackVowel, __pyx_k_isBackVowel, sizeof(__pyx_k_isBackVowel), 0, 0, 1, 1},
  {&__pyx_n_u_kbit, __pyx_k_kbit, sizeof(__pyx_k_kbit), 0, 1, 0, 1},
  {&__pyx_n_u_kcal, __pyx_k_kcal, sizeof(__pyx_k_kcal), 0, 1, 0, 1},
  {&__pyx_n_u_kg, __pyx_k_kg, sizeof(__pyx_k_kg), 0, 1, 0, 1},
  {&__pyx_n_u_khz, __pyx_k_khz, sizeof(__pyx_k_khz), 0, 1, 0, 1},
  {&__pyx_n_u_kva, __pyx_k_kva, sizeof(__pyx_k_kva), 0, 1, 0, 1},
  {&__pyx_n_u_kw, __pyx_k_kw, sizeof(__pyx_k_kw), 0, 1, 0, 1},
  {&__pyx_n_u_kwh, __pyx_k_kwh, sizeof(__pyx_k_kwh), 0, 1, 0, 1},
  {&__pyx_n_u_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 1, 0, 1},
  {&__pyx_n_u_l_2, __pyx_k_l_2, sizeof(__pyx_k_l_2), 0, 1, 0, 1},
  {&__pyx_n_u_l_3, __pyx_k_l_3, sizeof(__pyx_k_l_3), 0, 1, 0, 1},
  {&__pyx_n_s_lastVowel, __pyx_k_lastVowel, sizeof(__pyx_k_lastVowel), 0, 0, 1, 1},
  {&__pyx_n_u_li, __pyx_k_li, sizeof(__pyx_k_li), 0, 1, 0, 1},
  {&__pyx_n_u_lik, __pyx_k_lik, sizeof(__pyx_k_lik), 0, 1, 0, 1},
  {&__pyx_n_u_litre, __pyx_k_litre, sizeof(__pyx_k_litre), 0, 1, 0, 1},
  {&__pyx_n_u_lk, __pyx_k_lk, sizeof(__pyx_k_lk), 0, 1, 0, 1},
  {&__pyx_n_u_lk_2, __pyx_k_lk_2, sizeof(__pyx_k_lk_2), 0, 1, 0, 1},
  {&__pyx_n_s_loadDictionaries, __pyx_k_loadDictionaries, sizeof(__pyx_k_loadDictionaries), 0, 0, 1, 1},
  {&__pyx_n_u_lt, __pyx_k_lt, sizeof(__pyx_k_lt), 0, 1, 0, 1},
  {&__pyx_n_u_lu, __pyx_k_lu, sizeof(__pyx_k_lu), 0, 1, 0, 1},
  {&__pyx_n_u_luk, __pyx_k_luk, sizeof(__pyx_k_luk), 0, 1, 0, 1},
  {&__pyx_n_u_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 1, 0, 1},
  {&__pyx_n_u_m2, __pyx_k_m2, sizeof(__pyx_k_m2), 0, 1, 0, 1},
  {&__pyx_n_u_m3, __pyx_k_m3, sizeof(__pyx_k_m3), 0, 1, 0, 1},
  {&__pyx_n_u_m_2, __pyx_k_m_2, sizeof(__pyx_k_m_2), 0, 1, 0, 1},
  {&__pyx_n_u_m_3, __pyx_k_m_3, sizeof(__pyx_k_m_3), 0, 1, 0, 1},
  {&__pyx_n_u_mah, __pyx_k_mah, sizeof(__pyx_k_mah), 0, 1, 0, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_u_mb, __pyx_k_mb, sizeof(__pyx_k_mb), 0, 1, 0, 1},
  {&__pyx_n_u_mbit, __pyx_k_mbit, sizeof(__pyx_k_mbit), 0, 1, 0, 1},
  {&__pyx_n_u_mbps, __pyx_k_mbps, sizeof(__pyx_k_mbps), 0, 1, 0, 1},
  {&__pyx_n_u_mdr, __pyx_k_mdr, sizeof(__pyx_k_mdr), 0, 1, 0, 1},
  {&__pyx_n_u_mdr_2, __pyx_k_mdr_2, sizeof(__pyx_k_mdr_2), 0, 1, 0, 1},
  {&__pyx_n_s_mergedCandidatesList, __pyx_k_mergedCandidatesList, sizeof(__pyx_k_mergedCandidatesList), 0, 0, 1, 1},
  {&__pyx_n_u_metre, __pyx_k_metre, sizeof(__pyx_k_metre), 0, 1, 0, 1},
  {&__pyx_n_u_mg, __pyx_k_mg, sizeof(__pyx_k_mg), 0, 1, 0, 1},
  {&__pyx_n_u_mhz, __pyx_k_mhz, sizeof(__pyx_k_mhz), 0, 1, 0, 1},
  {&__pyx_n_u_mi, __pyx_k_mi, sizeof(__pyx_k_mi), 0, 1, 0, 1},
  {&__pyx_n_u_midir, __pyx_k_midir, sizeof(__pyx_k_midir), 0, 1, 0, 1},
  {&__pyx_n_u_misin, __pyx_k_misin, sizeof(__pyx_k_misin), 0, 1, 0, 1},
  {&__pyx_n_u_misindir, __pyx_k_misindir, sizeof(__pyx_k_misindir), 0, 1, 0, 1},
  {&__pyx_n_u_misiniz, __pyx_k_misiniz, sizeof(__pyx_k_misiniz), 0, 1, 0, 1},
  {&__pyx_n_u_misinizdir, __pyx_k_misinizdir, sizeof(__pyx_k_misinizdir), 0, 1, 0, 1},
  {&__pyx_n_u_miydi, __pyx_k_miydi, sizeof(__pyx_k_miydi), 0, 1, 0, 1},
  {&__pyx_n_u_miydik, __pyx_k_miydik, sizeof(__pyx_k_miydik), 0, 1, 0, 1},
  {&__pyx_n_u_miydiler, __pyx_k_miydiler, sizeof(__pyx_k_miydiler), 0, 1, 0, 1},
  {&__pyx_n_u_miydim, __pyx_k_miydim, sizeof(__pyx_k_miydim), 0, 1, 0, 1},
  {&__pyx_n_u_miydin, __pyx_k_miydin, sizeof(__pyx_k_miydin), 0, 1, 0, 1},
  {&__pyx_n_u_miydiniz, __pyx_k_miydiniz, sizeof(__pyx_k_miydiniz), 0, 1, 0, 1},
  {&__pyx_n_u_miyim, __pyx_k_miyim, sizeof(__pyx_k_miyim), 0, 1, 0, 1},
  {&__pyx_n_u_miyimdir, __pyx_k_miyimdir, sizeof(__pyx_k_miyimdir), 0, 1, 0, 1},
  {&__pyx_n_u_miyiz, __pyx_k_miyiz, sizeof(__pyx_k_miyiz), 0, 1, 0, 1},
  {&__pyx_n_u_miyizdir, __pyx_k_miyizdir, sizeof(__pyx_k_miyizdir), 0, 1, 0, 1},
  {&__pyx_n_u_miymi, __pyx_k_miymi, sizeof(__pyx_k_miymi), 0, 1, 0, 1},
  {&__pyx_n_u_miymiim, __pyx_k_miymiim, sizeof(__pyx_k_miymiim), 0, 1, 0, 1},
  {&__pyx_n_u_miymiiz, __pyx_k_miymiiz, sizeof(__pyx_k_miymiiz), 0, 1, 0, 1},
  {&__pyx_n_u_miymiler, __pyx_k_miymiler, sizeof(__pyx_k_miymiler), 0, 1, 0, 1},
  {&__pyx_n_u_miymisin, __pyx_k_miymisin, sizeof(__pyx_k_miymisin), 0, 1, 0, 1},
  {&__pyx_n_u_ml, __pyx_k_ml, sizeof(__pyx_k_ml), 0, 1, 0, 1},
  {&__pyx_n_u_mm, __pyx_k_mm, sizeof(__pyx_k_mm), 0, 1, 0, 1},
  {&__pyx_n_u_mm2, __pyx_k_mm2, sizeof(__pyx_k_mm2), 0, 1, 0, 1},
  {&__pyx_n_u_mm3, __pyx_k_mm3, sizeof(__pyx_k_mm3), 0, 1, 0, 1},
  {&__pyx_n_u_mp, __pyx_k_mp, sizeof(__pyx_k_mp), 0, 1, 0, 1},
  {&__pyx_n_u_ms, __pyx_k_ms, sizeof(__pyx_k_ms), 0, 1, 0, 1},
  {&__pyx_n_u_msn, __pyx_k_msn, sizeof(__pyx_k_msn), 0, 1, 0, 1},
  {&__pyx_n_u_msn_2, __pyx_k_msn_2, sizeof(__pyx_k_msn_2), 0, 1, 0, 1},
  {&__pyx_n_u_msndr, __pyx_k_msndr, sizeof(__pyx_k_msndr), 0, 1, 0, 1},
  {&__pyx_n_u_msndr_2, __pyx_k_msndr_2, sizeof(__pyx_k_msndr_2), 0, 1, 0, 1},
  {&__pyx_n_u_msnz, __pyx_k_msnz, sizeof(__pyx_k_msnz), 0, 1, 0, 1},
  {&__pyx_n_u_msnz_2, __pyx_k_msnz_2, sizeof(__pyx_k_msnz_2), 0, 1, 0, 1},
  {&__pyx_n_u_msnzdr, __pyx_k_msnzdr, sizeof(__pyx_k_msnzdr), 0, 1, 0, 1},
  {&__pyx_n_u_msnzdr_2, __pyx_k_msnzdr_2, sizeof(__pyx_k_msnzdr_2), 0, 1, 0, 1},
  {&__pyx_n_u_mt, __pyx_k_mt, sizeof(__pyx_k_mt), 0, 1, 0, 1},
  {&__pyx_n_u_mu, __pyx_k_mu, sizeof(__pyx_k_mu), 0, 1, 0, 1},
  {&__pyx_n_u_mudur, __pyx_k_mudur, sizeof(__pyx_k_mudur), 0, 1, 0, 1},
  {&__pyx_n_s_multiWord, __pyx_k_multiWord, sizeof(__pyx_k_multiWord), 0, 0, 1, 1},
  {&__pyx_n_u_musun, __pyx_k_musun, sizeof(__pyx_k_musun), 0, 1, 0, 1},
  {&__pyx_n_u_musundur, __pyx_k_musundur, sizeof(__pyx_k_musundur), 0, 1, 0, 1},
  {&__pyx_n_u_musunuz, __pyx_k_musunuz, sizeof(__pyx_k_musunuz), 0, 1, 0, 1},
  {&__pyx_n_u_musunuzdur, __pyx_k_musunuzdur, sizeof(__pyx_k_musunuzdur), 0, 1, 0, 1},
  {&__pyx_n_u_muydu, __pyx_k_muydu, sizeof(__pyx_k_muydu), 0, 1, 0, 1},
  {&__pyx_n_u_muyduk, __pyx_k_muyduk, sizeof(__pyx_k_muyduk), 0, 1, 0, 1},
  {&__pyx_n_u_muydular, __pyx_k_muydular, sizeof(__pyx_k_muydular), 0, 1, 0, 1},
  {&__pyx_n_u_muydum, __pyx_k_muydum, sizeof(__pyx_k_muydum), 0, 1, 0, 1},
  {&__pyx_n_u_muydun, __pyx_k_muydun, sizeof(__pyx_k_muydun), 0, 1, 0, 1},
  {&__pyx_n_u_muydunuz, __pyx_k_muydunuz, sizeof(__pyx_k_muydunuz), 0, 1, 0, 1},
  {&__pyx_n_u_muymu, __pyx_k_muymu, sizeof(__pyx_k_muymu), 0, 1, 0, 1},
  {&__pyx_n_u_muymular, __pyx_k_muymular, sizeof(__pyx_k_muymular), 0, 1, 0, 1},
  {&__pyx_n_u_muymusun, __pyx_k_muymusun, sizeof(__pyx_k_muymusun), 0, 1, 0, 1},
  {&__pyx_n_u_muymuum, __pyx_k_muymuum, sizeof(__pyx_k_muymuum), 0, 1, 0, 1},
  {&__pyx_n_u_muymuuz, __pyx_k_muymuuz, sizeof(__pyx_k_muymuuz), 0, 1, 0, 1},
  {&__pyx_n_u_muyum, __pyx_k_muyum, sizeof(__pyx_k_muyum), 0, 1, 0, 1},
  {&__pyx_n_u_muyumdur, __pyx_k_muyumdur, sizeof(__pyx_k_muyumdur), 0, 1, 0, 1},
  {&__pyx_n_u_muyuz, __pyx_k_muyuz, sizeof(__pyx_k_muyuz), 0, 1, 0, 1},
  {&__pyx_n_u_muyuzdur, __pyx_k_muyuzdur, sizeof(__pyx_k_muyuzdur), 0, 1, 0, 1},
  {&__pyx_n_u_mv, __pyx_k_mv, sizeof(__pyx_k_mv), 0, 1, 0, 1},
  {&__pyx_n_u_myd, __pyx_k_myd, sizeof(__pyx_k_myd), 0, 1, 0, 1},
  {&__pyx_n_u_myd_2, __pyx_k_myd_2, sizeof(__pyx_k_myd_2), 0, 1, 0, 1},
  {&__pyx_n_u_mydk, __pyx_k_mydk, sizeof(__pyx_k_mydk), 0, 1, 0, 1},
  {&__pyx_n_u_mydk_2, __pyx_k_mydk_2, sizeof(__pyx_k_mydk_2), 0, 1, 0, 1},
  {&__pyx_n_u_mydlar, __pyx_k_mydlar, sizeof(__pyx_k_mydlar), 0, 1, 0, 1},
  {&__pyx_n_u_mydler, __pyx_k_mydler, sizeof(__pyx_k_mydler), 0, 1, 0, 1},
  {&__pyx_n_u_mydm, __pyx_k_mydm, sizeof(__pyx_k_mydm), 0, 1, 0, 1},
  {&__pyx_n_u_mydm_2, __pyx_k_mydm_2, sizeof(__pyx_k_mydm_2), 0, 1, 0, 1},
  {&__pyx_n_u_mydn, __pyx_k_mydn, sizeof(__pyx_k_mydn), 0, 1, 0, 1},
  {&__pyx_n_u_mydn_2, __pyx_k_mydn_2, sizeof(__pyx_k_mydn_2), 0, 1, 0, 1},
  {&__pyx_n_u_mydnz, __pyx_k_mydnz, sizeof(__pyx_k_mydnz), 0, 1, 0, 1},
  {&__pyx_n_u_mydnz_2, __pyx_k_mydnz_2, sizeof(__pyx_k_mydnz_2), 0, 1, 0, 1},
  {&__pyx_n_u_mym, __pyx_k_mym, sizeof(__pyx_k_mym), 0, 1, 0, 1},
  {&__pyx_n_u_mym_2, __pyx_k_mym_2, sizeof(__pyx_k_mym_2), 0, 1, 0, 1},
  {&__pyx_n_u_mym_3, __pyx_k_mym_3, sizeof(__pyx_k_mym_3), 0, 1, 0, 1},
  {&__pyx_n_u_mym_4, __pyx_k_mym_4, sizeof(__pyx_k_mym_4), 0, 1, 0, 1},
  {&__pyx_n_u_mymdr, __pyx_k_mymdr, sizeof(__pyx_k_mymdr), 0, 1, 0, 1},
  {&__pyx_n_u_mymdr_2, __pyx_k_mymdr_2, sizeof(__pyx_k_mymdr_2), 0, 1, 0, 1},
  {&__pyx_n_u_mymlar, __pyx_k_mymlar, sizeof(__pyx_k_mymlar), 0, 1, 0, 1},
  {&__pyx_n_u_mymler, __pyx_k_mymler, sizeof(__pyx_k_mymler), 0, 1, 0, 1},
  {&__pyx_n_u_mymm, __pyx_k_mymm, sizeof(__pyx_k_mymm), 0, 1, 0, 1},
  {&__pyx_n_u_mymm_2, __pyx_k_mymm_2, sizeof(__pyx_k_mymm_2), 0, 1, 0, 1},
  {&__pyx_n_u_mymsn, __pyx_k_mymsn, sizeof(__pyx_k_mymsn), 0, 1, 0, 1},
  {&__pyx_n_u_mymsn_2, __pyx_k_mymsn_2, sizeof(__pyx_k_mymsn_2), 0, 1, 0, 1},
  {&__pyx_n_u_mymz, __pyx_k_mymz, sizeof(__pyx_k_mymz), 0, 1, 0, 1},
  {&__pyx_n_u_mymz_2, __pyx_k_mymz_2, sizeof(__pyx_k_mymz_2), 0, 1, 0, 1},
  {&__pyx_n_u_myz, __pyx_k_myz, sizeof(__pyx_k_myz), 0, 1, 0, 1},
  {&__pyx_n_u_myz_2, __pyx_k_myz_2, sizeof(__pyx_k_myz_2), 0, 1, 0, 1},
  {&__pyx_n_u_myzdr, __pyx_k_myzdr, sizeof(__pyx_k_myzdr), 0, 1, 0, 1},
  {&__pyx_n_u_myzdr_2, __pyx_k_myzdr_2, sizeof(__pyx_k_myzdr_2), 0, 1, 0, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_nextWord, __pyx_k_nextWord, sizeof(__pyx_k_nextWord), 0, 0, 1, 1},
  {&__pyx_n_u_ohm, __pyx_k_ohm, sizeof(__pyx_k_ohm), 0, 1, 0, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_u_oz, __pyx_k_oz, sizeof(__pyx_k_oz), 0, 1, 0, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pkg_resources, __pyx_k_pkg_resources, sizeof(__pyx_k_pkg_resources), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_u_ppm, __pyx_k_ppm, sizeof(__pyx_k_ppm), 0, 1, 0, 1},
  {&__pyx_n_s_previousWord, __pyx_k_previousWord, sizeof(__pyx_k_previousWord), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_SimpleSpellChecke, __pyx_k_pyx_unpickle_SimpleSpellChecke, sizeof(__pyx_k_pyx_unpickle_SimpleSpellChecke), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_questionSuffixList, __pyx_k_questionSuffixList, sizeof(__pyx_k_questionSuffixList), 0, 0, 1, 1},
  {&__pyx_n_u_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 1, 0, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_randrange, __pyx_k_randrange, sizeof(__pyx_k_randrange), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
  {&__pyx_n_s_readlines, __pyx_k_readlines, sizeof(__pyx_k_readlines), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_resource_filename, __pyx_k_resource_filename, sizeof(__pyx_k_resource_filename), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_rindex, __pyx_k_rindex, sizeof(__pyx_k_rindex), 0, 0, 1, 1},
  {&__pyx_n_u_rpm, __pyx_k_rpm, sizeof(__pyx_k_rpm), 0, 1, 0, 1},
  {&__pyx_n_u_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 1, 0, 1},
  {&__pyx_n_s_sentence, __pyx_k_sentence, sizeof(__pyx_k_sentence), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shortcuts, __pyx_k_shortcuts, sizeof(__pyx_k_shortcuts), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_u_sn, __pyx_k_sn, sizeof(__pyx_k_sn), 0, 1, 0, 1},
  {&__pyx_n_s_spellCheck, __pyx_k_spellCheck, sizeof(__pyx_k_spellCheck), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_splitCandidatesList, __pyx_k_splitCandidatesList, sizeof(__pyx_k_splitCandidatesList), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_u_tb, __pyx_k_tb, sizeof(__pyx_k_tb), 0, 1, 0, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_u_tl, __pyx_k_tl, sizeof(__pyx_k_tl), 0, 1, 0, 1},
  {&__pyx_n_s_toCapital, __pyx_k_toCapital, sizeof(__pyx_k_toCapital), 0, 0, 1, 1},
  {&__pyx_n_s_toLower, __pyx_k_toLower, sizeof(__pyx_k_toLower), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_u_utf8, __pyx_k_utf8, sizeof(__pyx_k_utf8), 0, 1, 0, 1},
  {&__pyx_n_u_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 1, 0, 1},
  {&__pyx_n_u_va, __pyx_k_va, sizeof(__pyx_k_va), 0, 1, 0, 1},
  {&__pyx_n_u_volt, __pyx_k_volt, sizeof(__pyx_k_volt), 0, 1, 0, 1},
  {&__pyx_n_u_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 1, 0, 1},
  {&__pyx_n_u_watt, __pyx_k_watt, sizeof(__pyx_k_watt), 0, 1, 0, 1},
  {&__pyx_n_s_word, __pyx_k_word, sizeof(__pyx_k_word), 0, 0, 1, 1},
  {&__pyx_n_s_wordName, __pyx_k_wordName, sizeof(__pyx_k_wordName), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 74, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 380, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "SpellChecker/SimpleSpellChecker.pyx":307
 *                     if TurkishLanguage.isBackVowel(Word.lastVowel(new_word_name)):
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():
 *                             result.addWord(Word("de"))             # <<<<<<<<<<<<<<
 *                         else:
 *                             result.addWord(Word("da"))
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_n_u_de); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "SpellChecker/SimpleSpellChecker.pyx":309
 *                             result.addWord(Word("de"))
 *                         else:
 *                             result.addWord(Word("da"))             # <<<<<<<<<<<<<<
 *                     else:
 *                         if txt_word.notObeysVowelHarmonyDuringAgglutination():
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_n_u_da); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "(tree fragment)":1
 * def __pyx_unpickle_SimpleSpellChecker(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__11 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__11, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_SimpleSpellChecke, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyUnicode_Type_rindex.type = (PyObject*)&PyUnicode_Type;
  __pyx_umethod_PyUnicode_Type_strip.type = (PyObject*)&PyUnicode_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_23252194 = PyInt_FromLong(23252194L); if (unlikely(!__pyx_int_23252194)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("SpellChecker.SpellChecker"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker = __Pyx_ImportType(__pyx_t_1, "SpellChecker.SpellChecker", "SpellChecker", sizeof(struct __pyx_obj_12SpellChecker_12SpellChecker_SpellChecker), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_12SpellChecker_12SpellChecker_SpellChecker = (struct __pyx_vtabstruct_12SpellChecker_12SpellChecker_SpellChecker*)__Pyx_GetVtable(__pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker->tp_dict); if (unlikely(!__pyx_vtabptr_12SpellChecker_12SpellChecker_SpellChecker)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker = &__pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.__pyx_base = *__pyx_vtabptr_12SpellChecker_12SpellChecker_SpellChecker;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.__pyx_base.spellCheck = (struct __pyx_obj_6Corpus_8Sentence_Sentence *(*)(struct __pyx_obj_12SpellChecker_12SpellChecker_SpellChecker *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_spellCheck;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.__pyx___generateCandidateList = (PyObject *(*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___generateCandidateList;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.candidateList = (PyObject *(*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_candidateList;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.forcedMisspellCheck = (int (*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedMisspellCheck;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.forcedBackwardMergeCheck = (int (*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedBackwardMergeCheck;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.forcedForwardMergeCheck = (int (*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedForwardMergeCheck;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.forcedSplitCheck = (int (*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedSplitCheck;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.forcedShortcutCheck = (int (*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedShortcutCheck;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.forcedDeDaSplitCheck = (int (*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedDeDaSplitCheck;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.forcedSuffixMergeCheck = (int (*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedSuffixMergeCheck;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.forcedHyphenMergeCheck = (int (*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedHyphenMergeCheck;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.forcedQuestionSuffixSplitCheck = (int (*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_6Corpus_8Sentence_Sentence *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_forcedQuestionSuffixSplitCheck;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.loadDictionaries = (PyObject *(*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_loadDictionaries;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.getCorrectForm = (PyObject *(*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_getCorrectForm;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.mergedCandidatesList = (PyObject *(*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_10Dictionary_4Word_Word *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_mergedCandidatesList;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.splitCandidatesList = (PyObject *(*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_splitCandidatesList;
  __pyx_vtable_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.getSplitPair = (PyObject *(*)(struct __pyx_obj_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker *, struct __pyx_obj_10Dictionary_4Word_Word *, int __pyx_skip_dispatch))__pyx_f_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker_getSplitPair;
  __pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.tp_base = __pyx_ptype_12SpellChecker_12SpellChecker_SpellChecker;
  if (PyType_Ready(&__pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.tp_dictoffset && __pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 12, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___init__.doc = __pyx_doc_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_12SpellChecker_18SimpleSpellChecker_18SimpleSpellChecker___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker.tp_dict, __pyx_vtabptr_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SimpleSpellChecker, (PyObject *)&__pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_ptype_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker = &__pyx_type_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("Dictionary.Word"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10Dictionary_4Word_Word = __Pyx_ImportType(__pyx_t_1, "Dictionary.Word", "Word", sizeof(struct __pyx_obj_10Dictionary_4Word_Word), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10Dictionary_4Word_Word) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_vtabptr_10Dictionary_4Word_Word = (struct __pyx_vtabstruct_10Dictionary_4Word_Word*)__Pyx_GetVtable(__pyx_ptype_10Dictionary_4Word_Word->tp_dict); if (unlikely(!__pyx_vtabptr_10Dictionary_4Word_Word)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("Corpus.Sentence"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6Corpus_8Sentence_Sentence = __Pyx_ImportType(__pyx_t_1, "Corpus.Sentence", "Sentence", sizeof(struct __pyx_obj_6Corpus_8Sentence_Sentence), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6Corpus_8Sentence_Sentence) __PYX_ERR(3, 4, __pyx_L1_error)
  __pyx_vtabptr_6Corpus_8Sentence_Sentence = (struct __pyx_vtabstruct_6Corpus_8Sentence_Sentence*)__Pyx_GetVtable(__pyx_ptype_6Corpus_8Sentence_Sentence->tp_dict); if (unlikely(!__pyx_vtabptr_6Corpus_8Sentence_Sentence)) __PYX_ERR(3, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("Dictionary.Trie.TrieNode"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10Dictionary_4Trie_8TrieNode_TrieNode = __Pyx_ImportType(__pyx_t_1, "Dictionary.Trie.TrieNode", "TrieNode", sizeof(struct __pyx_obj_10Dictionary_4Trie_8TrieNode_TrieNode), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10Dictionary_4Trie_8TrieNode_TrieNode) __PYX_ERR(4, 4, __pyx_L1_error)
  __pyx_vtabptr_10Dictionary_4Trie_8TrieNode_TrieNode = (struct __pyx_vtabstruct_10Dictionary_4Trie_8TrieNode_TrieNode*)__Pyx_GetVtable(__pyx_ptype_10Dictionary_4Trie_8TrieNode_TrieNode->tp_dict); if (unlikely(!__pyx_vtabptr_10Dictionary_4Trie_8TrieNode_TrieNode)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("Dictionary.TxtWord"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10Dictionary_7TxtWord_TxtWord = __Pyx_ImportType(__pyx_t_1, "Dictionary.TxtWord", "TxtWord", sizeof(struct __pyx_obj_10Dictionary_7TxtWord_TxtWord), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10Dictionary_7TxtWord_TxtWord) __PYX_ERR(5, 4, __pyx_L1_error)
  __pyx_vtabptr_10Dictionary_7TxtWord_TxtWord = (struct __pyx_vtabstruct_10Dictionary_7TxtWord_TxtWord*)__Pyx_GetVtable(__pyx_ptype_10Dictionary_7TxtWord_TxtWord->tp_dict); if (unlikely(!__pyx_vtabptr_10Dictionary_7TxtWord_TxtWord)) __PYX_ERR(5, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("Dictionary.Trie.Trie"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10Dictionary_4Trie_4Trie_Trie = __Pyx_ImportType(__pyx_t_1, "Dictionary.Trie.Trie", "Trie", sizeof(struct __pyx_obj_10Dictionary_4Trie_4Trie_Trie), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10Dictionary_4Trie_4Trie_Trie) __PYX_ERR(6, 6, __pyx_L1_error)
  __pyx_vtabptr_10Dictionary_4Trie_4Trie_Trie = (struct __pyx_vtabstruct_10Dictionary_4Trie_4Trie_Trie*)__Pyx_GetVtable(__pyx_ptype_10Dictionary_4Trie_4Trie_Trie->tp_dict); if (unlikely(!__pyx_vtabptr_10Dictionary_4Trie_4Trie_Trie)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("Dictionary.Dictionary"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10Dictionary_10Dictionary_Dictionary = __Pyx_ImportType(__pyx_t_1, "Dictionary.Dictionary", "Dictionary", sizeof(struct __pyx_obj_10Dictionary_10Dictionary_Dictionary), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10Dictionary_10Dictionary_Dictionary) __PYX_ERR(7, 4, __pyx_L1_error)
  __pyx_vtabptr_10Dictionary_10Dictionary_Dictionary = (struct __pyx_vtabstruct_10Dictionary_10Dictionary_Dictionary*)__Pyx_GetVtable(__pyx_ptype_10Dictionary_10Dictionary_Dictionary->tp_dict); if (unlikely(!__pyx_vtabptr_10Dictionary_10Dictionary_Dictionary)) __PYX_ERR(7, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("Dictionary.TxtDictionary"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10Dictionary_13TxtDictionary_TxtDictionary = __Pyx_ImportType(__pyx_t_1, "Dictionary.TxtDictionary", "TxtDictionary", sizeof(struct __pyx_obj_10Dictionary_13TxtDictionary_TxtDictionary), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10Dictionary_13TxtDictionary_TxtDictionary) __PYX_ERR(8, 7, __pyx_L1_error)
  __pyx_vtabptr_10Dictionary_13TxtDictionary_TxtDictionary = (struct __pyx_vtabstruct_10Dictionary_13TxtDictionary_TxtDictionary*)__Pyx_GetVtable(__pyx_ptype_10Dictionary_13TxtDictionary_TxtDictionary->tp_dict); if (unlikely(!__pyx_vtabptr_10Dictionary_13TxtDictionary_TxtDictionary)) __PYX_ERR(8, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("DataStructure.Cache.LRUCache"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_13DataStructure_5Cache_8LRUCache_LRUCache = __Pyx_ImportType(__pyx_t_1, "DataStructure.Cache.LRUCache", "LRUCache", sizeof(struct __pyx_obj_13DataStructure_5Cache_8LRUCache_LRUCache), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_13DataStructure_5Cache_8LRUCache_LRUCache) __PYX_ERR(9, 1, __pyx_L1_error)
  __pyx_vtabptr_13DataStructure_5Cache_8LRUCache_LRUCache = (struct __pyx_vtabstruct_13DataStructure_5Cache_8LRUCache_LRUCache*)__Pyx_GetVtable(__pyx_ptype_13DataStructure_5Cache_8LRUCache_LRUCache->tp_dict); if (unlikely(!__pyx_vtabptr_13DataStructure_5Cache_8LRUCache_LRUCache)) __PYX_ERR(9, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("MorphologicalAnalysis.InflectionalGroup"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup = __Pyx_ImportType(__pyx_t_1, "MorphologicalAnalysis.InflectionalGroup", "InflectionalGroup", sizeof(struct __pyx_obj_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup) __PYX_ERR(10, 1, __pyx_L1_error)
  __pyx_vtabptr_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup = (struct __pyx_vtabstruct_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup*)__Pyx_GetVtable(__pyx_ptype_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup->tp_dict); if (unlikely(!__pyx_vtabptr_21MorphologicalAnalysis_17InflectionalGroup_InflectionalGroup)) __PYX_ERR(10, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("MorphologicalAnalysis.MorphologicalParse"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse = __Pyx_ImportType(__pyx_t_1, "MorphologicalAnalysis.MorphologicalParse", "MorphologicalParse", sizeof(struct __pyx_obj_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse) __PYX_ERR(11, 6, __pyx_L1_error)
  __pyx_vtabptr_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse = (struct __pyx_vtabstruct_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse*)__Pyx_GetVtable(__pyx_ptype_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse->tp_dict); if (unlikely(!__pyx_vtabptr_21MorphologicalAnalysis_18MorphologicalParse_MorphologicalParse)) __PYX_ERR(11, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("MorphologicalAnalysis.State"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_21MorphologicalAnalysis_5State_State = __Pyx_ImportType(__pyx_t_1, "MorphologicalAnalysis.State", "State", sizeof(struct __pyx_obj_21MorphologicalAnalysis_5State_State), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_21MorphologicalAnalysis_5State_State) __PYX_ERR(12, 1, __pyx_L1_error)
  __pyx_vtabptr_21MorphologicalAnalysis_5State_State = (struct __pyx_vtabstruct_21MorphologicalAnalysis_5State_State*)__Pyx_GetVtable(__pyx_ptype_21MorphologicalAnalysis_5State_State->tp_dict); if (unlikely(!__pyx_vtabptr_21MorphologicalAnalysis_5State_State)) __PYX_ERR(12, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("MorphologicalAnalysis.FsmParse"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_21MorphologicalAnalysis_8FsmParse_FsmParse = __Pyx_ImportType(__pyx_t_1, "MorphologicalAnalysis.FsmParse", "FsmParse", sizeof(struct __pyx_obj_21MorphologicalAnalysis_8FsmParse_FsmParse), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_21MorphologicalAnalysis_8FsmParse_FsmParse) __PYX_ERR(13, 5, __pyx_L1_error)
  __pyx_vtabptr_21MorphologicalAnalysis_8FsmParse_FsmParse = (struct __pyx_vtabstruct_21MorphologicalAnalysis_8FsmParse_FsmParse*)__Pyx_GetVtable(__pyx_ptype_21MorphologicalAnalysis_8FsmParse_FsmParse->tp_dict); if (unlikely(!__pyx_vtabptr_21MorphologicalAnalysis_8FsmParse_FsmParse)) __PYX_ERR(13, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("MorphologicalAnalysis.FiniteStateMachine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine = __Pyx_ImportType(__pyx_t_1, "MorphologicalAnalysis.FiniteStateMachine", "FiniteStateMachine", sizeof(struct __pyx_obj_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine) __PYX_ERR(14, 4, __pyx_L1_error)
  __pyx_vtabptr_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine = (struct __pyx_vtabstruct_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine*)__Pyx_GetVtable(__pyx_ptype_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine->tp_dict); if (unlikely(!__pyx_vtabptr_21MorphologicalAnalysis_18FiniteStateMachine_FiniteStateMachine)) __PYX_ERR(14, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("MorphologicalAnalysis.MetamorphicParse"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse = __Pyx_ImportType(__pyx_t_1, "MorphologicalAnalysis.MetamorphicParse", "MetamorphicParse", sizeof(struct __pyx_obj_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse) __PYX_ERR(15, 6, __pyx_L1_error)
  __pyx_vtabptr_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse = (struct __pyx_vtabstruct_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse*)__Pyx_GetVtable(__pyx_ptype_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse->tp_dict); if (unlikely(!__pyx_vtabptr_21MorphologicalAnalysis_16MetamorphicParse_MetamorphicParse)) __PYX_ERR(15, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("MorphologicalAnalysis.FsmMorphologicalAnalyzer"); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer = __Pyx_ImportType(__pyx_t_1, "MorphologicalAnalysis.FsmMorphologicalAnalyzer", "FsmMorphologicalAnalyzer", sizeof(struct __pyx_obj_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer) __PYX_ERR(16, 13, __pyx_L1_error)
  __pyx_vtabptr_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer = (struct __pyx_vtabstruct_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer*)__Pyx_GetVtable(__pyx_ptype_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer->tp_dict); if (unlikely(!__pyx_vtabptr_21MorphologicalAnalysis_24FsmMorphologicalAnalyzer_FsmMorphologicalAnalyzer)) __PYX_ERR(16, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("Language.Language"); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8Language_8Language_Language = __Pyx_ImportType(__pyx_t_1, "Language.Language", "Language", sizeof(struct __pyx_obj_8Language_8Language_Language), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8Language_8Language_Language) __PYX_ERR(17, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("Language.TurkishLanguage"); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_8Language_15TurkishLanguage_TurkishLanguage = __Pyx_ImportType(__pyx_t_1, "Language.TurkishLanguage", "TurkishLanguage", sizeof(struct __pyx_obj_8Language_15TurkishLanguage_TurkishLanguage), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_8Language_15TurkishLanguage_TurkishLanguage) __PYX_ERR(18, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("MorphologicalAnalysis.FsmParseList"); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList = __Pyx_ImportType(__pyx_t_1, "MorphologicalAnalysis.FsmParseList", "FsmParseList", sizeof(struct __pyx_obj_21MorphologicalAnalysis_12FsmParseList_FsmParseList), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList) __PYX_ERR(19, 4, __pyx_L1_error)
  __pyx_vtabptr_21MorphologicalAnalysis_12FsmParseList_FsmParseList = (struct __pyx_vtabstruct_21MorphologicalAnalysis_12FsmParseList_FsmParseList*)__Pyx_GetVtable(__pyx_ptype_21MorphologicalAnalysis_12FsmParseList_FsmParseList->tp_dict); if (unlikely(!__pyx_vtabptr_21MorphologicalAnalysis_12FsmParseList_FsmParseList)) __PYX_ERR(19, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("SpellChecker.Candidate"); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_12SpellChecker_9Candidate_Candidate = __Pyx_ImportType(__pyx_t_1, "SpellChecker.Candidate", "Candidate", sizeof(struct __pyx_obj_12SpellChecker_9Candidate_Candidate), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_12SpellChecker_9Candidate_Candidate) __PYX_ERR(20, 3, __pyx_L1_error)
  __pyx_vtabptr_12SpellChecker_9Candidate_Candidate = (struct __pyx_vtabstruct_12SpellChecker_9Candidate_Candidate*)__Pyx_GetVtable(__pyx_ptype_12SpellChecker_9Candidate_Candidate->tp_dict); if (unlikely(!__pyx_vtabptr_12SpellChecker_9Candidate_Candidate)) __PYX_ERR(20, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initSimpleSpellChecker(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initSimpleSpellChecker(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_SimpleSpellChecker(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_SimpleSpellChecker(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_SimpleSpellChecker(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'SimpleSpellChecker' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_SimpleSpellChecker(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("SimpleSpellChecker", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_SpellChecker__SimpleSpellChecker) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "SpellChecker.SimpleSpellChecker")) {
      if (unlikely(PyDict_SetItemString(modules, "SpellChecker.SimpleSpellChecker", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "SpellChecker/SimpleSpellChecker.pyx":1
 * from random import randrange             # <<<<<<<<<<<<<<
 * import re
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_randrange);
  __Pyx_GIVEREF(__pyx_n_s_randrange);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_randrange);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_random, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_randrange); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_randrange, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":2
 * from random import randrange
 * import re             # <<<<<<<<<<<<<<
 * 
 * from Dictionary.TxtWord cimport TxtWord
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_re, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_re, __pyx_t_2) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":8
 * from MorphologicalAnalysis.FsmParseList cimport FsmParseList
 * 
 * import pkg_resources             # <<<<<<<<<<<<<<
 * from SpellChecker.Candidate cimport Candidate
 * from SpellChecker.Operator import Operator
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pkg_resources, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pkg_resources, __pyx_t_2) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":10
 * import pkg_resources
 * from SpellChecker.Candidate cimport Candidate
 * from SpellChecker.Operator import Operator             # <<<<<<<<<<<<<<
 * 
 * cdef class SimpleSpellChecker(SpellChecker):
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Operator);
  __Pyx_GIVEREF(__pyx_n_s_Operator);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Operator);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_SpellChecker_Operator, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Operator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Operator, __pyx_t_2) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":14
 * cdef class SimpleSpellChecker(SpellChecker):
 * 
 *     __shortcuts: list = ["cc", "cm2", "cm", "gb", "ghz", "gr", "gram", "hz", "inc", "inch", "in", "kg", "kw", "kva",             # <<<<<<<<<<<<<<
 *                          "litre", "lt", "m2", "m3", "mah", "mb", "metre", "mg", "mhz", "ml", "mm", "mp", "ms",
 *                          "mt", "mv", "tb", "tl", "va", "volt", "watt", "ah", "hp", "oz", "rpm", "dpi", "ppm", "ohm",
 */
  __pyx_t_1 = PyList_New(57); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_cc);
  __Pyx_GIVEREF(__pyx_n_u_cc);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_cc);
  __Pyx_INCREF(__pyx_n_u_cm2);
  __Pyx_GIVEREF(__pyx_n_u_cm2);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_cm2);
  __Pyx_INCREF(__pyx_n_u_cm);
  __Pyx_GIVEREF(__pyx_n_u_cm);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_cm);
  __Pyx_INCREF(__pyx_n_u_gb);
  __Pyx_GIVEREF(__pyx_n_u_gb);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_gb);
  __Pyx_INCREF(__pyx_n_u_ghz);
  __Pyx_GIVEREF(__pyx_n_u_ghz);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_ghz);
  __Pyx_INCREF(__pyx_n_u_gr);
  __Pyx_GIVEREF(__pyx_n_u_gr);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_gr);
  __Pyx_INCREF(__pyx_n_u_gram);
  __Pyx_GIVEREF(__pyx_n_u_gram);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_gram);
  __Pyx_INCREF(__pyx_n_u_hz);
  __Pyx_GIVEREF(__pyx_n_u_hz);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_hz);
  __Pyx_INCREF(__pyx_n_u_inc);
  __Pyx_GIVEREF(__pyx_n_u_inc);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_inc);
  __Pyx_INCREF(__pyx_n_u_inch);
  __Pyx_GIVEREF(__pyx_n_u_inch);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_inch);
  __Pyx_INCREF(__pyx_n_u_in);
  __Pyx_GIVEREF(__pyx_n_u_in);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_u_in);
  __Pyx_INCREF(__pyx_n_u_kg);
  __Pyx_GIVEREF(__pyx_n_u_kg);
  PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_u_kg);
  __Pyx_INCREF(__pyx_n_u_kw);
  __Pyx_GIVEREF(__pyx_n_u_kw);
  PyList_SET_ITEM(__pyx_t_1, 12, __pyx_n_u_kw);
  __Pyx_INCREF(__pyx_n_u_kva);
  __Pyx_GIVEREF(__pyx_n_u_kva);
  PyList_SET_ITEM(__pyx_t_1, 13, __pyx_n_u_kva);
  __Pyx_INCREF(__pyx_n_u_litre);
  __Pyx_GIVEREF(__pyx_n_u_litre);
  PyList_SET_ITEM(__pyx_t_1, 14, __pyx_n_u_litre);
  __Pyx_INCREF(__pyx_n_u_lt);
  __Pyx_GIVEREF(__pyx_n_u_lt);
  PyList_SET_ITEM(__pyx_t_1, 15, __pyx_n_u_lt);
  __Pyx_INCREF(__pyx_n_u_m2);
  __Pyx_GIVEREF(__pyx_n_u_m2);
  PyList_SET_ITEM(__pyx_t_1, 16, __pyx_n_u_m2);
  __Pyx_INCREF(__pyx_n_u_m3);
  __Pyx_GIVEREF(__pyx_n_u_m3);
  PyList_SET_ITEM(__pyx_t_1, 17, __pyx_n_u_m3);
  __Pyx_INCREF(__pyx_n_u_mah);
  __Pyx_GIVEREF(__pyx_n_u_mah);
  PyList_SET_ITEM(__pyx_t_1, 18, __pyx_n_u_mah);
  __Pyx_INCREF(__pyx_n_u_mb);
  __Pyx_GIVEREF(__pyx_n_u_mb);
  PyList_SET_ITEM(__pyx_t_1, 19, __pyx_n_u_mb);
  __Pyx_INCREF(__pyx_n_u_metre);
  __Pyx_GIVEREF(__pyx_n_u_metre);
  PyList_SET_ITEM(__pyx_t_1, 20, __pyx_n_u_metre);
  __Pyx_INCREF(__pyx_n_u_mg);
  __Pyx_GIVEREF(__pyx_n_u_mg);
  PyList_SET_ITEM(__pyx_t_1, 21, __pyx_n_u_mg);
  __Pyx_INCREF(__pyx_n_u_mhz);
  __Pyx_GIVEREF(__pyx_n_u_mhz);
  PyList_SET_ITEM(__pyx_t_1, 22, __pyx_n_u_mhz);
  __Pyx_INCREF(__pyx_n_u_ml);
  __Pyx_GIVEREF(__pyx_n_u_ml);
  PyList_SET_ITEM(__pyx_t_1, 23, __pyx_n_u_ml);
  __Pyx_INCREF(__pyx_n_u_mm);
  __Pyx_GIVEREF(__pyx_n_u_mm);
  PyList_SET_ITEM(__pyx_t_1, 24, __pyx_n_u_mm);
  __Pyx_INCREF(__pyx_n_u_mp);
  __Pyx_GIVEREF(__pyx_n_u_mp);
  PyList_SET_ITEM(__pyx_t_1, 25, __pyx_n_u_mp);
  __Pyx_INCREF(__pyx_n_u_ms);
  __Pyx_GIVEREF(__pyx_n_u_ms);
  PyList_SET_ITEM(__pyx_t_1, 26, __pyx_n_u_ms);
  __Pyx_INCREF(__pyx_n_u_mt);
  __Pyx_GIVEREF(__pyx_n_u_mt);
  PyList_SET_ITEM(__pyx_t_1, 27, __pyx_n_u_mt);
  __Pyx_INCREF(__pyx_n_u_mv);
  __Pyx_GIVEREF(__pyx_n_u_mv);
  PyList_SET_ITEM(__pyx_t_1, 28, __pyx_n_u_mv);
  __Pyx_INCREF(__pyx_n_u_tb);
  __Pyx_GIVEREF(__pyx_n_u_tb);
  PyList_SET_ITEM(__pyx_t_1, 29, __pyx_n_u_tb);
  __Pyx_INCREF(__pyx_n_u_tl);
  __Pyx_GIVEREF(__pyx_n_u_tl);
  PyList_SET_ITEM(__pyx_t_1, 30, __pyx_n_u_tl);
  __Pyx_INCREF(__pyx_n_u_va);
  __Pyx_GIVEREF(__pyx_n_u_va);
  PyList_SET_ITEM(__pyx_t_1, 31, __pyx_n_u_va);
  __Pyx_INCREF(__pyx_n_u_volt);
  __Pyx_GIVEREF(__pyx_n_u_volt);
  PyList_SET_ITEM(__pyx_t_1, 32, __pyx_n_u_volt);
  __Pyx_INCREF(__pyx_n_u_watt);
  __Pyx_GIVEREF(__pyx_n_u_watt);
  PyList_SET_ITEM(__pyx_t_1, 33, __pyx_n_u_watt);
  __Pyx_INCREF(__pyx_n_u_ah);
  __Pyx_GIVEREF(__pyx_n_u_ah);
  PyList_SET_ITEM(__pyx_t_1, 34, __pyx_n_u_ah);
  __Pyx_INCREF(__pyx_n_u_hp);
  __Pyx_GIVEREF(__pyx_n_u_hp);
  PyList_SET_ITEM(__pyx_t_1, 35, __pyx_n_u_hp);
  __Pyx_INCREF(__pyx_n_u_oz);
  __Pyx_GIVEREF(__pyx_n_u_oz);
  PyList_SET_ITEM(__pyx_t_1, 36, __pyx_n_u_oz);
  __Pyx_INCREF(__pyx_n_u_rpm);
  __Pyx_GIVEREF(__pyx_n_u_rpm);
  PyList_SET_ITEM(__pyx_t_1, 37, __pyx_n_u_rpm);
  __Pyx_INCREF(__pyx_n_u_dpi);
  __Pyx_GIVEREF(__pyx_n_u_dpi);
  PyList_SET_ITEM(__pyx_t_1, 38, __pyx_n_u_dpi);
  __Pyx_INCREF(__pyx_n_u_ppm);
  __Pyx_GIVEREF(__pyx_n_u_ppm);
  PyList_SET_ITEM(__pyx_t_1, 39, __pyx_n_u_ppm);
  __Pyx_INCREF(__pyx_n_u_ohm);
  __Pyx_GIVEREF(__pyx_n_u_ohm);
  PyList_SET_ITEM(__pyx_t_1, 40, __pyx_n_u_ohm);
  __Pyx_INCREF(__pyx_n_u_kwh);
  __Pyx_GIVEREF(__pyx_n_u_kwh);
  PyList_SET_ITEM(__pyx_t_1, 41, __pyx_n_u_kwh);
  __Pyx_INCREF(__pyx_n_u_kcal);
  __Pyx_GIVEREF(__pyx_n_u_kcal);
  PyList_SET_ITEM(__pyx_t_1, 42, __pyx_n_u_kcal);
  __Pyx_INCREF(__pyx_n_u_kbit);
  __Pyx_GIVEREF(__pyx_n_u_kbit);
  PyList_SET_ITEM(__pyx_t_1, 43, __pyx_n_u_kbit);
  __Pyx_INCREF(__pyx_n_u_mbit);
  __Pyx_GIVEREF(__pyx_n_u_mbit);
  PyList_SET_ITEM(__pyx_t_1, 44, __pyx_n_u_mbit);
  __Pyx_INCREF(__pyx_n_u_gbit);
  __Pyx_GIVEREF(__pyx_n_u_gbit);
  PyList_SET_ITEM(__pyx_t_1, 45, __pyx_n_u_gbit);
  __Pyx_INCREF(__pyx_n_u_bit);
  __Pyx_GIVEREF(__pyx_n_u_bit);
  PyList_SET_ITEM(__pyx_t_1, 46, __pyx_n_u_bit);
  __Pyx_INCREF(__pyx_n_u_byte);
  __Pyx_GIVEREF(__pyx_n_u_byte);
  PyList_SET_ITEM(__pyx_t_1, 47, __pyx_n_u_byte);
  __Pyx_INCREF(__pyx_n_u_mbps);
  __Pyx_GIVEREF(__pyx_n_u_mbps);
  PyList_SET_ITEM(__pyx_t_1, 48, __pyx_n_u_mbps);
  __Pyx_INCREF(__pyx_n_u_gbps);
  __Pyx_GIVEREF(__pyx_n_u_gbps);
  PyList_SET_ITEM(__pyx_t_1, 49, __pyx_n_u_gbps);
  __Pyx_INCREF(__pyx_n_u_cm3);
  __Pyx_GIVEREF(__pyx_n_u_cm3);
  PyList_SET_ITEM(__pyx_t_1, 50, __pyx_n_u_cm3);
  __Pyx_INCREF(__pyx_n_u_mm2);
  __Pyx_GIVEREF(__pyx_n_u_mm2);
  PyList_SET_ITEM(__pyx_t_1, 51, __pyx_n_u_mm2);
  __Pyx_INCREF(__pyx_n_u_mm3);
  __Pyx_GIVEREF(__pyx_n_u_mm3);
  PyList_SET_ITEM(__pyx_t_1, 52, __pyx_n_u_mm3);
  __Pyx_INCREF(__pyx_n_u_khz);
  __Pyx_GIVEREF(__pyx_n_u_khz);
  PyList_SET_ITEM(__pyx_t_1, 53, __pyx_n_u_khz);
  __Pyx_INCREF(__pyx_n_u_ft);
  __Pyx_GIVEREF(__pyx_n_u_ft);
  PyList_SET_ITEM(__pyx_t_1, 54, __pyx_n_u_ft);
  __Pyx_INCREF(__pyx_n_u_db);
  __Pyx_GIVEREF(__pyx_n_u_db);
  PyList_SET_ITEM(__pyx_t_1, 55, __pyx_n_u_db);
  __Pyx_INCREF(__pyx_n_u_sn);
  __Pyx_GIVEREF(__pyx_n_u_sn);
  PyList_SET_ITEM(__pyx_t_1, 56, __pyx_n_u_sn);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker->tp_dict, __pyx_n_s_shortcuts, __pyx_t_1) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker);

  /* "SpellChecker/SimpleSpellChecker.pyx":19
 *                          "kwh", "kcal", "kbit", "mbit", "gbit", "bit", "byte", "mbps", "gbps", "cm3", "mm2", "mm3",
 *                          "khz", "ft", "db", "sn"]
 *     __conditionalShortcuts: list = ["g", "v", "m", "l", "w", "s"]             # <<<<<<<<<<<<<<
 *     __questionSuffixList: list = ["mi", "m", "mu", "m", "miyim", "misin", "miyiz", "midir", "miydi", "mym", "msn",
 *                                   "myz", "mdr", "myd", "muyum", "musun", "muyuz", "mudur", "muydu", "mym",
 */
  __pyx_t_1 = PyList_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_g);
  __Pyx_GIVEREF(__pyx_n_u_g);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_g);
  __Pyx_INCREF(__pyx_n_u_v);
  __Pyx_GIVEREF(__pyx_n_u_v);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_v);
  __Pyx_INCREF(__pyx_n_u_m);
  __Pyx_GIVEREF(__pyx_n_u_m);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_m);
  __Pyx_INCREF(__pyx_n_u_l_3);
  __Pyx_GIVEREF(__pyx_n_u_l_3);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_l_3);
  __Pyx_INCREF(__pyx_n_u_w);
  __Pyx_GIVEREF(__pyx_n_u_w);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_w);
  __Pyx_INCREF(__pyx_n_u_s);
  __Pyx_GIVEREF(__pyx_n_u_s);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_s);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker->tp_dict, __pyx_n_s_conditionalShortcuts, __pyx_t_1) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker);

  /* "SpellChecker/SimpleSpellChecker.pyx":20
 *                          "khz", "ft", "db", "sn"]
 *     __conditionalShortcuts: list = ["g", "v", "m", "l", "w", "s"]
 *     __questionSuffixList: list = ["mi", "m", "mu", "m", "miyim", "misin", "miyiz", "midir", "miydi", "mym", "msn",             # <<<<<<<<<<<<<<
 *                                   "myz", "mdr", "myd", "muyum", "musun", "muyuz", "mudur", "muydu", "mym",
 *                                   "msn", "myz", "mdr", "myd", "miydim", "miydin", "miydik", "miymi", "mydm",
 */
  __pyx_t_1 = PyList_New(84); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_mi);
  __Pyx_GIVEREF(__pyx_n_u_mi);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_mi);
  __Pyx_INCREF(__pyx_n_u_m_2);
  __Pyx_GIVEREF(__pyx_n_u_m_2);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_m_2);
  __Pyx_INCREF(__pyx_n_u_mu);
  __Pyx_GIVEREF(__pyx_n_u_mu);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_mu);
  __Pyx_INCREF(__pyx_n_u_m_3);
  __Pyx_GIVEREF(__pyx_n_u_m_3);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_m_3);
  __Pyx_INCREF(__pyx_n_u_miyim);
  __Pyx_GIVEREF(__pyx_n_u_miyim);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_miyim);
  __Pyx_INCREF(__pyx_n_u_misin);
  __Pyx_GIVEREF(__pyx_n_u_misin);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_misin);
  __Pyx_INCREF(__pyx_n_u_miyiz);
  __Pyx_GIVEREF(__pyx_n_u_miyiz);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_miyiz);
  __Pyx_INCREF(__pyx_n_u_midir);
  __Pyx_GIVEREF(__pyx_n_u_midir);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_midir);
  __Pyx_INCREF(__pyx_n_u_miydi);
  __Pyx_GIVEREF(__pyx_n_u_miydi);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_miydi);
  __Pyx_INCREF(__pyx_n_u_mym);
  __Pyx_GIVEREF(__pyx_n_u_mym);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_mym);
  __Pyx_INCREF(__pyx_n_u_msn);
  __Pyx_GIVEREF(__pyx_n_u_msn);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_u_msn);
  __Pyx_INCREF(__pyx_n_u_myz);
  __Pyx_GIVEREF(__pyx_n_u_myz);
  PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_u_myz);
  __Pyx_INCREF(__pyx_n_u_mdr);
  __Pyx_GIVEREF(__pyx_n_u_mdr);
  PyList_SET_ITEM(__pyx_t_1, 12, __pyx_n_u_mdr);
  __Pyx_INCREF(__pyx_n_u_myd);
  __Pyx_GIVEREF(__pyx_n_u_myd);
  PyList_SET_ITEM(__pyx_t_1, 13, __pyx_n_u_myd);
  __Pyx_INCREF(__pyx_n_u_muyum);
  __Pyx_GIVEREF(__pyx_n_u_muyum);
  PyList_SET_ITEM(__pyx_t_1, 14, __pyx_n_u_muyum);
  __Pyx_INCREF(__pyx_n_u_musun);
  __Pyx_GIVEREF(__pyx_n_u_musun);
  PyList_SET_ITEM(__pyx_t_1, 15, __pyx_n_u_musun);
  __Pyx_INCREF(__pyx_n_u_muyuz);
  __Pyx_GIVEREF(__pyx_n_u_muyuz);
  PyList_SET_ITEM(__pyx_t_1, 16, __pyx_n_u_muyuz);
  __Pyx_INCREF(__pyx_n_u_mudur);
  __Pyx_GIVEREF(__pyx_n_u_mudur);
  PyList_SET_ITEM(__pyx_t_1, 17, __pyx_n_u_mudur);
  __Pyx_INCREF(__pyx_n_u_muydu);
  __Pyx_GIVEREF(__pyx_n_u_muydu);
  PyList_SET_ITEM(__pyx_t_1, 18, __pyx_n_u_muydu);
  __Pyx_INCREF(__pyx_n_u_mym_2);
  __Pyx_GIVEREF(__pyx_n_u_mym_2);
  PyList_SET_ITEM(__pyx_t_1, 19, __pyx_n_u_mym_2);
  __Pyx_INCREF(__pyx_n_u_msn_2);
  __Pyx_GIVEREF(__pyx_n_u_msn_2);
  PyList_SET_ITEM(__pyx_t_1, 20, __pyx_n_u_msn_2);
  __Pyx_INCREF(__pyx_n_u_myz_2);
  __Pyx_GIVEREF(__pyx_n_u_myz_2);
  PyList_SET_ITEM(__pyx_t_1, 21, __pyx_n_u_myz_2);
  __Pyx_INCREF(__pyx_n_u_mdr_2);
  __Pyx_GIVEREF(__pyx_n_u_mdr_2);
  PyList_SET_ITEM(__pyx_t_1, 22, __pyx_n_u_mdr_2);
  __Pyx_INCREF(__pyx_n_u_myd_2);
  __Pyx_GIVEREF(__pyx_n_u_myd_2);
  PyList_SET_ITEM(__pyx_t_1, 23, __pyx_n_u_myd_2);
  __Pyx_INCREF(__pyx_n_u_miydim);
  __Pyx_GIVEREF(__pyx_n_u_miydim);
  PyList_SET_ITEM(__pyx_t_1, 24, __pyx_n_u_miydim);
  __Pyx_INCREF(__pyx_n_u_miydin);
  __Pyx_GIVEREF(__pyx_n_u_miydin);
  PyList_SET_ITEM(__pyx_t_1, 25, __pyx_n_u_miydin);
  __Pyx_INCREF(__pyx_n_u_miydik);
  __Pyx_GIVEREF(__pyx_n_u_miydik);
  PyList_SET_ITEM(__pyx_t_1, 26, __pyx_n_u_miydik);
  __Pyx_INCREF(__pyx_n_u_miymi);
  __Pyx_GIVEREF(__pyx_n_u_miymi);
  PyList_SET_ITEM(__pyx_t_1, 27, __pyx_n_u_miymi);
  __Pyx_INCREF(__pyx_n_u_mydm);
  __Pyx_GIVEREF(__pyx_n_u_mydm);
  PyList_SET_ITEM(__pyx_t_1, 28, __pyx_n_u_mydm);
  __Pyx_INCREF(__pyx_n_u_mydn);
  __Pyx_GIVEREF(__pyx_n_u_mydn);
  PyList_SET_ITEM(__pyx_t_1, 29, __pyx_n_u_mydn);
  __Pyx_INCREF(__pyx_n_u_mydk);
  __Pyx_GIVEREF(__pyx_n_u_mydk);
  PyList_SET_ITEM(__pyx_t_1, 30, __pyx_n_u_mydk);
  __Pyx_INCREF(__pyx_n_u_mym_3);
  __Pyx_GIVEREF(__pyx_n_u_mym_3);
  PyList_SET_ITEM(__pyx_t_1, 31, __pyx_n_u_mym_3);
  __Pyx_INCREF(__pyx_n_u_muydum);
  __Pyx_GIVEREF(__pyx_n_u_muydum);
  PyList_SET_ITEM(__pyx_t_1, 32, __pyx_n_u_muydum);
  __Pyx_INCREF(__pyx_n_u_muydun);
  __Pyx_GIVEREF(__pyx_n_u_muydun);
  PyList_SET_ITEM(__pyx_t_1, 33, __pyx_n_u_muydun);
  __Pyx_INCREF(__pyx_n_u_muyduk);
  __Pyx_GIVEREF(__pyx_n_u_muyduk);
  PyList_SET_ITEM(__pyx_t_1, 34, __pyx_n_u_muyduk);
  __Pyx_INCREF(__pyx_n_u_muymu);
  __Pyx_GIVEREF(__pyx_n_u_muymu);
  PyList_SET_ITEM(__pyx_t_1, 35, __pyx_n_u_muymu);
  __Pyx_INCREF(__pyx_n_u_mydm_2);
  __Pyx_GIVEREF(__pyx_n_u_mydm_2);
  PyList_SET_ITEM(__pyx_t_1, 36, __pyx_n_u_mydm_2);
  __Pyx_INCREF(__pyx_n_u_mydn_2);
  __Pyx_GIVEREF(__pyx_n_u_mydn_2);
  PyList_SET_ITEM(__pyx_t_1, 37, __pyx_n_u_mydn_2);
  __Pyx_INCREF(__pyx_n_u_mydk_2);
  __Pyx_GIVEREF(__pyx_n_u_mydk_2);
  PyList_SET_ITEM(__pyx_t_1, 38, __pyx_n_u_mydk_2);
  __Pyx_INCREF(__pyx_n_u_mym_4);
  __Pyx_GIVEREF(__pyx_n_u_mym_4);
  PyList_SET_ITEM(__pyx_t_1, 39, __pyx_n_u_mym_4);
  __Pyx_INCREF(__pyx_n_u_misiniz);
  __Pyx_GIVEREF(__pyx_n_u_misiniz);
  PyList_SET_ITEM(__pyx_t_1, 40, __pyx_n_u_misiniz);
  __Pyx_INCREF(__pyx_n_u_msnz);
  __Pyx_GIVEREF(__pyx_n_u_msnz);
  PyList_SET_ITEM(__pyx_t_1, 41, __pyx_n_u_msnz);
  __Pyx_INCREF(__pyx_n_u_musunuz);
  __Pyx_GIVEREF(__pyx_n_u_musunuz);
  PyList_SET_ITEM(__pyx_t_1, 42, __pyx_n_u_musunuz);
  __Pyx_INCREF(__pyx_n_u_msnz_2);
  __Pyx_GIVEREF(__pyx_n_u_msnz_2);
  PyList_SET_ITEM(__pyx_t_1, 43, __pyx_n_u_msnz_2);
  __Pyx_INCREF(__pyx_n_u_miyimdir);
  __Pyx_GIVEREF(__pyx_n_u_miyimdir);
  PyList_SET_ITEM(__pyx_t_1, 44, __pyx_n_u_miyimdir);
  __Pyx_INCREF(__pyx_n_u_misindir);
  __Pyx_GIVEREF(__pyx_n_u_misindir);
  PyList_SET_ITEM(__pyx_t_1, 45, __pyx_n_u_misindir);
  __Pyx_INCREF(__pyx_n_u_miyizdir);
  __Pyx_GIVEREF(__pyx_n_u_miyizdir);
  PyList_SET_ITEM(__pyx_t_1, 46, __pyx_n_u_miyizdir);
  __Pyx_INCREF(__pyx_n_u_miydiniz);
  __Pyx_GIVEREF(__pyx_n_u_miydiniz);
  PyList_SET_ITEM(__pyx_t_1, 47, __pyx_n_u_miydiniz);
  __Pyx_INCREF(__pyx_n_u_miydiler);
  __Pyx_GIVEREF(__pyx_n_u_miydiler);
  PyList_SET_ITEM(__pyx_t_1, 48, __pyx_n_u_miydiler);
  __Pyx_INCREF(__pyx_n_u_miymiim);
  __Pyx_GIVEREF(__pyx_n_u_miymiim);
  PyList_SET_ITEM(__pyx_t_1, 49, __pyx_n_u_miymiim);
  __Pyx_INCREF(__pyx_n_u_miymiiz);
  __Pyx_GIVEREF(__pyx_n_u_miymiiz);
  PyList_SET_ITEM(__pyx_t_1, 50, __pyx_n_u_miymiiz);
  __Pyx_INCREF(__pyx_n_u_mymdr);
  __Pyx_GIVEREF(__pyx_n_u_mymdr);
  PyList_SET_ITEM(__pyx_t_1, 51, __pyx_n_u_mymdr);
  __Pyx_INCREF(__pyx_n_u_msndr);
  __Pyx_GIVEREF(__pyx_n_u_msndr);
  PyList_SET_ITEM(__pyx_t_1, 52, __pyx_n_u_msndr);
  __Pyx_INCREF(__pyx_n_u_myzdr);
  __Pyx_GIVEREF(__pyx_n_u_myzdr);
  PyList_SET_ITEM(__pyx_t_1, 53, __pyx_n_u_myzdr);
  __Pyx_INCREF(__pyx_n_u_mydnz);
  __Pyx_GIVEREF(__pyx_n_u_mydnz);
  PyList_SET_ITEM(__pyx_t_1, 54, __pyx_n_u_mydnz);
  __Pyx_INCREF(__pyx_n_u_mydlar);
  __Pyx_GIVEREF(__pyx_n_u_mydlar);
  PyList_SET_ITEM(__pyx_t_1, 55, __pyx_n_u_mydlar);
  __Pyx_INCREF(__pyx_n_u_mymm);
  __Pyx_GIVEREF(__pyx_n_u_mymm);
  PyList_SET_ITEM(__pyx_t_1, 56, __pyx_n_u_mymm);
  __Pyx_INCREF(__pyx_n_u_mymz);
  __Pyx_GIVEREF(__pyx_n_u_mymz);
  PyList_SET_ITEM(__pyx_t_1, 57, __pyx_n_u_mymz);
  __Pyx_INCREF(__pyx_n_u_muyumdur);
  __Pyx_GIVEREF(__pyx_n_u_muyumdur);
  PyList_SET_ITEM(__pyx_t_1, 58, __pyx_n_u_muyumdur);
  __Pyx_INCREF(__pyx_n_u_musundur);
  __Pyx_GIVEREF(__pyx_n_u_musundur);
  PyList_SET_ITEM(__pyx_t_1, 59, __pyx_n_u_musundur);
  __Pyx_INCREF(__pyx_n_u_muyuzdur);
  __Pyx_GIVEREF(__pyx_n_u_muyuzdur);
  PyList_SET_ITEM(__pyx_t_1, 60, __pyx_n_u_muyuzdur);
  __Pyx_INCREF(__pyx_n_u_muydunuz);
  __Pyx_GIVEREF(__pyx_n_u_muydunuz);
  PyList_SET_ITEM(__pyx_t_1, 61, __pyx_n_u_muydunuz);
  __Pyx_INCREF(__pyx_n_u_muydular);
  __Pyx_GIVEREF(__pyx_n_u_muydular);
  PyList_SET_ITEM(__pyx_t_1, 62, __pyx_n_u_muydular);
  __Pyx_INCREF(__pyx_n_u_muymuum);
  __Pyx_GIVEREF(__pyx_n_u_muymuum);
  PyList_SET_ITEM(__pyx_t_1, 63, __pyx_n_u_muymuum);
  __Pyx_INCREF(__pyx_n_u_muymuuz);
  __Pyx_GIVEREF(__pyx_n_u_muymuuz);
  PyList_SET_ITEM(__pyx_t_1, 64, __pyx_n_u_muymuuz);
  __Pyx_INCREF(__pyx_n_u_mymdr_2);
  __Pyx_GIVEREF(__pyx_n_u_mymdr_2);
  PyList_SET_ITEM(__pyx_t_1, 65, __pyx_n_u_mymdr_2);
  __Pyx_INCREF(__pyx_n_u_msndr_2);
  __Pyx_GIVEREF(__pyx_n_u_msndr_2);
  PyList_SET_ITEM(__pyx_t_1, 66, __pyx_n_u_msndr_2);
  __Pyx_INCREF(__pyx_n_u_myzdr_2);
  __Pyx_GIVEREF(__pyx_n_u_myzdr_2);
  PyList_SET_ITEM(__pyx_t_1, 67, __pyx_n_u_myzdr_2);
  __Pyx_INCREF(__pyx_n_u_mydnz_2);
  __Pyx_GIVEREF(__pyx_n_u_mydnz_2);
  PyList_SET_ITEM(__pyx_t_1, 68, __pyx_n_u_mydnz_2);
  __Pyx_INCREF(__pyx_n_u_mydler);
  __Pyx_GIVEREF(__pyx_n_u_mydler);
  PyList_SET_ITEM(__pyx_t_1, 69, __pyx_n_u_mydler);
  __Pyx_INCREF(__pyx_n_u_mymm_2);
  __Pyx_GIVEREF(__pyx_n_u_mymm_2);
  PyList_SET_ITEM(__pyx_t_1, 70, __pyx_n_u_mymm_2);
  __Pyx_INCREF(__pyx_n_u_mymz_2);
  __Pyx_GIVEREF(__pyx_n_u_mymz_2);
  PyList_SET_ITEM(__pyx_t_1, 71, __pyx_n_u_mymz_2);
  __Pyx_INCREF(__pyx_n_u_miymisin);
  __Pyx_GIVEREF(__pyx_n_u_miymisin);
  PyList_SET_ITEM(__pyx_t_1, 72, __pyx_n_u_miymisin);
  __Pyx_INCREF(__pyx_n_u_miymiler);
  __Pyx_GIVEREF(__pyx_n_u_miymiler);
  PyList_SET_ITEM(__pyx_t_1, 73, __pyx_n_u_miymiler);
  __Pyx_INCREF(__pyx_n_u_mymsn);
  __Pyx_GIVEREF(__pyx_n_u_mymsn);
  PyList_SET_ITEM(__pyx_t_1, 74, __pyx_n_u_mymsn);
  __Pyx_INCREF(__pyx_n_u_mymlar);
  __Pyx_GIVEREF(__pyx_n_u_mymlar);
  PyList_SET_ITEM(__pyx_t_1, 75, __pyx_n_u_mymlar);
  __Pyx_INCREF(__pyx_n_u_muymusun);
  __Pyx_GIVEREF(__pyx_n_u_muymusun);
  PyList_SET_ITEM(__pyx_t_1, 76, __pyx_n_u_muymusun);
  __Pyx_INCREF(__pyx_n_u_muymular);
  __Pyx_GIVEREF(__pyx_n_u_muymular);
  PyList_SET_ITEM(__pyx_t_1, 77, __pyx_n_u_muymular);
  __Pyx_INCREF(__pyx_n_u_mymsn_2);
  __Pyx_GIVEREF(__pyx_n_u_mymsn_2);
  PyList_SET_ITEM(__pyx_t_1, 78, __pyx_n_u_mymsn_2);
  __Pyx_INCREF(__pyx_n_u_mymler);
  __Pyx_GIVEREF(__pyx_n_u_mymler);
  PyList_SET_ITEM(__pyx_t_1, 79, __pyx_n_u_mymler);
  __Pyx_INCREF(__pyx_n_u_misinizdir);
  __Pyx_GIVEREF(__pyx_n_u_misinizdir);
  PyList_SET_ITEM(__pyx_t_1, 80, __pyx_n_u_misinizdir);
  __Pyx_INCREF(__pyx_n_u_msnzdr);
  __Pyx_GIVEREF(__pyx_n_u_msnzdr);
  PyList_SET_ITEM(__pyx_t_1, 81, __pyx_n_u_msnzdr);
  __Pyx_INCREF(__pyx_n_u_musunuzdur);
  __Pyx_GIVEREF(__pyx_n_u_musunuzdur);
  PyList_SET_ITEM(__pyx_t_1, 82, __pyx_n_u_musunuzdur);
  __Pyx_INCREF(__pyx_n_u_msnzdr_2);
  __Pyx_GIVEREF(__pyx_n_u_msnzdr_2);
  PyList_SET_ITEM(__pyx_t_1, 83, __pyx_n_u_msnzdr_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker->tp_dict, __pyx_n_s_questionSuffixList, __pyx_t_1) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_12SpellChecker_18SimpleSpellChecker_SimpleSpellChecker);

  /* "(tree fragment)":1
 * def __pyx_unpickle_SimpleSpellChecker(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_12SpellChecker_18SimpleSpellChecker_1__pyx_unpickle_SimpleSpellChecker, NULL, __pyx_n_s_SpellChecker_SimpleSpellChecker); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_SimpleSpellChecke, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "SpellChecker/SimpleSpellChecker.pyx":1
 * from random import randrange             # <<<<<<<<<<<<<<
 * import re
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init SpellChecker.SimpleSpellChecker", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init SpellChecker.SimpleSpellChecker");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyUnicode_Substring */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Substring(
            PyObject* text, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length;
    if (unlikely(__Pyx_PyUnicode_READY(text) == -1)) return NULL;
    length = __Pyx_PyUnicode_GET_LENGTH(text);
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    if (stop <= start)
        return __Pyx_NewRef(__pyx_empty_unicode);
#if CYTHON_PEP393_ENABLED
    return PyUnicode_FromKindAndData(PyUnicode_KIND(text),
        PyUnicode_1BYTE_DATA(text) + start*PyUnicode_KIND(text), stop-start);
#else
    return PyUnicode_FromUnicode(PyUnicode_AS_UNICODE(text)+start, stop-start);
#endif
}

/* GetItemIntUnicode */
static CYTHON_INLINE Py_UCS4 __Pyx_GetItemInt_Unicode_Fast(PyObject* ustring, Py_ssize_t i,
                                                           int wraparound, int boundscheck) {
    Py_ssize_t length;
    if (unlikely(__Pyx_PyUnicode_READY(ustring) < 0)) return (Py_UCS4)-1;
    if (wraparound | boundscheck) {
        length = __Pyx_PyUnicode_GET_LENGTH(ustring);
        if (wraparound & unlikely(i < 0)) i += length;
        if ((!boundscheck) || likely(__Pyx_is_valid_index(i, length))) {
            return __Pyx_PyUnicode_READ_CHAR(ustring, i);
        } else {
            PyErr_SetString(PyExc_IndexError, "string index out of range");
            return (Py_UCS4)-1;
        }
    } else {
        return __Pyx_PyUnicode_READ_CHAR(ustring, i);
    }
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* pop_index */
static PyObject* __Pyx__PyObject_PopNewIndex(PyObject* L, PyObject* py_ix) {
    PyObject *r;
    if (unlikely(!py_ix)) return NULL;
    r = __Pyx__PyObject_PopIndex(L, py_ix);
    Py_DECREF(py_ix);
    return r;
}
static PyObject* __Pyx__PyObject_PopIndex(PyObject* L, PyObject* py_ix) {
    return __Pyx_PyObject_CallMethod1(L, __pyx_n_s_pop, py_ix);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static PyObject* __Pyx__PyList_PopIndex(PyObject* L, PyObject* py_ix, Py_ssize_t ix) {
    Py_ssize_t size = PyList_GET_SIZE(L);
    if (likely(size > (((PyListObject*)L)->allocated >> 1))) {
        Py_ssize_t cix = ix;
        if (cix < 0) {
            cix += size;
        }
        if (likely(__Pyx_is_valid_index(cix, size))) {
            PyObject* v = PyList_GET_ITEM(L, cix);
            __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
            size -= 1;
            memmove(&PyList_GET_ITEM(L, cix), &PyList_GET_ITEM(L, cix+1), (size_t)(size-cix)*sizeof(PyObject*));
            return v;
        }
    }
    if (py_ix == Py_None) {
        return __Pyx__PyObject_PopNewIndex(L, PyInt_FromSsize_t(ix));
    } else {
        return __Pyx__PyObject_PopIndex(L, py_ix);
    }
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* unicode_tailmatch */
static int __Pyx_PyUnicode_TailmatchTuple(PyObject* s, PyObject* substrings,
                                          Py_ssize_t start, Py_ssize_t end, int direction) {
    Py_ssize_t i, count = PyTuple_GET_SIZE(substrings);
    for (i = 0; i < count; i++) {
        Py_ssize_t result;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        result = PyUnicode_Tailmatch(s, PyTuple_GET_ITEM(substrings, i),
                                     start, end, direction);
#else
        PyObject* sub = PySequence_ITEM(substrings, i);
        if (unlikely(!sub)) return -1;
        result = PyUnicode_Tailmatch(s, sub, start, end, direction);
        Py_DECREF(sub);
#endif
        if (result) {
            return (int) result;
        }
    }
    return 0;
}
static int __Pyx_PyUnicode_Tailmatch(PyObject* s, PyObject* substr,
                                     Py_ssize_t start, Py_ssize_t end, int direction) {
    if (unlikely(PyTuple_Check(substr))) {
        return __Pyx_PyUnicode_TailmatchTuple(s, substr, start, end, direction);
    }
    return (int) PyUnicode_Tailmatch(s, substr, start, end, direction);
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_dealloc != current_tp_dealloc)
        type = type->tp_base;
    while (type && type->tp_dealloc == current_tp_dealloc)
        type = type->tp_base;
    if (type)
        type->tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* PyUCS4InUnicode */
#if PY_VERSION_HEX < 0x03090000
#if Py_UNICODE_SIZE == 2
static int __Pyx_PyUnicodeBufferContainsUCS4_SP(Py_UNICODE* buffer, Py_ssize_t length, Py_UCS4 character) {
    Py_UNICODE high_val, low_val;
    Py_UNICODE* pos;
    high_val = (Py_UNICODE) (0xD800 | (((character - 0x10000) >> 10) & ((1<<10)-1)));
    low_val  = (Py_UNICODE) (0xDC00 | ( (character - 0x10000)        & ((1<<10)-1)));
    for (pos=buffer; pos < buffer+length-1; pos++) {
        if (unlikely((high_val == pos[0]) & (low_val == pos[1]))) return 1;
    }
    return 0;
}
#endif
static int __Pyx_PyUnicodeBufferContainsUCS4_BMP(Py_UNICODE* buffer, Py_ssize_t length, Py_UCS4 character) {
    Py_UNICODE uchar;
    Py_UNICODE* pos;
    uchar = (Py_UNICODE) character;
    for (pos=buffer; pos < buffer+length; pos++) {
        if (unlikely(uchar == pos[0])) return 1;
    }
    return 0;
}
#endif
static CYTHON_INLINE int __Pyx_UnicodeContainsUCS4(PyObject* unicode, Py_UCS4 character) {
#if CYTHON_PEP393_ENABLED
    const int kind = PyUnicode_KIND(unicode);
    if (likely(kind != PyUnicode_WCHAR_KIND)) {
        Py_ssize_t i;
        const void* udata = PyUnicode_DATA(unicode);
        const Py_ssize_t length = PyUnicode_GET_LENGTH(unicode);
        for (i=0; i < length; i++) {
            if (unlikely(character == PyUnicode_READ(kind, udata, i))) return 1;
        }
        return 0;
    }
#elif PY_VERSION_HEX >= 0x03090000
    #error Cannot use "UChar in Unicode" in Python 3.9 without PEP-393 unicode strings.
#endif
#if PY_VERSION_HEX < 0x03090000
#if Py_UNICODE_SIZE == 2
    if (unlikely(character > 65535)) {
        return __Pyx_PyUnicodeBufferContainsUCS4_SP(
            PyUnicode_AS_UNICODE(unicode),
            PyUnicode_GET_SIZE(unicode),
            character);
    } else
#endif
    {
        return __Pyx_PyUnicodeBufferContainsUCS4_BMP(
            PyUnicode_AS_UNICODE(unicode),
            PyUnicode_GET_SIZE(unicode),
            character);
    }
#endif
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
